/* Various Full (Max,Plus) matrix manipulation procedures */

#include <stdlib.h>
#include <stdio.h>
#include "boolean.h"
#include "maxpluscore.h"
#include "maxplusmat.h"
#include "fullmat.h"

#define TOOLNAME		"FullMat"

/* constructor and destructor */
Full_MaxPlus_Matrix New_Full_MaxPlus_Matrix( int Rows, 
					     int Cols,
					     int Init_Mode )
{
  Full_MaxPlus_Matrix	Res;
  int			i;
  boolean		Alloc_Ok;

  Res = (Full_MaxPlus_Matrix) malloc( sizeof( struct FULL_MAXPLUS_MATRIX ) );
  if ( (Full_MaxPlus_Matrix)NULL != Res ) {
    Alloc_Ok = TRUE;
    Res->Rows = Rows;
    Res->Cols = Cols;
    Res->Vals = (double**) calloc( (size_t)Rows, sizeof(double*) );
    Alloc_Ok = ( (double**)NULL != Res->Vals );
    for ( i=0; Alloc_Ok && (i<Rows); i++ ) {
      Res->Vals[i] = (double*) calloc( (size_t)Cols, sizeof(double) );
      Alloc_Ok = ( (double*)NULL != Res->Vals[i] );
    }
    if ( Alloc_Ok ) {
      Alloc_Ok = Init_Full_MaxPlus_Matrix( Res, Init_Mode );
    }
    if ( !Alloc_Ok ) {
      Free_Full_MaxPlus_Matrix( Res );
      Res = (Full_MaxPlus_Matrix)NULL;
    }
  }

  return Res;
}

void Free_Full_MaxPlus_Matrix(Full_MaxPlus_Matrix A)
{
  int		i;

  if ( (Full_MaxPlus_Matrix)NULL != A ) {
    if ( (double**)NULL != A->Vals ) {
      for ( i=0; i<A->Rows; i++ ) {
	if ( (double*)NULL != A->Vals[i] ) {
	  free( A->Vals[i] );
	}
      }
      free( A->Vals );
    }
    free( A );
  }
}

boolean Init_Full_MaxPlus_Matrix(Full_MaxPlus_Matrix 	A,
				 int			Init_Mode)
{
  int		i;
  int		j;
  boolean 	Ok = TRUE;

  /* initialization */
  switch( Init_Mode ) {
  case ZERO:
    for ( i=0; i<A->Rows; i++ ) {
      for ( j=0; j<A->Cols; j++ ) {
	A->Vals[i][j] = M_INFINITY;
      }
    }
    break;
  case ONE:
    for ( i=0; i<A->Rows; i++ ) {
      for ( j=0; j<A->Cols; j++ ) {
	A->Vals[i][j] = 0.0;
      }
    }
    break;
  case IDENTITY:
    for ( i=0; i<A->Rows; i++ ) {
      for ( j=0; j<A->Cols; j++ ) {
	A->Vals[i][j] = ( i == j ? 0.0 : M_INFINITY );
      }
    }
    break;
  case NONE:
    /* none ! */
    break;
  default:
    printf("%s: error: illegal init mode (%d) for matrices.\n",
	   TOOLNAME, Init_Mode );
    Ok = FALSE;
  }

  return Ok;
}

/* Create a (max,+) and initialize it out of a double[][] array.	*/
Full_MaxPlus_Matrix Make_Full_MaxPlus_Matrix( int Rows, int Cols, 
					      double **Init )
{
  int			i;
  int			j;
  Full_MaxPlus_Matrix	Res;

  Res = New_Full_MaxPlus_Matrix( Rows, Cols, NONE );
  if ( (Full_MaxPlus_Matrix)NULL != Res ) {
    for ( i=0; i<Res->Rows; i++ )
      for ( j=0; j<Res->Cols; j++ )
	Res->Vals[i][j] = Init[i][j];
  }

  return Res;
}

Full_MaxPlus_Matrix Copy_Full_MaxPlus_Matrix( Full_MaxPlus_Matrix M )
{
  return Make_Full_MaxPlus_Matrix( M->Rows, M->Cols, M->Vals );
}

/* Retrieving an entry from the matrix. Trivial for full matrices but	*/
/* useful for code compatibility!					*/
double Get_Full_Matrix_Entry(Full_MaxPlus_Matrix	A, 
			     int			Row,
			     int			Col)
{
  if ( ( 0 <= Row ) && ( Row < A->Rows ) 
       && ( 0 <= Col ) && ( Col < A->Cols ) ) {
    return A->Vals[Row][Col];
  }
  else {
    printf("%s: warning: illegal entry (%d,%d) for a %d x %d matrix.",
	   TOOLNAME, Row, Col, A->Rows, A->Cols );
    printf(" eps assumed.\n");
    return M_INFINITY;
  }
}

/* Giving a value to a matrix entry. Quite trivial, except for bounds	*/
/* checking.								*/
boolean Set_Full_Matrix_Entry(Full_MaxPlus_Matrix	A, 
			      int			Row,
			      int			Col,
			      double			Val )
{
  if ( ( 0 <= Row ) && ( Row < A->Rows ) 
       && ( 0 <= Col ) && ( Col < A->Cols ) ) {
    A->Vals[Row][Col] = Val;
    return TRUE;
  }
  else {
    printf("Warning: setting value to (%d,%d) in a %d x %d matrix.",
	   Row, Col, A->Rows, A->Cols );
    printf(" Ignored.\n");
    return FALSE;
  }
}

/* Reading from a file, and adding an element to the matrix */
boolean Full_Read_Full_Matrix_Entry(FILE		*Input, 
				    Full_MaxPlus_Matrix	A, 
				    int			Row,
				    int			Col)
{
  char		Entry[256];
  double	Val;

  fscanf( Input, "%s", Entry );
  if ( 1 == sscanf( Entry, "%lf", &Val ) ) {
    if ( Val < M_INFINITY ) {
      fprintf(stderr,"%s: warning: matrix entry (%e) less than eps.",
	      TOOLNAME, Val );
      fprintf(stderr," eps assumed.\n");
      A->Vals[Col][Row] = M_INFINITY;
    }
    else
      A->Vals[Col][Row] = Val;
    return TRUE;
  }
  else if ( !strcmp( Entry, "eps" ) ) {
    A->Vals[Col][Row] = M_INFINITY;
    return TRUE;
  }
  else {
    return FALSE;
  }
}

boolean Full_Read_SP_Matrix_Entry(FILE			*Input, 
				  Full_MaxPlus_Matrix	A)
{
  double	Val;
  int		Row;
  int		Col;

  if ( 3 == fscanf( Input, "%d %d %lf", &Row, &Col, &Val ) ) {
    if ( ( 0 <= Row ) && ( Row < A->Rows ) 
	 && ( 0 <= Col ) && ( Col < A->Cols ) ) {
      A->Vals[Row][Col] = Val;
      return TRUE;
    }
    else {
      printf("%s: Warning: failed to set entry (%d,%d) to %f.\n", 
	     TOOLNAME, Row, Col, Val );
      return FALSE;
    }
  }
  else {
    return FALSE;
  }
}

boolean Full_Read_Full_MaxPlus_Matrix( FILE		*Input,
				       Full_MaxPlus_Matrix A)
{
  int	i;
  int	j;
  boolean Ok;

  Ok = TRUE;

  for ( i=0; (i<A->Rows) && Ok; i++ ) {
    for ( j=0; (j<A->Cols) && Ok; j++ ) {
      if ( !Full_Read_Full_Matrix_Entry( Input, A, i, j ) ) {
	printf("%s: error: unexpected value (%d,%d) for matrix. Abort.\n",
	       TOOLNAME, i, j );
	Ok = FALSE;
      }
    }
  }
  return Ok;
}

boolean Full_Read_SP_MaxPlus_Matrix( FILE		*Input,
				     Full_MaxPlus_Matrix A )
{
  boolean Ok;
  char	Token[128];

  Ok = TRUE;

  do {
    Ok = Full_Read_SP_Matrix_Entry( Input, A );
  } while ( Ok );
  /* skip stopping mark */
  fscanf( Input, "%s", Token );

  return TRUE;
}

boolean Full_Read_MaxPlus_Matrix( FILE			*Input,
				  Full_MaxPlus_Matrix	A )
{
  int		i;
  int		j;
  boolean 	Ok;
  double	Val;
  char		Token[128];

  /* I write one single procedure with reading the two types of matrices */
  /* because I could not solve the problem of the first token, which	*/
  /* may be either the specifyer "sp" or an entry of the matrix.	*/

  /* Reading the first token to determine if the matrix is sparse. */
  fscanf( Input, "%128s", Token );
  if ( !strcmp( Token, "sp" ) ) {
    do {
      Ok = Full_Read_SP_Matrix_Entry( Input, A );
    } while ( Ok );
    /* skip stopping mark */
    fscanf( Input, "%s", Token );
    Ok = TRUE;
  }
  else {
    /* Must be full. Proceed with the reading, but skip the first read. */
    for ( i=0; Ok && (i<A->Rows); i++ ) {
      for ( j=0; Ok && (j<A->Cols); j++ ) {
	if ( ( i != 0 ) || ( j != 0 ) ) 
	  fscanf( Input, "%s", Token );

	/* convert value and store it */
	if ( 1 == sscanf( Token, "%lf", &Val ) ) {
	  if ( Val < M_INFINITY ) {
	    fprintf(stderr, "%s: warning: matrix entry (%e) less than eps.",
		    TOOLNAME, Val );
	    fprintf(stderr," eps assumed.\n");
	    A->Vals[i][j] = M_INFINITY;
	  }
	  else
	    A->Vals[i][j] = Val;
	}
	else if ( !strcmp( Token, "eps" ) ) {
	  A->Vals[i][j] = M_INFINITY;
	}
	else {
	  printf("%s: illegal matrix entry (%d,%d): %s.\n", 
		 TOOLNAME, i, j, Token );
	  Ok = FALSE;
	}
      }
    }
  }

  return Ok;
}

/************************************************************************/
/* Writing out a matrix							*/
/************************************************************************/
void Write_Full_MaxPlus_Matrix(Full_MaxPlus_Matrix	A,
			       FILE			*Output,
			       int			Format,
			       char			*Name )
{
  int i;
  int j;

  if ( (Full_MaxPlus_Matrix)NULL == A ) {
    fprintf( stderr, "%s: error: cannot print null matrix.\n", TOOLNAME );
    return;
  }

  if ( Format == FORMAT_SPARSE )
    fprintf( Output, "sp\n" );
  else if ( Format == FORMAT_SCILAB ) 
    fprintf( Output, "%s=#([\n\t", Name );
  else if ( Format == FORMAT_DEBUG ) 
    fprintf( Output, "Matrix %s\n", 
	     ( Name == (char*)NULL ? "<no name>" : Name ) );

  for ( i=0; i<A->Rows; i++ ) {

    switch ( Format ) {
    case FORMAT_FULL:
    case FORMAT_DEBUG:
      for ( j=0; j<A->Cols; j++ ) {
	if ( A->Vals[i][j] == M_INFINITY ) {
	  fprintf( Output, "    eps   ");
	}
	else
	  fprintf( Output, " %8.4f", A->Vals[i][j] );
	if ( (j+1) % 7 == 0 )
	  fprintf( Output, "\n");
      }
      fprintf( Output, "\n");
      break;
    case FORMAT_SPARSE:
      for ( j=0; j<A->Cols; j++ ) {
	if ( A->Vals[i][j] != M_INFINITY ) {
	  fprintf( Output, "%2d %2d %8.4f\n", i, j, A->Vals[i][j] );
	}
      }
      break;
    case FORMAT_SCILAB:
      for ( j=0; j<A->Cols; j++ ) {
	if ( A->Vals[i][j] == M_INFINITY ) 
	  fprintf( Output, "   %%0    ");
	else
	  fprintf( Output, " %8.4f", A->Vals[i][j] );
      }
      if ( i < A->Rows-1 )
	fprintf( Output, ";\n\t");
      else
	fprintf( Output, "])\n");
      break;
    }
  }

  if ( Format == FORMAT_SPARSE )
    fprintf( Output, "end\n" );

}
  
/************* Procedure to compute C = A \otimes B , size S ********/
void Full_MM_Mult(Full_MaxPlus_Matrix A, 
		  Full_MaxPlus_Matrix B,
		  Full_MaxPlus_Matrix C )
{
  int i;
  int j;
  int k;
  double x;

  /* checking size compatibility */
  if ( ( A->Cols == B->Rows ) 
       && ( A->Rows == C->Rows )
       && ( B->Cols == C->Cols ) ) {

    for ( i = 0; i < C->Rows; i++ )
      for ( j = 0; j < C->Cols; j++ ) {

	x = M_INFINITY;

	for ( k=0; k < A->Cols; k++ ) {
	  x = max( plus(A->Vals[i][k],B->Vals[k][j]), x );
	}

	C->Vals[i][j] = x;
      }
  }
  else {
    fprintf( stderr, "%s: error: incompatible matrix sizes. No product.\n",
	     TOOLNAME );
  }
}

/************* Procedure to compute C = A \oplus B , size S ********/
void Full_MM_Sum(Full_MaxPlus_Matrix A, 
		 Full_MaxPlus_Matrix B,
		 Full_MaxPlus_Matrix C )
{
  int i;
  int j;

  /* checking size compatibility */
  if ( ( A->Rows == B->Rows ) 
       && ( A->Rows == C->Rows ) 
       && ( A->Cols == B->Cols )
       && ( A->Cols == C->Cols ) ) {

    for ( i = 0; i < A->Rows; i++ )
      for ( j = 0; j < B->Cols; j++ )
        C->Vals[i][j] = max( A->Vals[i][j], B->Vals[i][j]);
  }
  else {
    fprintf( stderr, "%s: error: incompatible matrix sizes. No sum.\n", 
	     TOOLNAME );
  }
}

/************* Procedure to compute C = A + B , size S ********/
void Full_MM_Standard_Sum(Full_MaxPlus_Matrix A, 
			  Full_MaxPlus_Matrix B,
			  Full_MaxPlus_Matrix C )
{
  int i;
  int j;

  if ( ( A->Rows == B->Rows ) 
       && ( A->Rows == C->Rows ) 
       && ( A->Cols == B->Cols )
       && ( A->Cols == C->Cols ) ) {

    for ( i = 0; i < A->Rows; i++ )
      for ( j = 0; j < A->Cols; j++ )
        C->Vals[i][j] = plus( A->Vals[i][j], B->Vals[i][j]);
  }
  else {
    fprintf( stderr, "%s: error: incompatible matrix sizes. No sum.\n", 
	     TOOLNAME );
  }
}

void Full_Standard_Scalar_Product( Full_MaxPlus_Matrix A, 
				   Full_MaxPlus_Matrix B, 
				   double 		lambda )
{
  int i;
  int j;

  /* checking size compatibility */
  if ( ( A->Rows == B->Rows ) 
       && ( A->Cols == B->Cols ) ) {

    for ( i = 0; i < A->Rows; i++ )
      for ( j = 0; j < A->Cols; j++ )
	B->Vals[i][j] = ( A->Vals[i][j] == M_INFINITY ?
			  M_INFINITY :
			  lambda * A->Vals[i][j] );
  }
  else {
    fprintf( stderr, 
	     "%s: error: incompatible matrix sizes. No scalar product.\n",
	     TOOLNAME);
  }
}

/******* The "norm" of a matrix is the maximal value of its elements */
double Full_Matrix_Norm(Full_MaxPlus_Matrix A)
{
  int i;
  int j;
  double Res;

  Res = M_INFINITY;

  for ( i = 0; i < A->Rows; i++ )
    for ( j = 0; j < A->Cols; j++ ) {

      Res = max( Res, A->Vals[i][j] );
    }

  return Res;
}

double Full_Matrix_Min(Full_MaxPlus_Matrix A)
{
  int i;
  int j;
  double Res;

  Res = A->Vals[0][0];

  for ( i = 0; i < A->Rows; i++ )
    for ( j = 0; j < A->Cols; j++ ) {

      Res = min( Res, A->Vals[i][j] );
    }

  return Res;
}
