/* program/module for computinf bounds on the growth rate of an		*/
/* interaction graph.							*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "boolean.h"
#include "parseargs.h"
#include "i_graph.h"

#define	TOOLNAME	"gr_bound"
#define VERSION		"1.1"

#define max(x,y)        ( (x>y) ? x : y )
#define min(x,y)        ( (x<y) ? x : y )

#define CODE(x)		printf(x);
#define CODE1(x,y)	printf(x,y);
#define CODE2(x,y,z)	printf(x,y,z);
#define CODE3(x,y,z,t)	printf(x,y,z,t);
#define CODELN		CODE("\n")

#define MAX_NAME_LENGTH	256

/* Global declarations */
extern char	*todays_date( void );

/* Forward declarations */
void	Compute_Bounds(Interaction_Graph);

boolean	Debug = FALSE;
double	Epsilon = 1e-6;

int main(int argc, char *argv[] )
{
  static char	In_File_Name[MAX_NAME_LENGTH] = "";
  static char	Out_File_Name[MAX_NAME_LENGTH] = "";

  int		i;
  boolean	Help = FALSE;
  boolean	Args_Ok = TRUE;
  Interaction_Graph	G;

  i = 1;

  if (argc>1)
    do {
      if GET_STRING( "-f", In_File_Name )

      else if GET_STRING( "-o", Out_File_Name )

      else if GET_REAL( "-epsilon", Epsilon )

      else if SWITCH( "-debug", Debug, TRUE )		       

      else if IS_VERSION

      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-f <file>] [-o <file>]", TOOLNAME );
	printf(" [-epsilon <real>]");
	printf(" [-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    } while ( Args_Ok && ( i < argc ) );

  if (Args_Ok) {
    if ( strlen(In_File_Name) != 0 )
      freopen( In_File_Name, "r", stdin );

    G = Read_Interaction_Graph( stdin );

    if ( (Interaction_Graph)NULL != G ) {
      Compute_Bounds( G );
      return 0;
    }
    else
      fprintf( stderr, "%s: bad input data. No computation.\n", TOOLNAME );

    return 0;
  }
  else if ( Help )
    return 0;
  else
    return -1;
}

/* Spectral radius of a graph given by its adj matrix (estimated) */
double Spectral_Radius( Boolean_Matrix A, double Epsilon )
{
  int		i;
  int		j;
  double	*X;
  double	*Y;
  double	*Tmp;
  double	Rho;
  double	Rho_E;
  double	Dist;
  int 		Max_Iter_Nb = 500;
  int		Iter_Nb = 0;

  X = (double*) calloc( (unsigned)A->Size, sizeof(double) );
  Y = (double*) calloc( (unsigned)A->Size, sizeof(double) );
  for ( i=0; i<A->Size; i++ )
    X[i] = 1.0;

  do { 
    for ( i=0; i<A->Size; i++ ) {
      Y[i] = 0.0;
#ifdef USE_SP_BOOL_MATRICES
      for ( j=0; j<A->Nb_Elts[i]; j++ )
	Y[i] += X[A->Elts[i][j]];
#else
      for ( j=0; j<A->Size; j++ )
	Y[i] += A->Vals[i][j] * X[j];
#endif
    }
    /* normalize Y vector */
    Rho = Y[0];
    Dist = 0.0;
    for ( i=0; i<A->Size; i++ ) {
      Y[i] = Y[i] / Rho;
      Dist += fabs( X[i] - Y[i] );
    }
    /* swap X and Y */
    Tmp = X; X = Y; Y = Tmp;
    Iter_Nb ++;

  } while ( ( Max_Iter_Nb > Iter_Nb ) && ( Dist > Epsilon ) );

  /* The value of Rho could be returned as an ESTIMATE of the sp. rad.	*/
  /* Calculating the Collatz-Wielandt function at the X vector gives 	*/
  /* an UPPER BOUND of the spectral radius (cf. Minc).			*/
  for ( i=0; i<A->Size; i++ ) {
    Y[i] = 0.0;
#ifdef USE_SP_BOOL_MATRICES
    for ( j=0; j<A->Nb_Elts[i]; j++ )
      Y[i] += X[A->Elts[i][j]];
#else
    for ( j=0; j<A->Size; j++ )
      Y[i] += A->Vals[i][j] * X[j];
#endif
  }
  Rho_E = Rho;
  Rho = A->Size;
  for ( i=0; i<A->Size; i++ ) {
    if ( X[i] != 0.0 ) {
      if ( Y[i] / X[i] < Rho )
	Rho = Y[i] / X[i];
    }
  }
  
  if ( Debug ) {
    printf("Number of iterations = %d\n", Iter_Nb );
    printf("Spectral radius <= %f (estimated %f)\n", Rho, Rho_E );
  }

  if ( Max_Iter_Nb == Iter_Nb ) {
    printf("%s: warning: max number of iterations reached. ", TOOLNAME );
    printf("Spectral radius may be inaccurate.\n");
  }
    
  return Rho;
    
}

/* The Brilman f function (see Theorem 10 p. 81). */
double f( double x ) 
{
  if ( x == 0.0 ) 
    return 0.0;
  else if ( x == 1.0 )
    return 1.0;
  else if ( ( x > 0.0 ) && ( x < 1.0 ) )
    return exp( log(x) + (1.0-x)*log(1.0-x)/x );
  else /* illegal value. return default */
    return 0.0;
}

/* Derivative of the f function						*/
/* The logarithmic derivative gives: 	f'/f = -log(1-x)/x^2.		*/
double f1( double x ) 
{
  if ( x == 0.0 ) 
    return exp(-1.0);
  else if ( ( x > 0.0 ) && ( x < 1.0 ) )
    return -f(x)*x*x/log(1.0-x);
  else /* illegal value. return default */
    return 0.0;
}

/* inverse of the function. obtained by dichoyomy. Newton's algo on	*/
/* function f(x)-y does not work well because of the infinite slope 	*/
/* at x=1 :-( 								*/
double inv_f( double y )
{
  double X0;
  double X1;
  double F;
  /*  double F1; */
  double Xn;

/* Newton's algo... now useless but kept in case of. */

  if ( Debug ) {
    printf("Computing f^(-1) ( %f )\n", y );
  }
  if ( ( y < 0.0 ) || ( y > 1.0 ) ) {
    fprintf( stderr, 
	     "%s: error: attempt to compute f^(-1) with out of range value %f.\n", 
	     TOOLNAME, y );
    return 0.0;
  }

  Xn = 0.0;
  F = -y;	/* f ( 0.0 ) = 0.0 */

/*   do { */
/*     F1 = f1( Xn ); */
/*     if ( Debug ) { */
/*       printf("NEWT x=%f f=%f f'=%f\n", Xn, F, F1 ); */
/*     } */
/*     Xn += - F / F1; */
/*     F = f( Xn ) - y; */

/*   } while ( fabs( F ) > Epsilon ); */

  X0 = 0.0;
  X1 = 1.0;
  Xn = 0.5;
  F = f( Xn ) - y;

  do { 
    if ( F < 0.0 ) {
      X0 += (Xn-X0)/2.0;
    }
    else if ( F > 0.0 ) {
      X1 -= (X1-Xn)/2.0;
    }
    Xn = (X0+X1)/2.0;
    F = f( Xn ) - y;
    if ( Debug ) {
      printf("DICHO x=[%f,%f] f=%f\n", X0, X1, F );
    }
  } while( fabs(F) > Epsilon );

  return Xn;

}

void Compute_Bounds(Interaction_Graph G)
{
  int		i;
  int		j;
  double	UB;
  double	LB;
  boolean	Uniform;
  double	Sum;
  double	Max;
  double	Tot_Weight;

  Uniform = Is_Uniform( G );

  if ( Uniform ) {
    /* Upper bound based on Brilman's f function and the maximal degree	*/
    /* see proof of Corol. 3.6.1 					*/
    UB = inv_f( Max_Degree( G->A ) / (double)G->Size );
    if ( Debug ) {
      printf("UB using max degree (=%d): %f\n", 
	     Max_Degree( G->A ), UB );
    }	     
    /* The better bound with the spectral radius will come later...	*/
    UB = inv_f( Spectral_Radius( G->A, Epsilon ) / (double)G->Size );

  } else {
    /* Weaker upper bound for general graphs: Corol. 3.6.2.		*/
    Tot_Weight = 0;
    Max = 0.0;
    for ( i=0; i<G->Size; i++ ) {
      Tot_Weight += G->Weight[i];
      Sum = 0.0;
      for ( j=0; j<G->Size; j++ ) {
	if ( Is_Adj( G->A, i, j ) ) {
	  Sum += G->Weight[j];
	}
      }
      if ( Sum > Max )
	Max = Sum;
    }

    UB = exp(1.0) * Max / Tot_Weight;
  }

  if ( Uniform ) {
    /* Lower bound based on average degree: Theorem 11			*/
    LB = Avg_Degree( G->A ) / (double)G->Size;
  } 
  else {
    /* General lower bound: Corol. 3.6.3 p. 86.  			*/
    /* Note that the interpretation of p in Brilman is NOT weights, so	*/
    /* that gamma(G,p) of Brilman is our gamma * ||p||.			*/
    Tot_Weight = 0;
    Sum = 0.0;
    for ( i=0; i<G->Size; i++ ) {
      Tot_Weight += G->Weight[i];
      for ( j=0; j<G->Size; j++ ) {
	if ( Is_Adj( G->A, i, j ) ) {
	  Sum += G->Weight[i] * G->Weight[j];
	}
      }
    }

    LB = Sum / Tot_Weight / Tot_Weight;
  }
  
  printf("%f %f\n", LB, UB );
}
