#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "boolean.h"
#include "parseargs.h"
#include "tetris.h"
#include "mc.h"

#define TOOLNAME	"tet_bound"
/* #define VERSION 	"1.0 (14/12/98)" */

#define MAX_NAME_LENGTH 256
#define MAX(x,y) ( x > y ? x : y )
#define MIN(x,y) ( x < y ? x : y )
#define SCAN(ptr,list) for ( ptr = list; ptr != NULL; ptr=ptr->Next )

typedef struct STATE_DESC {
  int	Number;
  int*	Array;
} *State_Desc;

typedef struct STATE_LIST {
  State_Desc	Val;
  struct STATE_LIST *Next;
} *State_List;

typedef struct TRANSITION_LIST {
  int		*Destination;
  double 	Proba;
  int		Reward;
  struct TRANSITION_LIST *Next;
} *Transition_List;

typedef struct TRUNCATION_LIST {
  int		Origin;
  double	Proba;
  int		Reward;
  struct TRUNCATION_LIST *Next;
} *Truncation_List;

typedef struct FIFO_LIST {
  int		Nb_Elements;
  State_List	First;
  State_List	Last;
} *Fifo_List;

typedef struct STORE {
  int		Nb_Items;
  State_Desc	*Item;
} *Store;

typedef struct HASH_TABLE {
  int	Max_Nb_Items;
  int	Nb_Items;
  int	Nb_Boxes;
  Store *Box;
} *Hash_Table;

typedef enum { NEW, OLD } Record_Type;

typedef struct ANSWER_DESC {
  Record_Type	Status;
  State_Desc	State;
} ANSWER_DESC;

/* external declaration, from "spmat.c" of the MC package, not the
  (max,+) one! */
extern void VM_Mult(double *X, SP_Real_Matrix Matrix, double *Res );

/* forward declarations */
ANSWER_DESC Lookup_State( int *Target, int Dimension, Hash_Table Table );
void Insert_Fifo( State_Desc S, Fifo_List F );
void Explore( Tetris Problem, Markov_Chain MC, Truncation_List *TR, 
	      boolean, boolean );
void Write_Config( int *, int, FILE * );
void Compute_Markov_Bounds( Markov_Chain Chain, Truncation_List TR,
			    boolean Progress );
void Compute_Simple_Lower_Bound( Tetris Problem );
Truncation_List Add_Truncation_List( Truncation_List, int, double, int );
int Truncation_List_Length( Truncation_List );

/* global variables */
extern boolean Debug;

/***********************************************************************/
/*                Programme Principal                                  */
/***********************************************************************/

int main(int argc, char *argv[], char *argp[] )
{
  static char		In_File_Name[MAX_NAME_LENGTH] = "";
  static char		Out_File_Name[MAX_NAME_LENGTH] = "";

  int 			i;
  int			N_Max = 200;
  boolean		Progress = FALSE;
  boolean		Dump = FALSE;
  boolean		Stats = FALSE;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  Tetris		Problem;
  Markov_Chain		MC;
  Truncation_List	TR = (Truncation_List) NULL;

  /************************* Parsing *****************************/
  i = 1;

  if (argc>1)
    do {

      if GET_STRING( "-f", In_File_Name )
      else if GET_STRING( "-o", Out_File_Name )
      else if SWITCH( "-progress", Progress, TRUE )
      else if SWITCH( "-stats", Stats, TRUE )
      else if SWITCH( "-dump", Dump, TRUE )
      else if SWITCH( "-debug", Debug, TRUE )
      else if GET_INT( "-nmax", N_Max )
      else if IS_VERSION

      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {
	i++;
	fprintf(stdout,"usage: %s [-f inputfile] [-o outputfile] [-nmax int]",
		TOOLNAME );
	fprintf(stdout,"\n\t[-stats] [-dump] [-debug] [-h] [-help] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
      }
      else 
	i++;
    } while( Args_Ok && ( i < argc ) );

  if (Args_Ok) {

    if ( strlen(Out_File_Name) != 0 )
      File_Ok = (NULL != freopen( Out_File_Name, "w", stdout ) );

    if ( !File_Ok ) {
      fprintf(stderr,"%s: could not open file %s\n", 
	      TOOLNAME, Out_File_Name);
      File_Ok = FALSE;
    }
  }


  /************* start of the program if everything is Ok *******************/

  if ( Args_Ok && File_Ok ) {

    Problem = Read_Tetris( In_File_Name );

    if ( (Tetris)NULL != Problem ) {

      if ( Debug ) 
	Write_Tetris( Problem, stdout );

      /* renormalize weights. This is needed so often: better do it once. */
      Normalize_Weights( Problem );

      /* perform exploration of state space */
      if ( Progress ) {
	fprintf( stderr, "Allocating Markov Chain.\n");
	fflush( stderr );
      }
      MC = New_Markov_Chain( N_Max, DISCRETE );
      if ( Progress ) {
	fprintf( stderr, "Exploring Markov Chain.\n");
	fflush( stderr );
      }
      Explore( Problem, MC, &TR, Progress, Stats );

      if ( Dump ) {
	/* just dump the chain and exit */
	Write_Markov_Chain( stdout, MC );

	return 0;
      }
	
      if ( Debug ) {
	fprintf( stdout, "Generator:\n" );
	Write_Matrix( stdout, MC->Generator, /* full? */ FALSE );
	fprintf( stdout, "Rewards:\n" );
	Write_Matrix( stdout, MC->Reward, /* full? */ FALSE );
      }

      if ( Progress ) {
	fprintf( stderr, "Computing Bounds.\n");
	fflush( stderr );
      }

      /* Solve... */
      Compute_Simple_Lower_Bound( Problem );
      Compute_Markov_Bounds( MC, TR, Progress );

      /* computation successful */
      /* cleanup */
      Free_Tetris( Problem );
      /* Free_Markov_Chain( MC ) does not exist... */

      return 0;
    }
    else
      return -2;
  }
  else if ( Help ) {
    return 0;
  }
  else {
    return -1;
  }
}

/* Simple lower bound using the "load" of columns */
void Compute_Simple_Lower_Bound( Tetris Problem )
{
  int		col;
  int		piece;
  double	Total_Weight;
  double	Res;
  double	Rho;
  Piece_Desc	The_Piece;
  
  /* compute total weight, in case they hare not normalized */
  Total_Weight = 0.0;
  for ( piece=0; piece<Problem->Nb_Pieces; piece++ )
    Total_Weight += Problem->Weight[piece];

  /* compute max of all expected "loads" of columns */
  for ( col=0; col<Problem->Nb_Columns; col++ ) {
    Rho = 0.0;
    for ( piece=0; piece<Problem->Nb_Pieces; piece++ ) {
      The_Piece = Problem->Piece[piece];
      if ( The_Piece->LoProf[col] != -1 ) {
	Rho += Problem->Weight[piece] / Total_Weight 
	  * ( The_Piece->HiProf[col] - The_Piece->LoProf[col] );
      }
    }
    if ( Rho > Res )
      Res = Rho;
  }

  printf( "Gamma >= %8.4f\n", Res ); 
}

/************************************************************************/
/* Below are all procedures for generating the Markov chain and 	*/
/* computing its expected reward.					*/
/************************************************************************/
/* Hashing function for the states */
int Hash( int *S, int Size )
{
  int   i;
  int   Max;
  int	Res;
  /* hash value is the location of the "height" of the configuration */
  Max = S[0];
  Res = 0;
  for ( i=1; i<Size; i++ ) {
    if ( S[i] > Max ) {
      Res = i;
      Max = S[i];
    }
  }

  return Res;
}

boolean Compare( int *A, int *B, int Dimension ) 
{
  boolean Equal;
  int	i;

  Equal = TRUE;
  for ( i=0; Equal && ( i < Dimension ); i++ )
    Equal = ( A[i] == B[i] );

  return Equal;
}

int Max_Value( int *A, int Dimension ) 
{
  int	i;
  int	Res;

  Res = A[0];
  for ( i=1; i < Dimension; i++ )
    Res = ( A[i] > Res ? A[i] : Res );

  return Res;
}

int Min_Value( int *A, int Dimension ) 
{
  int	i;
  int	Res;

  Res = A[0];
  for ( i=1; i < Dimension; i++ )
    Res = ( A[i] < Res ? A[i] : Res );

  return Res;
}

Transition_List Next( int *Configuration, Tetris Problem, int *State_Buffer )
{
  int			i;
  int			Pc_Num;
  int			col;
  int			Contact;
  int			Delta;
  Transition_List	Res = (Transition_List)NULL;
  Transition_List	New;
  Piece_Desc		The_Piece;

  for ( Pc_Num=0; Pc_Num<Problem->Nb_Pieces; Pc_Num++ ) {
    The_Piece = Problem->Piece[Pc_Num];

    /* calculate new configuration: first find touching point. This is
     not trivial because I have to handle separatly the first column,
     and be careful that the piece may not span this column.	*/
    for ( col = 0; ( col < Problem->Nb_Columns ) 
	    && ( The_Piece->HiProf[col] == -1 ); col++ );
    if ( col == Problem->Nb_Columns ) {
      fprintf( stderr, "%s: error: empty piece encountered (#%d). Exit.\n",
	       TOOLNAME, i );
      exit(-4);
    }
    if ( col == 0 ) {
      Contact = -The_Piece->LoProf[0];
    } 
    else {
      Contact = Configuration[col-1] - The_Piece->LoProf[col];
    }
    for ( col++; col < Problem->Nb_Columns; col++ ) {
      if ( The_Piece->HiProf[col] != -1 ) {
	Contact = MAX( Contact, 
		       Configuration[col-1] - The_Piece->LoProf[col] );
      }
    }

    /* computing the new altitude in column 0, which is also the	*/
    /* height increment.						*/
    if ( The_Piece->HiProf[0] != -1 )
      Delta = Contact + The_Piece->HiProf[0];
    else 
      Delta = 0;

    for ( i=0; i<Problem->Nb_Columns-1; i++ ) {
      if ( The_Piece->LoProf[i+1] != -1 ) {
	State_Buffer[i] = Contact + The_Piece->HiProf[i+1] - Delta;
      }
      else {
	State_Buffer[i] = Configuration[i] - Delta;
      }
    }

    New = (Transition_List) malloc( sizeof( struct TRANSITION_LIST ) );
    New->Destination = (int*) calloc( (size_t)(Problem->Nb_Columns-1), 
				      sizeof(int) );

    for ( i=0; i<Problem->Nb_Columns-1; i++ ) {
      New->Destination[i] = State_Buffer[i];
    }
    if ( Debug ) {
      printf( "Next config: " );
      Write_Config( State_Buffer, Problem->Nb_Columns-1, stdout );
      printf("\n");
    }

    New->Proba = Problem->Weight[Pc_Num];
    New->Reward = Delta;
    /* insert new state in list */
    New->Next = Res;
    Res = New;
  }

  return Res;
}

void Add_Markov_Transition( Markov_Chain MC,
			    int Orig, int Dest, 
			    double Proba, int Reward ) 
{
  /* I use "Set" and not "Add" because "Set" allows replications and
     this is needed for the case of different rewards on the same 
     transition.							*/
  Set_Sparse_Matrix_Entry( MC->Generator, Orig-1, Dest-1, Proba );
  Set_Sparse_Matrix_Entry( MC->Reward, Orig-1, Dest-1, (double)Reward );
  if ( Debug ) {
    printf("%d -> %d wp %8.6f +%d\n", Orig, Dest, Proba, Reward );
  }
}

void Explore( Tetris Problem, Markov_Chain MC, Truncation_List *TR,
	      boolean Progress, boolean Stats )
{
  int			i;
  int			Dimension;
  int   		Max_Nb_States = MC->Generator->Size;
  int			Cur_Nb_States;
  int			*State_Buffer;
  int			Dest;
  int			MaxVal;
  int			MinVal;
  Hash_Table 		Table;
  Transition_List	Sons;
  Transition_List	S_Scan;
  ANSWER_DESC		Answer;
  State_Desc		Cur_State;
  Fifo_List		Unexplored_States;

  Dimension = Problem->Nb_Columns-1;

  /* Prepare the hash table */
  Table = (Hash_Table) malloc( sizeof( struct HASH_TABLE ) );
  Table->Max_Nb_Items = Max_Nb_States;
  Table->Nb_Items = 0;
  Table->Nb_Boxes = Dimension;
  Table->Box = (Store*) calloc( (size_t)Table->Nb_Boxes, sizeof(Store) );
  for ( i=0; i<Table->Nb_Boxes; i++ ) {
    Table->Box[i] = (Store) malloc( sizeof( struct STORE ) );
    Table->Box[i]->Nb_Items = 0;
    Table->Box[i]->Item = (State_Desc*)NULL;
  }

  State_Buffer = (int*) calloc( (size_t)Dimension, sizeof(int) );
  /* State_Buffer is initialized to (0,0,..,0) which is just fine */

  /* Create initial state and store it */
  Answer = Lookup_State( State_Buffer, Dimension, Table );
  Cur_State = Answer.State;
  Cur_Nb_States = 1;

  Unexplored_States = (Fifo_List) malloc( sizeof( struct STATE_LIST ) );
  Unexplored_States->First = (State_List)NULL;
  Unexplored_States->Last = (State_List)NULL;
  Unexplored_States->Nb_Elements = 0;

  Insert_Fifo( Cur_State, Unexplored_States );

  do {
    /* get next state and remove it from file */
    Cur_State = Unexplored_States->First->Val;
    Unexplored_States->First = Unexplored_States->First->Next;
    Unexplored_States->Nb_Elements --;

    /* find sons of this state */
    Sons = Next( Cur_State->Array, Problem, State_Buffer );

    /* sort the sons: are they new or not ? */
    SCAN( S_Scan, Sons ) {
      Answer = Lookup_State( S_Scan->Destination, Dimension, Table );
      if ( Answer.Status == NEW ) {
	if ( Cur_Nb_States < Max_Nb_States ) {
	  Insert_Fifo( Answer.State, Unexplored_States );
	  Cur_Nb_States++;
	  Dest = Answer.State->Number;
	  if ( Progress ) {
	    if ( Cur_Nb_States % 100 == 0 ) {
	      fprintf( stderr, "%d states explored (%4.2f%%). ",
		       Cur_Nb_States, 
		       100*(double)Cur_Nb_States/Max_Nb_States );
	      fprintf( stderr, "Number left to examine: %d.\n",
		       Unexplored_States->Nb_Elements );
	      fflush( stderr );
	    }	      
	  }
	}
	else {
	  /* Limit reached. Go back to zero, with modified "reward".
	     The reward is the max height of the pile if X0 was zero
	     before dropping. It is therefore the normal reward Delta(X0)
	     plus the height of the new state */
	  Dest = 1;
	  MaxVal = Max_Value( S_Scan->Destination, Dimension );
	  MinVal = Min_Value( S_Scan->Destination, Dimension );

	  S_Scan->Reward += MAX( 0, MaxVal );

	  /* Record truncation for the lower bound. Same as above for	*/
	  /* the min height of the pile. What is stored is a new 	*/
	  /* transition with a reward equal to min-max to compensate 	*/
	  /* for the normal reward.					*/
	  (*TR) = Add_Truncation_List( (*TR), 
				       Cur_State->Number,
				       S_Scan->Proba,
				       MIN( 0, MinVal ) - MAX( 0, MaxVal ) );
	}
      }
      else {
	Dest = Answer.State->Number;
      }
      Add_Markov_Transition( MC,
			     Cur_State->Number, Dest,
			     S_Scan->Proba, S_Scan->Reward );
    }
  } while ( Unexplored_States->First != (State_List)NULL );

  if ( Stats ) {
    printf("States explored: %d\n", Cur_Nb_States );
    printf("Transitions truncated: %d\n", Truncation_List_Length( (*TR) ) );
    /* Dump_Hash_Stats( Table ); */
    printf("Hash table statistics:\n");
    printf("Number of items = %d\n", Table->Nb_Items );
    printf("Number of boxes = %d\n", Table->Nb_Boxes );
    printf("Contents: \n");
    for ( i=0; i<Table->Nb_Boxes; i++ ) {
      printf("[%3d] %3d ", i, Table->Box[i]->Nb_Items );
      if ( (i+1) %7 == 0 )
	printf("\n");
    }
    printf("\n");
  }
  /* Free_Hash_Table( ) does not exist. Sorry for the waste... */

}

State_Desc New_State( int Array_Size, int *Array )
{
  int		i;
  State_Desc	Res;

  Res = malloc( sizeof( struct STATE_DESC ) );
  Res->Array = (int*)calloc( (size_t)Array_Size, sizeof(int) );
  Res->Number = -1;
  for ( i=0; i<Array_Size; i++ ) {
    Res->Array[i] = Array[i];
  }

  return Res;
}

ANSWER_DESC Lookup_State( int *Target, int Dimension, Hash_Table Table )
{
  int		i;
  int  		H;
  ANSWER_DESC	Answer;
  Store		The_Box;
  boolean	Found;

  H = Hash( Target, Dimension );
  The_Box = Table->Box[H];
  Found = FALSE;

  for( i=0; (!Found) && (i<The_Box->Nb_Items); ) {
    Found = Compare( Target, The_Box->Item[i]->Array, Dimension );
    if (!Found)
      i++;
  } 

  if ( Found ) {
    Answer.Status = OLD;
    Answer.State = The_Box->Item[i];
    if ( Debug ) {
      printf("  State exists: %d", Answer.State->Number );
      Write_Config( Answer.State->Array, Dimension, stdout );
      printf("\n");
    }
  }
  else {
    Answer.Status = NEW;
    if ( Table->Nb_Items < Table->Max_Nb_Items ) {
      /* create new state */
      Answer.State = New_State( Dimension, Target );
      Table->Nb_Items++;
      Answer.State->Number = Table->Nb_Items;
      The_Box->Nb_Items++;
      The_Box->Item = realloc( The_Box->Item, 
			       (size_t)The_Box->Nb_Items*sizeof(State_Desc) );
      The_Box->Item[The_Box->Nb_Items-1] = Answer.State;
      if ( Debug ) {
	printf("  New State: %d", Answer.State->Number );
	Write_Config( Answer.State->Array, Dimension, stdout );
	printf("\n");
      }
    }
    else {
      /* return empty state */
      Answer.State = (State_Desc)NULL;
    }
  }

  return Answer;
}

/* Insert a new state descriptor into a FIFO list. The list is assumed 	*/
/* to exist already.							*/
void Insert_Fifo( State_Desc S, Fifo_List F )
{
  State_List	New_Elt;

  New_Elt = (State_List) malloc( sizeof( struct STATE_LIST ) );
  New_Elt->Val = S;
  New_Elt->Next = (State_List)NULL;
  if ( (State_List)NULL != F->First ) {
    /* some elements are there */
    F->Last->Next = New_Elt;
    F->Last = New_Elt;
  }
  else {
    F->Last = New_Elt;
    F->First = New_Elt;
  }
  F->Nb_Elements++;
}

void Write_Config( int *Buffer, int Size, FILE *Out ) 
{
  int i;

  fprintf( Out, "( " );
  for ( i=0; i<Size; i++ )
    fprintf( Out, " %d ", Buffer[i] );
  fprintf( Out, " )");
}

double Expected_Reward( double *Pi, Markov_Chain MC ) 
{
  int		i;
  int		j;
  double	Res;

  Res = 0.0;
  for ( i=0; i<MC->Generator->Size; i++ ) {
    for ( j=0; j<MC->Generator->Nb_Elts[i]; j++ ) {
      Res += Pi[i] * MC->Generator->Vals[i][j] * MC->Reward->Vals[i][j];
    }
  }

  return Res;
}

/************************************************************************/
/* Procedure to compute the bounds by solving the Markov chain and	*/
/* computing the expected reward.					*/
/* Procedure stolen from mc_solve.c/Solve_Disc_Time_Chain and adapted	*/
/* stop on convergence of expected reward.				*/
/************************************************************************/
void Compute_Markov_Bounds( Markov_Chain Chain, 
			    Truncation_List TR,
			    boolean Progress )
{
  int		i;
  int		Size;
  int		Time;
  boolean	Continue;
  double	*Pi;
  double	*Res;
  double	*Tmp;
  double	err;
  double	Cur_Reward;
  double	New_Reward;
  int		T_Max = 1000;
  Truncation_List	L_Scan;

  Size = Chain->State_Space_Size;

  Pi = (double*) calloc( (size_t)Size, sizeof( double ) );
  Res = (double*) calloc( (size_t)Size, sizeof( double ) );

  /* initialize Pi */
  Pi[0] = 1.0;
  Cur_Reward = 0.0;

  Continue = TRUE;
  for ( Time=0; (Time<T_Max) && Continue; Time++ ) {
    VM_Mult( Pi, Chain->Generator, Res );
    
    if ( Debug ) {
      /* dump the current estimate of Pi */
      printf("# Pi(%d) = ", Time );
      for ( i=0; i<Size; i++ ) {
	printf("%8.6f ", Res[i] );
      }
      printf("\n");
    }

    New_Reward = Expected_Reward( Pi, Chain );
    err = fabs( New_Reward - Cur_Reward );
    if ( Progress ) {
      fprintf( stderr, "Iteration #%3d: gamma = %8.4f   distance %10e\n", 
	       Time, New_Reward, err );
      fflush( stderr );
    }
    if ( ( Time >= 10 ) && ( err < 1e-7 ) ) {
      Continue = FALSE;
    } 
    else {
      Cur_Reward = New_Reward;
      Tmp = Res; Res = Pi; Pi = Tmp;
    }
  }
  if ( Time >= T_Max ) {
    fprintf( stderr, "Warning: Maximum number of iterations found.");
    fprintf( stderr, "Distribution may not be correct.\n");
  }

  /* write result out */
  /*
  printf("# Stationary distribution\n");
  for ( i=0; i<Size; i++ ) {
    printf("%3d %8.6f\n", i, Pi[i] );
  }
  */
  printf( "Gamma <= %8.4f\n", New_Reward ); 

  /* Now recompute average reward with rewards of the truncation list. */
  SCAN( L_Scan, TR ) {
    Add_Markov_Transition( Chain,
			   L_Scan->Origin, 0, 
			   L_Scan->Proba, (double)L_Scan->Reward );
  }
  
  New_Reward = Expected_Reward( Pi, Chain );

  printf( "Gamma >= %8.4f\n", New_Reward ); 

  /* cleanup */
  free( Pi );
  free( Res );

}


/* Procedures related to Truncation Lists */
Truncation_List New_Truncation_List( int	Origin,
				     double	Proba,
				     int	Reward )
{
  Truncation_List	Res;

  Res = (Truncation_List) malloc( sizeof( struct TRUNCATION_LIST ) );
  Res->Origin = Origin;
  Res->Proba = Proba;
  Res->Reward = Reward;
  Res->Next = (Truncation_List) NULL;

  return Res;
  
}

Truncation_List Add_Truncation_List( Truncation_List TR,
				     int	Origin,
				     double	Proba,
				     int	Reward )
{
  Truncation_List	Elem;

  Elem = New_Truncation_List( Origin, Proba, Reward );
  Elem->Next = TR;

  return Elem;
  
}

int Truncation_List_Length( Truncation_List TR )
{
  int	Res;
  Truncation_List	L_Scan;

  Res = 0;

  SCAN( L_Scan, TR ) {
    Res++;
  }

  return Res;
}

