/* Various Sparse matrix manipulation procedures */

#include <stdlib.h>
#include <stdio.h>
#include "boolean.h"
#include "spmat.h"

/* The constructor */
SP_Real_Matrix New_SP_Real_Matrix(int	Size)
{
  int			i;
  SP_Real_Matrix	Res;

  Res = (SP_Real_Matrix) malloc( sizeof( struct SP_REAL_MATRIX ) );

  if ( (SP_Real_Matrix)NULL != Res ) {
    Res->Size = Size;

    Res->Nb_Elts = (int*) calloc( (size_t)Size, sizeof(int) );
    Res->Elts = (int**) calloc( (size_t)Size, sizeof(int*) );
    Res->Vals = (double**) calloc( (size_t)Size, sizeof(double*) );

    if ( ( (int*)NULL != Res->Nb_Elts )
	 && ( (int**)NULL != Res->Elts )
	 && ( (double**)NULL != Res->Vals ) ) {
      for ( i=0; i<Size; i++ ) {
	Res->Nb_Elts[i] = 0;
	Res->Elts[i] = (int*)NULL;
	Res->Vals[i] = (double*)NULL;
      }
    } 
    else {
      /* abort... without salvaging the memory (shame on me)... */
      free( Res );
      Res = (SP_Real_Matrix)NULL;
    }
  }

  return Res;
}

/* Inserting an element in the matrix.					*/
/* IMPORTANT NOTE: there is no check for the existence of the column	*/
/* This is because in the context of Markov Reward Processes, there 	*/
/* may be several transitions with different reward. It is simpler to	*/
/* handle this by allowing replicated entries in the neighborhood lists.*/
/* The multiplication algorithm handles this in a natural way.		*/
boolean Set_Sparse_Matrix_Entry(SP_Real_Matrix	Matrix, 
				int		Row,
				int		Col,
				double		Val)
{
  int		i;
  int		j;

  /* could be checking ranges ... oh well. */
  if ( ( 0 <= Row ) && ( Row < Matrix->Size ) ) {
    Matrix->Nb_Elts[Row]++;
    Matrix->Elts[Row] = 
      (int*)realloc( Matrix->Elts[Row],
		     (size_t)Matrix->Nb_Elts[Row]*sizeof(int) );
    Matrix->Vals[Row] = 
      (double*)realloc( Matrix->Vals[Row],
			(size_t)Matrix->Nb_Elts[Row]*sizeof(double) );
      
    if ( ( (int*)NULL == Matrix->Elts[Row] )
	 || ( (double*)NULL == Matrix->Vals[Row] ) ) {
      return FALSE;
    }
    else {
      Matrix->Elts[Row][Matrix->Nb_Elts[Row]-1] = Col;
      Matrix->Vals[Row][Matrix->Nb_Elts[Row]-1] = Val;
      return TRUE;
    }
  }
  else {
    printf("Warning: setting value to (%d,%d) in matrix of size %d.",
	   Row, Col, Matrix->Size );
    printf(" Ignored.\n");
    return FALSE;
  }
}

/* Adding a value to an element to the matrix */
boolean Add_To_Sparse_Matrix_Entry(SP_Real_Matrix	Matrix, 
				   int			Row,
				   int			Col,
				   double		Val)
{
  int		i;
  int		j;

  /* could be checking ranges ... oh well. */
  if ( ( 0 <= Row ) && ( Row < Matrix->Size ) ) {
    /* checking if the element is there already */
    for ( i=0; 
          ( i<Matrix->Nb_Elts[Row] ) && ( Matrix->Elts[Row][i]!= Col ); 
          i++ );
    if ( i<Matrix->Nb_Elts[Row] ) {
      /* found it. */
      Matrix->Vals[Row][i] += Val;
      return TRUE;
    }
    else { 
      /* not there. add element. I replicate the code of "Set_Sparse..." */
      /* for the sake of efficiency.					*/
      Matrix->Nb_Elts[Row]++;
      Matrix->Elts[Row] = 
	(int*)realloc( Matrix->Elts[Row],
		       (size_t)Matrix->Nb_Elts[Row]*sizeof(int) );
      Matrix->Vals[Row] = 
	(double*)realloc( Matrix->Vals[Row],
			  (size_t)Matrix->Nb_Elts[Row]*sizeof(double) );
      
      if ( ( (int*)NULL == Matrix->Elts[Row] )
	   || ( (double*)NULL == Matrix->Vals[Row] ) ) {
	return FALSE;
      }
      else {
	Matrix->Elts[Row][Matrix->Nb_Elts[Row]-1] = Col;
	Matrix->Vals[Row][Matrix->Nb_Elts[Row]-1] = Val;
	return TRUE;
      }
    }
  }
  else {
    printf("Warning: setting value to (%d,%d) in matrix of size %d.",
	   Row, Col, Matrix->Size );
    printf(" Ignored.\n");
    return FALSE;
  }
}

/* Reading from a file, and adding an element to the matrix */
boolean Read_Full_Matrix_Entry(FILE		*Input, 
			       SP_Real_Matrix	Matrix, 
			       int		Row,
			       int		Col)
{
  double	Val;

  if ( 1 == fscanf( Input, "%lf", &Val ) ) {
    if ( Set_Sparse_Matrix_Entry( Matrix, Row, Col, Val ) )
      return TRUE;
    else {
      printf("Warning: failed to set entry (%d,%d) to %f.\n",
	     Row, Col, Val );
      return FALSE;
    }
  }
  else {
    return FALSE;
  }
}

/* Reading from a file, and adding an element to the matrix */
boolean Read_Sparse_Matrix_Entry(FILE		*Input, 
				 SP_Real_Matrix	Matrix)
{
  double	Val;
  int		Row;
  int		Col;

  if ( 3 == fscanf( Input, "%d %d %lf", &Row, &Col, &Val ) ) {
    Set_Sparse_Matrix_Entry( Matrix, Row, Col, Val );
    return TRUE;
  }
  else {
    return FALSE;
  }
}

/* Retrieving an entry from the matrix. This is not optimized at all... */
double Get_Sparse_Matrix_Entry(SP_Real_Matrix	Matrix, 
			       int		Row,
			       int		Col)
{
  int		i;
  double	Res;

  Res = 0.0;
  for( i=0; i<Matrix->Nb_Elts[Row]; i++ ) {
    if ( Matrix->Elts[Row][i] == Col ) {
      Res = Matrix->Vals[Row][i];
    }
  }

  return Res;
}

/* Linear algebra... */
void VM_Mult(double 		*X,
	     SP_Real_Matrix	Matrix, 
	     double		*Res )
{
  int		i;
  int		j;
  int		k;

  /* (re) initialization */
  for ( i=0; i<Matrix->Size; i++ ) 
    Res[i] = 0.0;

  for ( i=0; i<Matrix->Size; i++ ) {
    for ( j=0; j<Matrix->Nb_Elts[i]; j++ ) {
      k = Matrix->Elts[i][j];
      Res[k] += X[i] * Matrix->Vals[i][j];
    }
  }
}

/* Summing a row */
double Row_Sum(SP_Real_Matrix	Matrix, 
	       int		Row)
{
  static int i;
  static double	Sum;

  Sum = 0.0;
  for( i=0; i<Matrix->Nb_Elts[Row]; i++) {
    Sum += Matrix->Vals[Row][i];
  }

  return Sum;
}

/* Writing it out... */
void Write_Matrix(FILE 			*Output,
		  SP_Real_Matrix	Matrix,
		  boolean		Full_Form)
{
  int		i;
  int		j;
  double	Val;
  
  for ( i=0; i<Matrix->Size; i++ ) {
    printf("[%2d]", Matrix->Nb_Elts[i] );
    if ( Full_Form ) {
      for ( j=0; j<Matrix->Size; j++ ) {
	Val = Get_Sparse_Matrix_Entry( Matrix, i, j );
	printf(" %8.4f", Val );
      }
    }
    else {
      for ( j=0; j<Matrix->Nb_Elts[i]; j++ ) {
	printf(" %d:%8.4f", Matrix->Elts[i][j], Matrix->Vals[i][j] );
      }
    }
    printf("\n");
  }
}

