#include <stddef.h>
#include <malloc.h>
#include <math.h>

#include "pn.h"

#define INFINI		1e12

#define max(x,y)        ( (x>y) ? x : y )
#define SCAN(x,y)	for(x=y;x!=NULL;x=x->Next)
#define IRINT(x)	(int)rint(x)

/* additionnal declaration, not in the standard "math.h" */
extern double rint(double x);

typedef struct LAYER_DESC {
  int	Nb_Nodes;
  int	*Node_Id;
} *Layer_Desc;

typedef struct LAYER_STRUCT {
  int		Nb_Levels;
  Layer_Desc	*Level;		/* array of layer descriptions */
} *Layer_Struct;

/*The value of the precision for integrality check of reals */
double	Epsilon = 1e-05;
/************************************************************************/
/* Functions for checking that the network is OK			*/
/************************************************************************/
boolean Is_Integer( double x )
{
  double	Delta;
  Delta = fabs( x - rint(x) );

  return ( Delta < Epsilon );
}

int gcd( int n, int m )
{
  if ( n == m )
    return n;
  else if ( n == 0 )
    return m;
  else if ( m == 0 )
    return n;
  else if ( n > m )
    return gcd( m, n % m );
  else
    return gcd( n, m % n );
}

boolean	Has_All_Deterministic_Transitions(
PetriNet 	*Reseau,
boolean		Testing,
int		*Bad_Objects )
{
  int		i;
  boolean	Is_Ok = TRUE;
  boolean	Is_Cst = TRUE;

  for ( i=0; ( Testing || Is_Ok ) && ( i < Reseau->Nb_Trans ); i++ ) 
    if ( Reseau->T[i].Tag == 0 ) {
      Is_Cst = ( Reseau->T[i].Tr_Law->Name == 'C' );

      if ( !Is_Cst ) {
	Is_Ok = FALSE;
	printf("Trans. %d (%s) has non constant delay.\n", 
	       i, Reseau->T[i].Name );
	if ( Testing ) {
	  /* Store the marker for transitions: 0, and the transition */
	  Bad_Objects[0]++;
	  Bad_Objects[2*Bad_Objects[0]-1] = 0;
	  Bad_Objects[2*Bad_Objects[0]] = i;
	}
      }
    }

  return Is_Ok;
}

boolean	Has_All_Deterministic_Places(
PetriNet 	*Reseau,
boolean		Testing,
int		*Bad_Objects )
{
  int		i;
  boolean	Is_Ok = TRUE;
  boolean	Is_Cst = TRUE;

  for ( i=0; ( Testing || Is_Ok ) && ( i < Reseau->Nb_Places ); i++ ) 
    if ( Reseau->P[i].Tag == 0 ) {
      Is_Cst = ( Reseau->P[i].Pl_Law->Name == 'C' );

      if ( !Is_Cst ) {
	Is_Ok = FALSE;
	printf("Place %d (%s) has non constant delay.\n", 
	       i, Reseau->P[i].Name );
	if ( Testing ) {
	  /* Store the marker for places: 1, and the place */
	  Bad_Objects[0]++;
	  Bad_Objects[2*Bad_Objects[0]-1] = 1;
	  Bad_Objects[2*Bad_Objects[0]] = i;
	}
      }
    }

  return Is_Ok;
}

boolean	Has_All_Deterministic_Objects(
PetriNet 	*Reseau,
boolean		Testing,
int		*Bad_Objects )
{
  boolean	Is_Ok = TRUE;

  /* checking transitions */
  Is_Ok = Has_All_Deterministic_Transitions( Reseau, Testing, Bad_Objects );

  if ( Is_Ok || Testing ) {
    /* checking places... */
    Is_Ok = Has_All_Deterministic_Places( Reseau, Testing, Bad_Objects );
  }

  return Is_Ok;
}
/************************************************************************/
/* Function for finding the time scale of a network which is a number	*/
/* dividing all times (firing, holding, initial lags).			*/
/* The argument "Scale" specifies that the time scale is searched as	*/
/* a multiple of 1/Scale. The function returns 1/the scale found.	*/
/* If "Testing" is true, objects not complying are stored in the array	*/
/* "Bad_Objects". 							*/
/* Note: this function assumes that all times are deterministic.	*/
/* If the switch "Sources_Also" is FALSE, then sources that are not	*/
/* integral are left out of the computation. It is assumed that they 	*/
/* will be later discretized according to the time scale of the other.	*/
/************************************************************************/
int	Find_Time_Scale(
PetriNet 	*Reseau,
boolean		Sources_Also,
int		Scale,
boolean		Testing,
int		*Bad_Objects )
{
  int		i, j;
  boolean	Is_Ok = TRUE;
  boolean	Is_Int;
  boolean	Is_Source;
  int		Pl_Num;
  double	Val;
  int		New_Scale;
  int		Trans_Scale;
  int		Nb_Params;
  Law_Desc	The_Law;

  New_Scale = Scale;

  /* check integrality of delays of transitions */
  for ( i=0; ( Testing || Is_Ok ) && ( i < Reseau->Nb_Trans ); i++ ) 
    if ( Reseau->T[i].Tag == 0 ) {

      /* decide whether it is a source or not. Might be useful later. */
      Is_Source = FALSE;
      if ( Reseau->T[i].Indegree == 0 )
	/* I assume this is meant to be a source... */
	Is_Source = TRUE;
      else if ( Reseau->T[i].Indegree == 1 ) {
	Pl_Num = Reseau->T[i].Inputs->Dest;
	if ( ( Reseau->P[Pl_Num].Indegree == 1 ) 
	     && ( Reseau->P[Pl_Num].Inputs->Dest == i ) )
	  Is_Source = TRUE;
      }

      /* analyze firing law. */
      The_Law = Reseau->T[i].Tr_Law;
      switch( The_Law->Name ) {
      case 'C':
	Val = Scale * The_Law->Parameters[0];
	Is_Int = Is_Integer( Val );
	Trans_Scale = IRINT( Val );
	break;

      case 'D':
	Nb_Params = IRINT( The_Law->Parameters[0] );
	Trans_Scale = Scale;
	Is_Int = TRUE;
	for ( j=1; j<= Nb_Params; j++ ) {
	  Val = Scale * The_Law->Parameters[j];
	  Is_Int = Is_Int && Is_Integer( Val );
	  if ( Is_Int ) 
	    Trans_Scale = gcd( Trans_Scale, IRINT( Val ) );
	}
	break;

      case 'G':
      case 'B':
	Is_Int = TRUE;
	Trans_Scale = Scale;
	break;

      default:
	Is_Int = FALSE;
      }

      if ( !Is_Int ) {
	/* if Source and not Sources Also, ignore non-integrality */
	if ( Is_Source && !Sources_Also ) {
	  if ( Debug ) {
	    printf("Trans. %d (%s) has non integer delay... but that's ok.\n", 
		   i, Reseau->T[i].Name );
	  }
	}
	else {
	  Is_Ok = FALSE;
	  printf("Trans. %d (%s) has non integer delay.\n", 
		 i, Reseau->T[i].Name );
	  if ( Testing ) {
	    /* Store the marker for transitions: 0, and the transition */
	    Bad_Objects[0]++;
	    Bad_Objects[2*Bad_Objects[0]-1] = 0;
	    Bad_Objects[2*Bad_Objects[0]] = i;
	  }
	}
      }
      else {
	New_Scale = gcd( New_Scale, Trans_Scale );
      }
    }

  if ( Is_Ok || Testing ) {
    for ( i=0; ( Testing || Is_Ok ) && ( i < Reseau->Nb_Places ); i++ ) 
      if ( Reseau->P[i].Tag == 0 ) {
	Val = Scale * Reseau->P[i].Pl_Law->Parameters[0];
	Is_Int = Is_Integer( Val );
	
	if ( !Is_Int ) {
	  Is_Ok = FALSE;
	  printf("Place %d (%s) has non integer delay.\n",
		 i, Reseau->P[i].Name );
	  if ( Testing ) {
	    /* Store the marker for places: 1, and the place */
	    Bad_Objects[0]++;
	    Bad_Objects[2*Bad_Objects[0]-1] = 1;
	    Bad_Objects[2*Bad_Objects[0]] = i;
	  }
	}
	else {
	  New_Scale = gcd( New_Scale, IRINT(Val) );
	}
      }
  }

  if ( Is_Ok ) {
    for ( i=0; Is_Ok && ( i < Reseau->Nb_Places ); i++ )
      if ( Reseau->P[i].Tag == 0 )
	for ( j=0; Is_Ok && ( j < Reseau->P[i].Init_Pop ); j++ ) {
	  Val = Scale * Reseau->P[i].Lag_Times[j];
	  Is_Int = Is_Integer( Val );
    
	  if ( !Is_Int ) {
	    Is_Ok = FALSE;
	    printf("Place %d (%s) has non integer lag times.\n", 
		   i, Reseau->P[i].Name );
	  }
	  else {
	    New_Scale = gcd( New_Scale, IRINT(Val) );
	  }
	}

  }

  if ( Is_Ok )
    return Scale / New_Scale;
  else
    return 0;
}

boolean Has_All_Routing_Places( Reseau, Testing, Bad_Objects )
PetriNet 	*Reseau;
boolean		Testing;
int		*Bad_Objects;
{
  int		i;
  boolean	Is_Ok = TRUE;
  boolean	Is_Routing = TRUE;

  for ( i=0; ( Testing || Is_Ok ) && ( i < Reseau->Nb_Places ); i++ )
    if ( Reseau->P[i].Tag == 0 ) {
      if ( Reseau->P[i].Routing != NULL ) {
	Is_Routing = 
	  ( ( Reseau->P[i].Routing->Type == 'R' )
	    || ( Reseau->P[i].Routing->Type == 'D' )
	    || ( Reseau->P[i].Outdegree <= 1 ) );

	if ( !Is_Routing ) {
	  Is_Ok = FALSE;
	  printf("Place %d (%s) has competition routing.\n", 
		 i, Reseau->P[i].Name );
	  if ( Testing ) {
	    /* Store the marker for places: 1, and the place */
	    Bad_Objects[0]++;
	    Bad_Objects[2*Bad_Objects[0]-1] = 1;
	    Bad_Objects[2*Bad_Objects[0]] = i;
	  }
	}
      }
    }

  return Is_Ok;
}

boolean Has_Simulable_Places( Reseau, Testing, Bad_Objects )
PetriNet 	*Reseau;
boolean		Testing;
int		*Bad_Objects;
{
  int		i;
  boolean	Is_Ok = TRUE;
  boolean	Is_Routing = TRUE;
  Liste_Link	L_Scan;
  Law_Desc	The_Law;

  for ( i=0; ( Testing || Is_Ok ) && ( i < Reseau->Nb_Places ); i++ )
    if ( Reseau->P[i].Tag == 0 ) {
      if ( Reseau->P[i].Routing != NULL ) {
	Is_Routing = 
	  ( ( Reseau->P[i].Routing->Type == 'R' )
	    || ( Reseau->P[i].Routing->Type == 'D' )
	    || ( Reseau->P[i].Routing->Type == 'P' )
	    || ( Reseau->P[i].Outdegree <= 1 ) );

	if ( Reseau->P[i].Routing->Type == 'P' ) {
	  /* check if all output transitions are instantaneous */
	  SCAN( L_Scan, Reseau->P[i].Outputs ) {
	    The_Law = Reseau->T[L_Scan->Dest].Tr_Law;
	    if ( ( The_Law->Name != 'C' ) 
		 || ( The_Law->Parameters[0] != 0.0 ) )
	      Is_Routing = FALSE;
	  }
	}

	if ( !Is_Routing ) {
	  Is_Ok = FALSE;

	  if ( !Testing || ( Bad_Objects == NULL ) ) {
	    printf("Place %d (%s) has competition routing.\n", 
		   i, Reseau->P[i].Name );
	  }
	  else if ( Testing ) {
	    /* Store the marker for places: 1, and the place */
	    Bad_Objects[0]++;
	    Bad_Objects[2*Bad_Objects[0]-1] = 1;
	    Bad_Objects[2*Bad_Objects[0]] = i;
	  }
	}
      }
    }

  return Is_Ok;
}
/************************************************************************/
/* A transition is said to be "integral" if:				*/
/* - either it is a source with a discrete distribution (cst, geom,...)	*/
/* - either it is a normal transition with a deterministic time 	*/
/* The sources are allowed to have any distribution if Source_Also	*/
/* is FALSE.								*/
/************************************************************************/
boolean Has_Integral_Transitions(
PetriNet 	*Reseau,
boolean		Testing,
boolean		Sources_Also,
int		*Bad_Objects )
{
  int		i;
  int		Pl_Num;
  boolean	Is_Ok = TRUE;
  boolean	Is_Source = TRUE;
  Law_Desc	The_Law;
  boolean	Is_Cst = TRUE;

  /* checking transitions */
  for ( i=0; ( Testing || Is_Ok ) && ( i < Reseau->Nb_Trans ); i++ ) 
    if ( Reseau->T[i].Tag == 0 ) {

      Is_Source = FALSE;
      if ( Reseau->T[i].Indegree == 0 )
	/* I assume this is meant to be a source... */
	Is_Source = TRUE;
      else if ( Reseau->T[i].Indegree == 1 ) {
	Pl_Num = Reseau->T[i].Inputs->Dest;
	if ( ( Reseau->P[Pl_Num].Indegree == 1 ) 
	     && ( Reseau->P[Pl_Num].Inputs->Dest == i ) )
	  Is_Source = TRUE;
      }

      The_Law = Reseau->T[i].Tr_Law;

      if ( Is_Source && Sources_Also ) {
	/* there remains to check the discrete nature of delays */
	Is_Cst = ( ( The_Law->Name == 'C' )
		   || ( The_Law->Name == 'G' )
		   || ( The_Law->Name == 'B' )
		   || ( The_Law->Name == 'D' )
		   );
      }
      else if ( !Is_Source )
	Is_Cst = ( The_Law->Name == 'C' );
      else /* a source, but Discretized... */
	Is_Cst = TRUE;

      if ( !Is_Cst ) {
	Is_Ok = FALSE;
	if ( !Testing ) 
	  printf("Trans. %d (%s) has non constant/discrete delay.\n", 
		 i, Reseau->T[i].Name );
	else {
	  /* Store the marker for transitions: 0, and the transition */
	  Bad_Objects[0]++;
	  Bad_Objects[2*Bad_Objects[0]-1] = 0;
	  Bad_Objects[2*Bad_Objects[0]] = i;
	}
      }
    }

  return Is_Ok;
}

boolean Has_Real_Sources(PetriNet 	*Reseau,
			 boolean	Testing,
			 int		*Bad_Objects )
{
  int		i;
  boolean	Is_Real;
  boolean	Is_Ok = TRUE;

  for( i=0; ( Testing || Is_Ok ) && ( i < Reseau->Nb_Trans ); i++ ) 
    if ( Reseau->T[i].Tag == 0 ) {
      Is_Real = ( Reseau->T[i].Indegree > 0 );

      if ( !Is_Real ) {
	Is_Ok = FALSE;
      
	printf("Transition %d (%s) has no input.\n", 
	       i, Reseau->T[i].Name );
	if ( Testing ) {
	  /* Store the marker for transitions: 0, and the transition */
	  Bad_Objects[0]++;
	  Bad_Objects[2*Bad_Objects[0]-1] = 0;
	  Bad_Objects[2*Bad_Objects[0]] = i;
	}
      }
    }

  return Is_Ok;
}

double	Max_Lag( Reseau )
PetriNet	*Reseau;
{
  int		i;
  int		j;
  double	Res;

  Res = -INFINI;

  for ( i=0; i<Reseau->Nb_Places; i++ )
    if ( Reseau->P[i].Tag == 0 ) {
      for( j=0; j<Reseau->P[i].Init_Pop; j++ )
	Res = max( Res, Reseau->P[i].Lag_Times[j] );
    }

  return Res;
}  

double	Max_Delay( Reseau )
PetriNet	*Reseau;
{
  int		i;
  double	Res;
  double	Delay1, Delay2;
  Liste_Link	L_Scan;

  Res = -INFINI;

  for ( i=0; i<Reseau->Nb_Places; i++ ) 
    if ( Reseau->P[i].Tag == 0 ) {
      Delay1 = Reseau->P[i].Pl_Law->Parameters[0];
      SCAN( L_Scan, Reseau->P[i].Inputs ) {
	Delay2 = Reseau->T[L_Scan->Dest].Tr_Law->Parameters[0];
	Res = max( Res, Delay1 + Delay2 );
      }
    }

  return Res;
}  

int	Nb_Available_Tokens( The_Place, Time )
Pl_Des		The_Place;
int		Time;
{
  int		j;
  int		Res;

  Res = 0;

  for( j=0; j<The_Place.Init_Pop; j++ )
      Res += ( The_Place.Lag_Times[j] <= Time ? 1 : 0 );

  return Res;
}  

/************************************************************************/
/* Procedures to perform the "topological analysis" of the structure 	*/
/* of delays.								*/
/************************************************************************/
/************************************************************************/
/* Procedure to extract a layer structure from an adjacency matrix Adj, */
/* of dimension The_Size, and store it in the structure "The_Layers".	*/ 
/* The space needed for the layer structure must have been allocated	*/
/* Returns the number of layers found, 0 if a cycle exists.		*/
/*									*/
/* Copied from the module "analyze_eg.c".				*/
/************************************************************************/
int find_layer_structure( Adj, The_Size, Level )
boolean		*Adj;
int		The_Size;
Layer_Desc	*Level;
{
  int		Layer_Nb;
  int		Left_Nodes, InDegree;
  boolean	Found_One;
  boolean	*Present;	/* matrice de booleens */
  Layer_Desc	Niveau;
  int		i,j;

  Present = (boolean*) calloc( (unsigned)The_Size, sizeof( boolean ) );

  for (i=0; i<The_Size; i++) {
    Present[i] = TRUE;
    Level[i] = NULL;
  }
	
  Left_Nodes = The_Size;
  Found_One = TRUE;
  for ( Layer_Nb = 0; (Left_Nodes > 0) && Found_One ; Layer_Nb++ ) {
    Level[Layer_Nb] = (Layer_Desc) malloc( sizeof( struct LAYER_DESC ) );
    Niveau = Level[Layer_Nb];
    Niveau->Nb_Nodes = 0;
    Niveau->Node_Id = (int*) calloc( (unsigned)Left_Nodes, sizeof (int) );

    /* look for nodes without fathers */
    Found_One = FALSE;
    for (j=0; j<The_Size; j++)
      if ( Present[j] ) {
	InDegree = 0;
	for (i=0; i<The_Size; i++)
	  if ( Present[i] && Adj[i*The_Size+j] )
	    InDegree++;
	if (InDegree == 0) {
	  /* node j has no fathers... */
	    Niveau->Node_Id[Niveau->Nb_Nodes] = j;
	    Niveau->Nb_Nodes++;
	    Found_One = TRUE;
	}
      }

    if ( Found_One ) {
      for (i=0; i<Niveau->Nb_Nodes; i++)
	Present[Niveau->Node_Id[i]] = FALSE;
      Left_Nodes -= Niveau->Nb_Nodes;
    }
  }

  free( Present );

  if ( Found_One )
    return Layer_Nb;
  else
    return 0;
}

/************************************************************************/
/* Procedure to find a cycle in graph A, knowing A is strongly		*/
/* connected, possibly ignoring "self loops" (according to the option	*/
/* "Skip_Recycles".							*/
/* Returns a pointer to the cycle as an array of integers, the		*/
/*   first one being the size of the cycle				*/
/*									*/
/* Copied from the module "analyze_eg.c".				*/
/************************************************************************/
int *find_cycle(A,The_Size,Skip_Recycles)
boolean	*A;
int	The_Size;
boolean Skip_Recycles;
{
  int 		i,j,k;
  boolean	*Visited;	/* matrice de booleens	*/
  boolean	*Present;
  int		*Outdeg;	/* matrice des degres	*/
  int		Start;
  int		*Cycle;
  int		*Chemin;
  int		Lg;
  int		Dep;

  Visited = (boolean*) calloc( (unsigned)The_Size, sizeof( boolean ) );
  Present = (boolean*) calloc( (unsigned)The_Size, sizeof( boolean ) );
  Outdeg = (int*) calloc( (unsigned)The_Size, sizeof( int ) );
  Chemin = (int*) calloc( (unsigned)The_Size, sizeof( int ) );

  /* initialization. Mark nodes not visited, compute outdegree	*/
  /* and mark present if outdegree <> 0				*/
  for (i=0; i<The_Size; i++) {
    Visited[i] = FALSE;
    Outdeg[i] = 0;
    for (j=0; j<The_Size; j++)
      if ( A[i*The_Size+j]
	   && ( ( i <= j ) || Present[j] )
	   && ( ( i != j ) || !Skip_Recycles ) )
	Outdeg[i]++;
    Present[i] = ( Outdeg[i] != 0 );
  }

  if ( Debug )
    printf("Looking for zero weight cycle\n");

  /* first pass: follow a path until visit to a previous node 	*/
  /* first find initial node with outdegree <> 0 		*/
  Dep = 0;
  Lg = 0;
  while ( Lg == 0 ) {	
    /* repeat until cycle was found */
    /* Find next node with outdegree <> 0 and let it be the departure	*/
    /* of the potential cycle.						*/
    for ( ; 
	  ( Dep < The_Size ) && Present[Dep] && ( Outdeg[Dep] == 0 );
	  Dep++ );
    if ( Dep == The_Size ) {
      printf("Error in analysis: no cycle found ?!?\n");
      Lg = -1;	/* to exit the while loop */
    }
    else if ( Debug )
      printf("... Dep=%d ", Dep );

    i = Dep;

    while ( ! Visited[i] ) {
      Visited[i] = TRUE;
      Chemin[Lg] = i;
      Lg ++;
      /* find a neighbor which is present and not a leaf */
      for ( j = 0;
	    ( j < The_Size )
	      && ( !Present[j] || ( Outdeg[j] == 0 ) 
		   || (Skip_Recycles && (j==i)) 
		   || !A[i*The_Size+j] );
	    j++ );
      if ( j == The_Size ) {	
	/* dead end: we have to backtrack. Remove i which is bad */
	Present[i] = FALSE;
	Lg --;
	/* find last "good" node, but declare it as unvisited */
	if ( Lg > 0 ) {
	  Lg --;
	  i = Chemin[Lg];
	  Visited[i] = FALSE;
	}
	else 
	  i = Dep;
	if ( Debug )
	  printf("x Back to %d ", i );
      }
      else {
	i = j;
	if ( Debug ) 
	  printf("-> %d ", i );
      }
    }
    /* here, we have found an already visited node */
    if ( i == Dep ) {
      /* Either we have found a cycle starting at Dep, in which case	*/
      /* Lg > 0 and we are done, or we found nothing and had to 	*/
      /* backtrack to  Dep. If so, we go on (top of the loop) with next	*/
      /* possible departure node = Dep+1.				*/
      Present[Dep] = FALSE;
      Dep++;
    }
  }
  /* A cycle has been found. Mark the end for the next phase 		*/
  Chemin[Lg] = i;

  /* second pass: store the cycle					*/
  Start = i;
  Cycle	= (int*) calloc( (unsigned)Lg+1, sizeof( int ) ); 
  /* find the start of the cycle in the path	*/
  for ( i = 0; Chemin[i] != Start; i++);
  Cycle[1] = Start;
  k = i;
  for ( i++ ; Chemin[i] != Start; i++ ) {
    Cycle[i-k+1] = Chemin[i];
  }
  Cycle[0] = Lg-k;

  free( (char*)Visited );
  free( (char*)Present );
  free( (char*)Chemin );
  free( (char*)Outdeg );

  return Cycle;
}

void print_trans_cycle( Reseau, Cycle )
PetriNet	*Reseau;
int 		*Cycle;
{
  int i;
  
  printf("%1d ", Cycle[1]+1 );
  if ( strlen( Reseau->T[Cycle[1]].Name ) != 0 )
    printf("(%s) ", Reseau->T[Cycle[1]].Name );

  for ( i=2 ; i <= Cycle[0]; i++ ) {
    printf("-> %1d ",Cycle[i]+1);
    if ( strlen( Reseau->T[Cycle[i]].Name ) != 0 )
      printf("(%s) ", Reseau->T[Cycle[i]].Name );
  }

  printf("-> %1d\n", Cycle[1]+1 );
}

void Dump_Layer_Struct( The_Layer_Struct )
Layer_Struct	The_Layer_Struct;
{
  int i, j;
  
  printf("Layer structure:\n");
  for ( i=0 ; i < The_Layer_Struct->Nb_Levels; i++ ) {
    printf("Layer %2d: ", i );
    for ( j=0; j<The_Layer_Struct->Level[i]->Nb_Nodes; j++ )
      printf("%1d ", The_Layer_Struct->Level[i]->Node_Id[j] );
    printf("\n");
  }

}
/************************************************************************/
/* Procedure to construct the analysis in layers of the "delay" 	*/
/* structure of a PN. 							*/
/* Upon success, returns TRUE and fills in "The_Layers".		*/
/* Upon failure (0 time cycle), returns false and dumps a cycle.	*/
/************************************************************************/
boolean topological_analysis( PetriNet		*Reseau,
			      Layer_Struct	The_Layer_Struct )
{
  boolean	*Adj0;		/* matrice d'adjacence	*/
  int		*Cycle;
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;

  int		The_Size;
  int		Nb_Layers;
  int		i, j, k, l;

  The_Size = Reseau->Nb_Trans;

  /* Create the adjacency matrix (BOO, HISS) of zero delay transitions */
  Adj0 = (boolean*) 
    calloc ( (unsigned) The_Size * The_Size, sizeof( boolean ) );

  for ( i=0; i<The_Size*The_Size; i++ )
    Adj0[i] = FALSE;

  /* Fill the adjacency matrix of zero delay transitions */
  for (i=0; i<The_Size; i++) 
    if ( Reseau->T[i].Tag == 0 ) {
      SCAN( L_Scan, Reseau->T[i].Inputs ) {
	j = L_Scan->Dest;

	if ( Reseau->P[j].Tag == 0 ) {
	  SCAN( M_Scan, Reseau->P[j].Inputs ) {
	    k  = M_Scan->Dest;
	    if ( Reseau->T[k].Tr_Law->Parameters[0] == 0.0 ) 
	      Adj0[k*The_Size+i] = TRUE;
	  }
	}
      }
    }
  /* Add the dependencies induced by Competition routing with priorities */
  /* Dependencies between objects that are tagged are left because this	*/
  /* makes my life simpler, and I don't think it will hurt.		*/
  for (i=0; i<Reseau->Nb_Places; i++)
    if ( ( Reseau->P[i].Tag == 0 ) 
	 && ( Reseau->P[i].Routing != NULL ) ) {
      if ( Reseau->P[i].Routing->Type == 'P' )
	for ( j=0; j<Reseau->P[i].Outdegree-1; j++ ) {
	  k = Reseau->P[i].Routing->Cycle[j]->Number;
	  l = Reseau->P[i].Routing->Cycle[j+1]->Number;
	  Adj0[k*The_Size+l] = TRUE;
	}
    }

  if ( Debug ) {
    printf("Adj matrix:\n");
    for (i=0; i<The_Size; i++ ) {
      for ( j=0; j<The_Size; j++ )
	printf("%1s", ( Adj0[j*The_Size+i] ? "x" : " " ) );
      printf("| %s\n", Reseau->T[i].Name );
    }
  }

  The_Layer_Struct->Level = 
    (Layer_Desc*) calloc ( (unsigned)The_Size, sizeof( Layer_Desc ) );

  Nb_Layers = find_layer_structure( Adj0, The_Size, The_Layer_Struct->Level );

  if ( Nb_Layers == 0 ) {
    printf("Error: the network has a zero time cycle.\n" );
    Cycle = find_cycle( Adj0, The_Size, FALSE );
    print_trans_cycle( Reseau, Cycle );

    free( Adj0 );

    return FALSE;
  }
  else {
    The_Layer_Struct->Nb_Levels = Nb_Layers;

    free( Adj0 );

    return TRUE;
  }
}

/* Cleanup procedure */
void Free_Layer_Struct( The_Layer_Struct )
Layer_Struct	The_Layer_Struct;
{
  int	i;

  for ( i=0; i<The_Layer_Struct->Nb_Levels; i++ ) {
    free( The_Layer_Struct->Level[i]->Node_Id);
    free( The_Layer_Struct->Level[i] );
  }

  free( The_Layer_Struct->Level );
  free( The_Layer_Struct );

}
