/****************************************************************/
/* module pour generer des reseaux ATM en langage QNAP.		*/
/* Buffers en entree et en sortie a chaque port de chaque switch*/
/* sessions avec priorite stricte selon leur ordre dans le      */
/* fichier de description                                       */
/****************************************************************/

#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>

#define TOOLNAME	"atm2pn"
#define VERSION		"v1.0"

typedef int boolean;

#define TRUE 1
#define FALSE 0

typedef struct LAW_DESC {
	char	Name;
	double*	Parameters;
  	struct LAW_DESC	**Law_Parameters;	/* table pour lois "annexes */
} LAW_DESC;
typedef LAW_DESC *Law;

extern char	*todays_date();

boolean Debug = FALSE;

int NbSession;
int NbSwitch;    

int             *capaIn;  /* input buffer capacity */
int             *capaOut; /* output buffer capacity */


int   ***route;     /* routes of each session */
Law   *SourceLaw;  /* description of the source of each session */
int   **NumSw;     /* names of switches crossed by a session */
int   *NbThrough;   /* nb of switches crossed by a session */
int   *inBufferI;  /* indices of transitions in the buffer In the switch */
int   *outBufferI;
int   *inBufferO;
int   *outBufferO;
int   **inside;  /* inside the switch */

int   *SwitchSize; /* taille des switches */
int   *Layer;     /* nb of layers in a switch */
int   *Nsessions; /* nb of Nsessions in a switch*/
int   **NumSe;     /* names of the sessions in a switch */

int   ***Num_Class;	/* number of class for session [i] in switch [j]*/
			/* column [k] */
int   ***Max_Class;	/* number of classes seen in switch [i], col [j]*/
			/* row [k] */


/************************************************************************/
/* Procedure to write the law. Copied from the same from pn2qnap.c	*/
/************************************************************************/
void fprint_law(
FILE		*OutFile,
Law		The_Law,
int		Obj_Id )
{
  double	Alpha;
  double	Beta;
  double	Delta;
  double	Lambda;
  double	s1;
  double	s2;

  switch( The_Law->Name ) {
  case 'E':
    fprintf(OutFile, "EXP( %8.4f );\n", The_Law->Parameters[0] );
    break;
  case 'C':
    fprintf(OutFile, "CST( %8.4f );\n", The_Law->Parameters[0] );
    break;
  case 'B':
    fprintf(OutFile, "IF RANDU > %8.6f THEN CST(0.0) ELSE CST(1.0);\n", 
	   The_Law->Parameters[0] );
    break;
  case 'g':
    fprintf(OutFile, "CST( FIX( EXP( -1/LOG(%8.4f) ) ) );\n", 
	   The_Law->Parameters[0] );
    break;
  case 'G':
    fprintf(OutFile, "CST( 1.0 + FIX( EXP( -1/LOG(%8.4f) ) ) );\n", 
	   The_Law->Parameters[0] );
    break;
  case 'U':
    fprintf(OutFile, "UNIFORM( %8.4f, %8.4f );\n",
	   The_Law->Parameters[0],
	   The_Law->Parameters[1] );
    break;
  case 'R':
    fprintf(OutFile, "ERLANG( %8.4f, %4.0f );\n",
	   The_Law->Parameters[1],
	   The_Law->Parameters[0] );
    break;
  case 'O':
    Alpha = 1.0 / The_Law->Parameters[1];
    Beta = 1.0 / The_Law->Parameters[2];
    Lambda = (1.0+Alpha/Beta) / The_Law->Parameters[0];
    Delta = sqrt( Alpha*Alpha + 2.0*Alpha*(Beta+Lambda)
		  + (Beta-Lambda)*(Beta-Lambda) );
    s1 = -( Alpha+Beta+Lambda+Delta )/2.0;
    s2 = -( Alpha+Beta+Lambda-Delta )/2.0;
    fprintf(OutFile, "IF RANDU < %8.6f THEN EXP( %8.4f ) ELSE EXP( %8.4f );\n",
	   (-Lambda-s2)/(s1-s2), -1.0/s1, -1.0/s2 );
    break;
   case 'H':
    fprintf(OutFile, "HEXP( %8.4f, %8.4f );\n",
	   The_Law->Parameters[0],
	   The_Law->Parameters[1] );
    break;
  case 'D':
    fprintf(stderr, "%s: error: can't handle timing law (%c)\n",
	     TOOLNAME, The_Law->Name );
    break;
  case 'P':
    /* this part assumes implicitly that the object is a transition	*/
    fprintf(OutFile, "BEGIN\n");
    fprintf(OutFile, "        TR_%d.NextArr := EXP( %8.4f );\n",
	   Obj_Id, The_Law->Parameters[0] );
    fprintf(OutFile, "        WHILE ( TR_%d.NextArr > TR_%d.NextChg )\n",
	   Obj_Id, Obj_Id );
    fprintf(OutFile, "        DO BEGIN\n");
    fprintf(OutFile, "            TR_%d.NextArr := TR_%d.NextArr - TR_%d.NextChg;\n",
	   Obj_Id, Obj_Id, Obj_Id );
    fprintf(OutFile, "            CST( TR_%d.NextChg );\n", Obj_Id );
    fprintf(OutFile, "            ");
    fprint_law(OutFile, The_Law->Law_Parameters[1], Obj_Id );
    fprintf(OutFile, "            TR_%d.NextChg := ", Obj_Id );
    fprint_law(OutFile, The_Law->Law_Parameters[0], Obj_Id );
    fprintf(OutFile, "        END;\n" );
    fprintf(OutFile, "        CST( TR_%d.NextArr );\n", Obj_Id );
    fprintf(OutFile, "        TR_%d.NextChg := TR_%d.NextChg - TR_%d.NextArr;\n", 
	   Obj_Id, Obj_Id, Obj_Id );
    fprintf(OutFile, "    END;\n" );
    break;
  default:
    fprintf( stderr, "%s: error: unknown timing law (%c)\n",
	     TOOLNAME, The_Law->Name );
  }
}


/*************************/
/* generate QNAP code  */
/*************************/

void gen_declarations( char *In_File_Name, FILE *OutFile, boolean AllBounded )
{
  int		i;
  int		j;
  int		Mx;
  char		*The_Date;

  fprintf(OutFile, "& QNAP2 model generated by %s version %s\n", 
	  TOOLNAME, VERSION );
  if ( strlen( In_File_Name ) > 0 )
    fprintf( OutFile, "& Origin: %s\n", In_File_Name );
  The_Date = todays_date();
  fprintf( OutFile, "& Date: %s\n", The_Date );
  free( The_Date );
  fprintf(OutFile, "\n");
  
  fprintf(OutFile, "/CONTROL/\n");
  fprintf(OutFile, "&    OPTION = TRACE;\n");
  fprintf(OutFile, "&    OPTION = SNOURCE;\n");
  fprintf(OutFile, "&    OPTION = NRESULT;\n");
  fprintf(OutFile, "\n");
  
  fprintf(OutFile, "/DECLARE/\n\n");

  fprintf(OutFile, "&   Classes for sessions ");
  fprintf(OutFile, "(declaring the maximum possible, i.e. 18!)\n");
  if ( NbSession > 0 )
    fprintf(OutFile, "    CLASS  ");
  else 
    fprintf(OutFile, "&   No Sessions!!!\n");

  Mx = ( NbSession > 18 ? 17 : NbSession-1 );
  for (i=0; i<=Mx;i++) {
    fprintf(OutFile, "Cl_%d", i );
    if ( i != Mx )
      fprintf(OutFile, ", ");
    if ( ( 0 == (i+1) % 8 ) && ( i != Mx ) )
      fprintf(OutFile, "\n           ");
  }
  fprintf(OutFile,";\n");
  fprintf(OutFile,"    CLASS   Token;\n");
  fprintf(OutFile,"    CLASS   NewCust;\n");

  fprintf(OutFile, "&   Sources for sessions, and leaky buckets\n");
  if ( NbSession > 0 ) {
    fprintf(OutFile, "    QUEUE Src(0:%d);\n", NbSession-1 );
    fprintf(OutFile, "    QUEUE TokSrc(0:%d);\n", NbSession-1 );
    fprintf(OutFile, "    QUEUE Join(0:%d);\n", NbSession-1 );
    fprintf(OutFile, "    QUEUE TokLoss(0:%d);\n", NbSession-1 );
    fprintf(OutFile, "    QUEUE AdmLoss(0:%d);\n", NbSession-1 );
  }
  else 
    fprintf(OutFile, "&   No Sessions!!!\n");


  fprintf(OutFile, "&   Links for sessions between Switches\n");

  for (i=0;i<NbSession;i++) {
    if ( NbThrough[i] > 1 ) {
      fprintf(OutFile, "    QUEUE  ");
      for (j=1;j<NbThrough[i];j++) {
	fprintf(OutFile, "L_%d_%d", i, NumSw[i][j] );
      }
      fprintf(OutFile,";\n");
    }
  }
  fprintf(OutFile, "&   Nodes within Switches and overflow stations.\n");
  fprintf(OutFile, "&   Note the Sw(col,row) convention\n");
  for (i=0; i<NbSwitch; i++) {
    fprintf(OutFile, "    QUEUE  Sw%i(0:%d,0:%d);\n",
	    i, Layer[i]+1, SwitchSize[i]-1 );
    fprintf(OutFile, "    QUEUE  OvfIn%d(0:%d),", i, SwitchSize[i]-1 );
    fprintf(OutFile, " OvfOut%d(0:%d);\n", i, SwitchSize[i]-1 );
  }
  if ( NbSwitch >= 100 )
    printf("%s: warning: too many switches. Names may not be unique.\n",
	   TOOLNAME );

  fprintf(OutFile, "&   Sinks.\n");
  fprintf(OutFile, "    QUEUE  Sink(0:%d);\n", NbSession - 1 );
  if ( AllBounded )
    fprintf(OutFile, "    QUEUE  GlobLoss;\n" );
  fprintf(OutFile, "\n" );

}

void gen_stations( FILE *OutFile, boolean AllBounded )
{

  int		i;
  int		j;
  int 		k;
  int		l;
  int		m;
  int		Count;
  int		NextSwitch;
  int		NextClass;

  /* all the sources */
  for (i=0;i<NbSession;i++) {
    fprintf(OutFile, "/STATION/\n" );
    fprintf(OutFile, "    NAME = Src(%d);\n", i );
    fprintf(OutFile, "    TYPE = SOURCE;\n" );
    fprintf(OutFile, "    SERVICE = BEGIN\n" );
    fprintf(OutFile, "        ");
    fprint_law( OutFile, SourceLaw[i], i );
    fprintf(OutFile, "        IF ( SERVNB( Join( %d ) ) + SERVNB(", i );
    fprintf(OutFile, " AdmLoss( %d ) ) < SERVNB(", i );
    fprintf(OutFile, " Src( %d ) ) - 100 )\n", i );
    fprintf(OutFile, "        THEN TRANSIT( AdmLoss( %d ) );\n", i );
    fprintf(OutFile, "    END;\n");
    fprintf(OutFile, "    TRANSIT = Join( %d ), NewCust;\n", i );
    fprintf(OutFile, "\n");
  }

  for (i=0;i<NbSession;i++) {
    fprintf(OutFile, "/STATION/\n" );
    fprintf(OutFile, "    NAME = TokSrc( %d );\n", i );
    fprintf(OutFile, "    TYPE = SOURCE;\n" );
    fprintf(OutFile, "    SERVICE = BEGIN\n");
    fprintf(OutFile, "        CST( 2.0 );\n" );
    fprintf(OutFile, "        IF ( SERVNB( Join( %d ) ) + SERVNB(", i );
    fprintf(OutFile, " TokLoss( %d ) ) < SERVNB(", i );
    fprintf(OutFile, " TokSrc( %d ) ) - 10 )\n", i );
    fprintf(OutFile, "        THEN TRANSIT( TokLoss( %d ) );\n", i );
    fprintf(OutFile, "    END;\n");
    fprintf(OutFile, "    TRANSIT = Join( %d ), Token;\n", i );
    fprintf(OutFile, "\n");
  }

  for (i=0;i<NbSession;i++) {
    fprintf(OutFile, "/STATION/\n" );
    fprintf(OutFile, "    NAME = Join( %d );\n", i );
    NextClass = Num_Class[ i ][ NumSw[i][0] ][0];
    fprintf(OutFile, "    FUSION = ( NewCust, 1, Token, 1 ) Cl_%d;\n",
	    NextClass );
    fprintf(OutFile, "    SERVICE = CST( 0.0 );\n" );
    fprintf(OutFile, "    TRANSIT = Sw%d( 0, %d ), Cl_%d, 1.0;\n",
	    NumSw[i][0], route[i][ NumSw[i][0] ][0]-1, NextClass );
    fprintf(OutFile, "\n");
  }

  fprintf(OutFile, "/STATION/\n" );
  fprintf(OutFile, "    NAME = TokLoss( 0 STEP 1 UNTIL %d ),", 
	  NbSession - 1 );
  fprintf(OutFile, " AdmLoss( 0 STEP 1 UNTIL %d );\n", 
	  NbSession - 1 );
  fprintf(OutFile, "    SERVICE = CST( 0.0 );\n" );
  fprintf(OutFile, "    TRANSIT = OUT;\n");
  fprintf(OutFile, "\n");

  fprintf(OutFile, "  \n& declaration of the switches \n\n");

  for (i=0;i<NbSwitch;i++) {
    fprintf(OutFile, "/STATION/\n" );
    fprintf(OutFile, "    NAME = Sw%d( 0 STEP 1 UNTIL %d,",
	    i, Layer[i]+1 );
    fprintf(OutFile, " 0 STEP 1 UNTIL %d );\n",  SwitchSize[i]-1 );
    fprintf(OutFile, "    SERVICE = CST( 1.0 );\n");
    fprintf(OutFile, "\n");
  }

  fprintf(OutFile, "  \n& declaration of the routing inside switches \n\n");

  for (i=0;i<NbSwitch;i++) {
    for( j=0; j<=Layer[i]+1; j++ ) {
      for ( k=0; k<SwitchSize[i]; k++ ) {
	/* Enumerate all routes that go through this queue */
	/* First time to count them, second time to print their ids */
	Count = 0;
	for ( l=0; l<NbSession; l++ ) {
	  for ( m=0; m<NbThrough[l]; m++ ) {
	    if ( NumSw[l][m] == i ) {
	      /* this is the right switch */
	      if ( route[l][i][j] == k+1 ) {
		/* this is the right row k in layer (column) j in switch i */
		Count++;
	      }
	    }
	  }
	}
	if ( Count > 0 ) {
	  if ( Debug ) {
	    printf("Found %d sessions going through Sw%d(%d,%d)\n",
		   Count, i, j, k );
	  }
	  fprintf(OutFile, "/STATION/\n");
	  fprintf(OutFile, "    NAME = Sw%d( %d, %d );\n", i, j, k );
	  if ( Count > 1 )
	    fprintf(OutFile, "    SCHED = PRIOR, PREEMPT;\n");
	  if ( j == 0 ) {
	    fprintf(OutFile, "    CAPACITY = %d;\n", capaIn[i] );
	    fprintf(OutFile, "    REJECT = TRANSIT( OvfIn%d(%d) );\n",
		    i, k );
	  }
	  else if ( j == Layer[i]+1 ) {
	    fprintf(OutFile, "    CAPACITY = %d;\n", capaOut[i] );
	    fprintf(OutFile, "    REJECT = TRANSIT( OvfOut%d(%d) );\n",
		    i, k );
	  }
	  else if ( AllBounded ) {
	    fprintf(OutFile, "    CAPACITY = %d;\n", capaOut[i] );
	    fprintf(OutFile, "    REJECT = TRANSIT( GlobLoss );\n" );
	  }

	  for ( l=0; l<NbSession; l++ ) {
	    for ( m=0; m<NbThrough[l]; m++ ) {
	      if ( NumSw[l][m] == i ) {
		/* this is the right switch */
		if ( route[l][i][j] == k+1 ) {
		  /* this is the right row k in layer (column) j in switch i */
		  NextClass = Num_Class[l][i][j];
		  if ( Count > 1 ) {
		    fprintf(OutFile, "    PRIOR( Cl_%d ) = %d;\n",
			   NextClass, NbSession - l );
		  }
		  if ( j <= Layer[i] ) {
		    /* routing is internal to the switch */
		    fprintf(OutFile, 
			    "    TRANSIT( Cl_%d ) = Sw%d( %d, %d )",
			    NextClass, i, j+1, route[l][i][j+1]-1 );
		    NextClass =  Num_Class[l][i][j+1];
		    fprintf(OutFile, ", Cl_%d;\n", NextClass );
		  }
		  else {
		    /* routing is to link or sink */
		    if ( m == NbThrough[l] - 1 ) {
		      fprintf(OutFile, "    TRANSIT( Cl_%d ) = Sink(%d);\n", 
			      NextClass, l );
		    }
		    else {
		      fprintf(OutFile, "    TRANSIT( Cl_%d ) = L_%d_%d;\n", 
			      NextClass, l, NumSw[l][m+1] );
		    }
		  }
		}
	      }
	    }
	  }
	}
	/* done with Sw i / col j / row k */
	fprintf(OutFile, "\n");
      }
    }
  }

  fprintf(OutFile, "\n& Declaration of overflow queues.\n");
  for ( i=0; i<NbSwitch; i++ ) {
    fprintf(OutFile, "/STATION/\n");
    fprintf(OutFile, "    NAME = OvfIn%d( 0 STEP 1 UNTIL %d );\n",
	    i, SwitchSize[i]-1 );
    fprintf(OutFile, "    SERVICE = CST( 0.0 );\n");
    fprintf(OutFile, "    TRANSIT = OUT;\n");
    fprintf(OutFile, "\n");
  }
  for ( i=0; i<NbSwitch; i++ ) {
    fprintf(OutFile, "/STATION/\n");
    fprintf(OutFile, "    NAME = OvfOut%d( 0 STEP 1 UNTIL %d );\n",
	    i, SwitchSize[i]-1 );
    fprintf(OutFile, "    SERVICE = CST( 0.0 );\n");
    fprintf(OutFile, "    TRANSIT = OUT;\n");
    fprintf(OutFile, "\n");
  }

  /* all the sinks (one per session ) */
  fprintf(OutFile, "\n& Declaration of the sinks (one per session )\n");
  fprintf(OutFile, "/STATION/\n");
  fprintf(OutFile, "    NAME = Sink( 0 STEP 1 UNTIL %d );\n",
	  NbSession-1 );
  fprintf(OutFile, "    SERVICE = CST( 0.0 );\n");
  fprintf(OutFile, "    TRANSIT = OUT;\n");

  if ( AllBounded ) {
    fprintf(OutFile, "\n& Declaration of the ultimate trap\n");
    fprintf(OutFile, "/STATION/\n");
    fprintf(OutFile, "    NAME = GlobLoss;\n" );
    fprintf(OutFile, "    SERVICE = CST( 0.0 );\n");
    fprintf(OutFile, "    TRANSIT = OUT;\n");
    fprintf(OutFile, "\n");
  }

  /* all the links between the switches */
  fprintf(OutFile, "\n& Declaration of the links\n");
  for (i=0;i<NbSession;i++) {
    for (j=1;j<NbThrough[i];j++) {
      NextSwitch = NumSw[i][j];
      fprintf(OutFile, "/STATION/\n");
      fprintf(OutFile, "    NAME = L_%d_%d;\n", i, NextSwitch );
      fprintf(OutFile, "    SERVICE = CST(0.0);\n");
      fprintf(OutFile, "    TRANSIT = Sw%d( 0, %d ), Cl_%d;\n", 
	      NextSwitch, route[i][ NextSwitch ][0]-1,
	      Num_Class[i][ NextSwitch ][0] );
    }
  }
}

void gen_control( FILE *OutFile, double T_Max )
{
  /***** parametre de la simulation ********/
  fprintf(OutFile, "\n");
  fprintf(OutFile, "/CONTROL/\n");
  fprintf(OutFile, "    TMAX = %12f;\n", T_Max );
  fprintf(OutFile, "\n");

  /******* display of the results *****/

  fprintf(OutFile, "/EXEC/\n");
  fprintf(OutFile, "    BEGIN;\n");
  fprintf(OutFile, "        SIMUL;\n");
  fprintf(OutFile, "        WRITELN(\"Total time: \", GETCPUT);\n");
  fprintf(OutFile, "    END;\n");
  fprintf(OutFile, "\n");

  fprintf(OutFile, "/END/\n");
}

/************************************************************************/
/* A procedure for reading a stochastic law description and returning	*/
/* a structure describing it.						*/
/************************************************************************/
Law Get_Law( int Session_Nb )
{
  int	Nb_Params;
  int	i;
  Law	Res;
  char	Token[4];

  Res = (Law) malloc( sizeof( struct LAW_DESC ) );
  scanf( "%2s", Token );
  /* 
     printf("Got[%s]\n", Token );
     */
  if ( strlen(Token) > 0 ) {
    Res->Name = Token[0];
    switch( Res->Name ) {
    case 'E':
    case 'C':
    case 'B':
    case 'g':
    case 'G':
      Res->Parameters = (double*) malloc( sizeof(double) );
      scanf( "%lf", &(Res->Parameters[0] ) );
      break;
    case 'R':
    case 'U':
    case 'H':
      Res->Parameters = (double*) malloc ( 2*sizeof(double) );
      scanf( "%lf", &(Res->Parameters[0] ) );
      scanf( "%lf", &(Res->Parameters[1] ) );
      break;
    case 'O':
      Res->Parameters = (double*) malloc( 3*sizeof( double ) );
      scanf( "%lf", &(Res->Parameters[0]) );
      scanf( "%lf", &(Res->Parameters[1]) );
      scanf( "%lf", &(Res->Parameters[2]) );
      break;
    case 'D':
      /* find the length of the (first) parameter list */
      Nb_Params = 0;
      scanf( "%d", &Nb_Params );
      Res->Parameters = (double*) calloc( 1 + 2*Nb_Params, sizeof( double ) );
      Res->Parameters[0] = Nb_Params;
	
      /* Store values of the discrete distribution */
      for( i=1; i<=2*Nb_Params; i++ )
	scanf( "%lf", &(Res->Parameters[i]) );
	
      break;
	
    case 'P':
      /* IPP-type law. */
      Res->Parameters = (double*) calloc( 1, sizeof( double ) );
      scanf( "%lf", &(Res->Parameters[0]) );
      Res->Law_Parameters = (LAW_DESC **) 
	calloc( 2, sizeof( struct LAW_DESC* ) );
      Res->Law_Parameters[0] = Get_Law( Session_Nb );
      Res->Law_Parameters[1] = Get_Law( Session_Nb );
      break;
	
    default:
      printf("%s: unknown law name: %c\n", TOOLNAME, Res->Name );
    }
  }
  else {
    printf("%s: Error reading law name for session number %d.\n",
	   TOOLNAME, Session_Nb );
    free( Res );
    Res = (Law)NULL;
  }

  return Res;

}

/**************************************/
/* programme principal                */
/**************************************/
int main(int	argc,
	 char	*argv[] )
{

  char Out_File_Name[128]= "";	  /* nom du fichiers QNAP    */
  char In_File_Name[128]= "";     /* nom du fichier d'entree */
      
  int  Args_OK=TRUE;
  int  InOut = TRUE;
  int i,j,k,J,exp2k;
  int PortIn,PortOut,Port;
  double 	aa;
  double	T_Max = 1000.0;
  boolean	AllBounded = FALSE;
  int		NextClass;
  int		The_Sw;
  int		The_Row;
  FILE  	*File;

  /************************* Parsing *****************************/
  i = 1;

  if (argc>1)
    do {

      if ( 0 == strcmp( "-f", argv[i] ) )
	{	
	  i++;
	  if ( i < argc )
	    {	
	      strncpy( In_File_Name, argv[i], 128 );
	      i++;
	    }
	  else 
	    {	
	      fprintf(stderr,"%s: usage:filename expected after -f\n",argv[0]);
	      Args_OK = FALSE;
	    }
	}
      else if ( 0 == strcmp( "-o", argv[i] ) )
	{	
	  i++;
	  if ( i < argc )
	    {	
	      strncpy( Out_File_Name, argv[i], 128 );
	      i++;
	    }
	  else 
	    {	
	      fprintf(stderr,"%s: usage:filename expected after -o\n",argv[0]);
	      Args_OK = FALSE;
	    }
	}
      else if ( 0 == strcmp( "-debug", argv[i] ) ) {
	i++;
	Debug = TRUE;
      }
      else if ( 0 == strcmp( "-allbounded", argv[i] ) ) {
	i++;
	AllBounded = TRUE;
      }
      else if ( 0 == strcmp( "-tmax", argv[i] ) ) {
	i++;
	if ( 1 != sscanf( argv[i], "%lf", &T_Max ) ) {
	  fprintf(stderr,"%s: usage: real expected after -tmax\n",
		  TOOLNAME );
	  Args_OK = FALSE;
	}
	else 
	  i++;
      }

      else if (( 0 == strcmp( "-h", argv[i] )) || 
	       ( 0 == strcmp( "-?", argv[i] )) ||
	       ( 0 == strcmp( "-help", argv[i] ))) {
	i++;
	fprintf(stdout,"usage: %s [-f inputfile] [-o outfile]", argv[0] );
	fprintf(stdout," [-tmax real] [-allbounded]\n");
	Args_OK = FALSE;
      }
      else {
	/* skip silently unknown argument */
	i++;
      }
    } while( i < argc );

  if (Args_OK)
    {
      if  ( ( strlen(In_File_Name) != 0 ) &&
	    (NULL == freopen( In_File_Name, "r", stdin ) ))
	{
	  fprintf(stderr,"%s: could not open file %s\n",
		  TOOLNAME, In_File_Name );
	  InOut = FALSE;
	}
    }


  /**************** start of the program if everything is OK *******************/

  if (Args_OK && InOut ) {

      /* read the input and alloc the arrays */
      scanf("%d",&NbSwitch);
      SwitchSize = (int *) calloc(NbSwitch,sizeof(int));
      Layer = (int *) calloc(NbSwitch,sizeof(int));
      capaIn = (int *) calloc(NbSwitch,sizeof(int));
      capaOut= (int *) calloc(NbSwitch,sizeof(int));

      for (i=0;i<NbSwitch;i++)
	{
	  scanf("%d",&(SwitchSize[i]));
	  scanf("%d",&(capaIn[i]));
	  scanf("%d",&(capaOut[i]));

	  /*  test power of 2 */
	  for( j=1; j<=SwitchSize[i]; j*=2 );
	  if ( (j/2) != SwitchSize[i] ) {
	    printf("%s: Error: size of switch %d (=%d) is not",
		   TOOLNAME, i, SwitchSize[i] );
	    printf(" a power of 2.\n");
	  }

	  aa=(double)SwitchSize[i];
	  aa=log(aa)/log(2);
	  aa=(int)rint(aa);
	  Layer[i]=aa-1;
	}


      scanf("%d",&NbSession);

      SourceLaw = (Law*) calloc( NbSession, sizeof(Law) );

      NumSw = (int **) calloc(NbSession,sizeof(int*));
      for (i=0;i<NbSession;i++)
	{
	  NumSw[i] = (int *) calloc(NbSwitch, sizeof(int));
	}

      NbThrough =(int *) calloc(NbSession, sizeof(int));

      inBufferI=(int *) calloc(NbSession, sizeof(int));

      outBufferI=(int *) calloc(NbSession, sizeof(int));

      inBufferO=(int *) calloc(NbSession, sizeof(int));

      outBufferO=(int *)calloc(NbSession, sizeof(int));

      inside = (int **) calloc(NbSession,sizeof(int*));
      for (i=0;i<NbSession;i++)
	{
	  inside[i]=(int *) calloc(NbSwitch, sizeof(int*));
	}


      Nsessions =(int *) calloc(NbSwitch, sizeof(int));
      for (j=0;j<NbSwitch;j++)
	{
	  Nsessions[j] = 0;
	}

      NumSe = (int **) calloc(NbSwitch,sizeof(int*));
      for (i=0;i<NbSwitch;i++)
	{
	  NumSe[i] = (int *) calloc(NbSession, sizeof(int));
	}

      route = (int ***) calloc(NbSession,sizeof(int**));
      Num_Class = (int ***) calloc(NbSession,sizeof(int**));
      for (i=0;i<NbSession;i++)
	{
	  route[i]=(int **) calloc(NbSwitch, sizeof(int*));
	  Num_Class[i]=(int **) calloc(NbSwitch, sizeof(int*));
    
	  for (j=0;j<NbSwitch;j++)
	    {
	      route[i][j]=(int*) calloc(Layer[j]+2, sizeof(int));
	      Num_Class[i][j]=(int*) calloc(Layer[j]+2, sizeof(int));
	    }
    
	}

      /* init the route/class array */
      for (i=0;i<NbSession;i++)
	{
	  for (j=0;j<NbSwitch;j++)
	    {
	      for (k=0;k<Layer[j]+1;k++)
		{
		  route[i][j][k]=-1;
		  Num_Class[i][j][k]=-1;
		}
	    }
	}

      Max_Class = (int ***) calloc(NbSwitch,sizeof(int**));
      for (i=0;i<NbSwitch;i++) {
	Max_Class[i] = (int **) calloc(2+Layer[i],sizeof(int*));
	for (j=0;j<=Layer[i]+1;j++) {
	  Max_Class[i][j] = (int *) calloc(SwitchSize[i],sizeof(int));
	}
      }

      /* construct the route array, the NumSw and the NumSe arrays */
      for (i=0;i<NbSession;i++)
	{
	  /* getting the type of source */
	  SourceLaw[i] = Get_Law( i );

	  scanf("%d",&NbThrough[i]);
	  for (J=0;J<NbThrough[i];J++)
	    {
	      scanf("%d",&j);
	      NumSe[j-1][i] = Nsessions[j-1];
	      Nsessions[j-1]++;
	      NumSw[i][J]=j-1;
	      scanf("%d",&PortIn);
	      scanf("%d",&PortOut);

	      /* XXX test de compatibilite pas fait */

	      /* Ancien algo ...
	      route[i][j-1][0] = PortIn;
	      Port=PortIn;
	      for (k=0;k<Layer[j-1];k++)
		{
		  exp2k=1<<k;
		  if ((PortOut-1) & exp2k)
		    {
		      Port = Port + (1-((Port-1) & exp2k))*exp2k;
		    }
		  else
		    {
		      Port = Port - ((Port-1) & exp2k)*exp2k;
		    }
		  route[i][j-1][k+1] = Port;
		}
		*/
/* Nouvel algo */
	      route[i][j-1][0] = PortIn;
	      Port=PortIn-1;
	      for (k=0;k<Layer[j-1];k++)
		{
		  exp2k=1<<k;
		  if ((PortOut-1) & exp2k) {
		    /* PortOut a un 1 dans le bit k */
		    if ( ! ( Port & exp2k ) ) {
		      /* Port a un 0: il faut mettre un 1 */
		      Port = Port + exp2k;
		    }
		  }
		  else {
		    /* PortOut a un 0 dans le bit k */
		    if ( Port & exp2k ) {
		      /* Port a un 1: il faut l'enlever */
		      Port = Port - exp2k;
		    }
		  }
		  route[i][j-1][k+1] = Port+1;
		}
	      route[i][j-1][Layer[j-1]+1] = PortOut;
	    }
	}

      for (i=0;i<NbSession;i++) {
	for (J=0;J<NbThrough[i];J++) {
	  The_Sw = NumSw[i][J];
	  for ( k=0; k<=Layer[ The_Sw ]+1; k++ ) {
	    The_Row = route[i][ The_Sw ][k]-1;
	    Num_Class[i][ The_Sw ][k] =
	      Max_Class[ The_Sw ][k][ The_Row ];
	    Max_Class[ The_Sw ][k][ The_Row ] ++;
	    NextClass = Max_Class[ The_Sw ][k][ The_Row ];
	    if ( NextClass >= 18 ) {
	      Max_Class[ The_Sw ][k][ The_Row ] = 17;
	      fprintf( stderr, "%s: warning: max nb of classes exceeded ",
		       TOOLNAME );
	      fprintf( stderr, "at source %d. 17 Assumed.\n", i );
	    }
	  }
	}
      }

      if ( Debug ) {
	printf("Routing table (row/class):\n");
	for( i=0;i<NbSession;i++ ) {
	  printf("Session %d:", i );
	  for ( J=0; J<NbThrough[i]; J++ ) {
	    printf(" Sw %d[ ", NumSw[i][J] );
	    for ( k=0; k<=Layer[NumSw[i][J]]+1; k++ )
	      printf("%d/%d ",
		     route[i][NumSw[i][J]][k],
		     Num_Class[i][NumSw[i][J]][k] );
	    printf("]");
	  }
	  printf("\n");
	}
      }
      /***** generate QNAP declarations *****/
      if ( strlen( Out_File_Name ) > 0 ) {
	File = fopen( Out_File_Name, "w" );
      }
      else
	File = stdout;
      
      gen_declarations( In_File_Name, File, AllBounded );

      /********* build the network *****/
      
      gen_stations( File, AllBounded );

      /********* make the simulation ****/

      gen_control( File, T_Max );

      return 0;
    }
  else 
    return -1;
}
