/* programme/module pour la simulation equationelle d'un PN avec	*/
/* temporisations deterministes et routage.			 	*/

#include <stddef.h>
#include <values.h>
#include <malloc.h>
#include <string.h>
#include <math.h>		/* for random number generation	*/
#include <sys/time.h>
#include <time.h>

#include "pn.h"

#define TOOLNAME	"pn_sim"

#define MIN(x,y)	( (x)<(y) ? (x) : (y) )
#define IRINT(x)	(int)rint(x)

#define MAX_RANDOM_VALUE          0x7FFFFFFF

typedef struct LAYER_DESC {
  int	Nb_Nodes;
  int	*Node_Id;
} *Layer_Desc;

typedef struct LAYER_STRUCT {
  int		Nb_Levels;
  Layer_Desc	*Level;		/* array of layer descriptions */
} *Layer_Struct;


#define INFINI		MAXINT

#define min(x,y)        ( (x<y) ? x : y )

#define SCAN(x,y)       for(x=y;x!=NULL;x=x->Next)

#ifndef ALPHA
extern long random(void);
#endif 

/* Du module analyze_pn.c */
extern double Max_Lag();
extern double Max_Delay();
extern int    Nb_Available_Tokens();

/* Tables for simulation */
int	**YY;		/* Nb of firings initiated 	*/
boolean	*Is_Routing;	/* Routing flag for places	*/
int	*Count;		/* Nb of tokens arrived in a place */
int	*Phase;		/* Phase for cyclic routing 	*/
int	**HH;		/* Place/Transition Routing function */
int	*Fire_Time;
int	*Hold_Time;

/*======================================================*/
/* Statistiques                                         */
/*======================================================*/
double U_0_1()
{
  return (double)random()/MAX_RANDOM_VALUE;
}

double Bernoulli(Proba)
double Proba;
{
  if (U_0_1() < Proba)
    return 1.0;
  else
    return 0.0;
}

double Max( x, y )
double x;
double y;
{
  return ( x>y ? x : y );
}

/************************************************************************/
/* Functions for the simulation itself					*/
/************************************************************************/
boolean init_simulation(PetriNet 	*Reseau,
			int		Scale,
			int		Buf_Size )
{
  int		i;
  int		j;
  int		The_Size;
  boolean	Alloc_Ok = TRUE;

  The_Size = Reseau->Nb_Trans;

  YY = (int**)calloc( (unsigned)Buf_Size, sizeof( int* ) );

  Alloc_Ok = ( NULL != YY );

  for ( i=0; Alloc_Ok && ( i<Buf_Size ); i++ ) {
    YY[i] = (int*)calloc( (unsigned)The_Size, sizeof( int ) );    
    Alloc_Ok = Alloc_Ok && ( YY[i] != NULL );
    if ( Alloc_Ok )
      for( j=0; j<The_Size; j++ )
	YY[i][j] = 0;
  }

  Fire_Time = (int*) calloc( (unsigned)The_Size, sizeof( int ) );
  Hold_Time = (int*) calloc( (unsigned)Reseau->Nb_Places, sizeof( int ) );
  Count = (int*) calloc( (unsigned)Reseau->Nb_Places, sizeof( int ) );
  Phase = (int*) calloc( (unsigned)Reseau->Nb_Places, sizeof( int ) );
  Is_Routing = (boolean*) calloc( (unsigned)Reseau->Nb_Places, 
				  sizeof( boolean ) );
  HH = (int**) calloc( (unsigned)Reseau->Nb_Places, sizeof( int* ) );

  Alloc_Ok = Alloc_Ok 
    && ( NULL != Fire_Time )
    && ( NULL != Hold_Time )
    && ( NULL != Count ) 
    && ( NULL != Phase ) 
    && ( NULL != HH );

  /* Initialize tables */

  if ( Alloc_Ok ) {
    for ( i=0; i < The_Size; i++ )
      Fire_Time[i] = IRINT( Scale * Reseau->T[i].Tr_Law->Parameters[0] );

    for ( i=0; i < Reseau->Nb_Places; i++ ) {
      Hold_Time[i] = IRINT( Scale * Reseau->P[i].Pl_Law->Parameters[0] );
      Count[i] = 0;
      Phase[i] = 0;
      if ( Reseau->P[i].Routing != NULL ) {
	if ( Reseau->P[i].Routing->Type != 'C' ) {
	  Is_Routing[i] = TRUE;

	  HH[i] = (int*) calloc( (unsigned)The_Size, sizeof( int ) );
	  Alloc_Ok = Alloc_Ok && ( NULL != HH[i] );
	}
	else {
	  Is_Routing[i] = FALSE;
	  HH[i] = NULL;
	}
      }
      else {
	Is_Routing[i] = FALSE;
	HH[i] = NULL;
      }
    }
    
    return Alloc_Ok;
  }
  else
    return FALSE;
    
}

/************************************************************************/
/* A function for computing the number of tokens present AND available	*/
/* in a place at a certain time.					*/
/************************************************************************/

int Place_Marking(Pl_Des	*The_Place,
		  int		Time,
		  int		Buf_Size)
{
  Liste_Link	L_Scan;
  int		Res;
  int		Delay;

  Res = Nb_Available_Tokens( *The_Place, Time );

  SCAN( L_Scan, The_Place->Inputs ) {
    Delay = Hold_Time[ The_Place->Pl_Id ] + Fire_Time[ L_Scan->Dest ];
    Res += ( Time >= Delay ? 
	     YY[(Time-Delay+Buf_Size) % Buf_Size][L_Scan->Dest] 
	     * L_Scan->Weight :
	     0 );
  }

  SCAN( L_Scan, The_Place->Outputs )
    Res -= YY[Time % Buf_Size][L_Scan->Dest] * L_Scan->Weight;

  return Res;
}

boolean Draw_Routing( Reseau, Place_Id, Trans_Id, Time, Buf_Size )
PetriNet	*Reseau;
int 		Place_Id;
int 		Trans_Id;
int		Time;
int		Buf_Size;
{
  int		j;
  int		Marking;
  int		E_Degree;
  double	U;
  Routing_Desc	The_Routing;
  Pl_Des	The_Place;
  Liste_Link	L_Scan;

  The_Routing = Reseau->P[Place_Id].Routing;

  switch ( The_Routing->Type ) {

  case 'R':
    U = U_0_1();
    L_Scan = Reseau->P[Place_Id].Outputs;
    U -= The_Routing->Probas[0];
    for ( j=0; ( U >= 0.0 ) && ( j < Reseau->P[Place_Id].Outdegree ); j++ ) {
      U -= The_Routing->Probas[j+1];
      L_Scan = L_Scan->Next;
    }

    /* j is the next routing choice */
    HH[Place_Id][ L_Scan->Dest ] ++;

    return TRUE;

    break;

  case 'D':
    HH[ Place_Id ][ The_Routing->Cycle[ Phase[ Place_Id ] ]->Number ] ++;
    Phase[ Place_Id ] ++;
    Phase[ Place_Id ] = ( Phase[ Place_Id ] % The_Routing->Period );

    return TRUE;

    break;

  case 'P':
    /* Compute Enabling Degree of transition Trans_Id */
    E_Degree = MAXINT;

    SCAN( L_Scan, Reseau->T[Trans_Id].Inputs ) {
      The_Place = Reseau->P[L_Scan->Dest];

      Marking = Place_Marking( &The_Place, Time, Buf_Size );

      E_Degree = MIN( E_Degree, Marking / L_Scan->Weight );
    }
    if ( E_Degree > 0 ) {
      HH[Place_Id][Trans_Id] += E_Degree;
      Count[Place_Id] += E_Degree;
      return TRUE;
    }
    else
      return FALSE;

    break;

  default:
    printf("%s: error: shouldn't have to compute routing for place %d",
	   TOOLNAME, Place_Id );
    if ( strlen( Reseau->P[Place_Id].Name ) > 0 )
      printf(" (%s)\n", Reseau->P[Place_Id].Name );
    else 
      printf("\n");

    return FALSE;
  }

}

/************************************************************************/
/* Simulate the evolution of the network by computing sucessively	*/
/* the number of firings at time n. The equations used are		*/
/* that "in N" (initiation of firings).					*/
/************************************************************************/
void simulate(
PetriNet 	*Reseau,
Layer_Struct	The_Layer_Struct,
int		Scale,
int		Buf_Size )
{	
  int		i;
  int		j;
  int		Time;
  int		L_Time = 0;	/* "local" time in the circular buffer */
  int		The_Size;
  int		Max_Time;
  int		x;
  int		Tokens;
  int		Tr_Id, Prev_Id;
  int		Max_Dependency;
  double	Tmp;
  int		Delay;
  Liste_Link	L_Scan;
  Liste_Link	M_Scan;
  int		*Idx;		/* table to store precomputed modulos */

  The_Size = Reseau->Nb_Trans;
  Max_Time = Reseau->Sim_Params->Nb_Firings * Scale;

  Tmp = Scale * Max( Max_Delay( Reseau ), Max_Lag( Reseau ) );
  Max_Dependency = IRINT( Tmp );

  if (! init_simulation(Reseau, Scale, Buf_Size ) ) {
    printf("Could not allocate space. Simulation not launched.\n");
  }
  else {
    /* compute modulos */
    Idx = (int*) calloc( (unsigned int)Buf_Size, sizeof(int) );
    for ( i=0; i<Buf_Size; i++ )
      Idx[i] = Buf_Size - i;
			     
    /* Start the simulation */

    /* compute Max_Delay first firing dates. No modulos, as Buf_Size 	*/
    /* is larger than Max_Delay	  			     		*/
    for (Time=0; (Time<=Max_Time) && (Time<Max_Dependency) ; Time++) {

      /* In order to compute enablings for competition, it is necessary	*/
      /* to initialize the firing counts of (some) transitions.		*/
      if ( Time > 0 )
	for ( i=0; i<Reseau->Nb_Trans; i++ )
	  YY[Time][i] = YY[Time-1][i];

      for ( i=0; i<The_Layer_Struct->Nb_Levels; i++)
	for ( j=0; j<The_Layer_Struct->Level[i]->Nb_Nodes; j++) {

	  Tr_Id = The_Layer_Struct->Level[i]->Node_Id[j];

	  x = MAXINT;

	  SCAN( L_Scan, Reseau->T[Tr_Id].Inputs ) {
	    Prev_Id = L_Scan->Dest;

	    Tokens = Nb_Available_Tokens( Reseau->P[Prev_Id], Time );

	    SCAN( M_Scan, Reseau->P[Prev_Id].Inputs ) {
	      Delay = Hold_Time[ Prev_Id ] + Fire_Time[ M_Scan->Dest ];

	      if ( Time >= Delay )
		Tokens += YY[ Time - Delay ][ M_Scan->Dest ]
		  * M_Scan->Weight;
	    }

	    if ( Is_Routing[ Prev_Id ] ) {
	      /* if routing, then ..->Routing is not null.		*/
	      if ( Reseau->P[Prev_Id].Routing->Type == 'P' ) {
		/* simply compute how many tokens go through.		*/
		Draw_Routing( Reseau, Prev_Id, Tr_Id, Time, Buf_Size );
	      }
	      else {
		/* Advance the number of routed tokens as far as possible */
		for( ; ( Count[ Prev_Id ] < Tokens ) &&
		       Draw_Routing( Reseau, Prev_Id, Tr_Id, Time, Buf_Size );
		     Count[ Prev_Id ]++ );
	      }

	      Tokens = HH[ Prev_Id ][ Tr_Id ];
	    }

	    Tokens = Tokens / L_Scan->Weight;

	    x = min( x, Tokens );
	  }

	  YY[Time][Tr_Id] = x;
	}
    }
    L_Time = Time;

    /* compute the rest... */
    for ( ; Time <= Max_Time ; Time++ ) {

      /* In order to compute enablings for competition, it is necessary	*/
      /* to initialize the firing counts of (some) transitions.		*/
      for ( i=0; i<Reseau->Nb_Trans; i++ )
	YY[L_Time][i] = YY[(L_Time+Buf_Size-1) % Buf_Size][i];

      for ( i=0; i<The_Layer_Struct->Nb_Levels; i++)
	for ( j=0; j<The_Layer_Struct->Level[i]->Nb_Nodes; j++) {

	  Tr_Id = The_Layer_Struct->Level[i]->Node_Id[j];

	  x = INFINI;

	  SCAN( L_Scan, Reseau->T[Tr_Id].Inputs ) {
	    Prev_Id = L_Scan->Dest;

	    Tokens = Reseau->P[Prev_Id].Init_Pop;

	    SCAN( M_Scan, Reseau->P[Prev_Id].Inputs ) {
	      Delay = Hold_Time[ Prev_Id ] + Fire_Time[ M_Scan->Dest ];

	      Tokens += YY[ ( L_Time - Delay + Buf_Size )
					  % Buf_Size ][ M_Scan->Dest ]
		* M_Scan->Weight;
	    }

	    if ( Is_Routing[ Prev_Id ] ) {
	      /* if routing, then ..->Routing is not null.		*/
	      if ( Reseau->P[Prev_Id].Routing->Type == 'P' ) {
		/* simply compute how many tokens go through.		*/
		Draw_Routing( Reseau, Prev_Id, Tr_Id, Time, Buf_Size );
	      }
	      else {
		/* Advance the number of routed tokens as far as possible */
		for( ; ( Count[ Prev_Id ] < Tokens ) &&
		       Draw_Routing( Reseau, Prev_Id, Tr_Id, Time, Buf_Size );
		     Count[ Prev_Id ]++ );
	      }

	      Tokens = HH[ Prev_Id ][ Tr_Id ];
	    }

	    Tokens = Tokens / L_Scan->Weight;

	    x = min( x, Tokens );

	  }

	  YY[L_Time][Tr_Id] = x;
	}

      L_Time++;
      if ( L_Time == Buf_Size )
	L_Time = 0;
    }
  } /* fin du "else" */
}

void dumpe_simulation(Reseau, Buf_Size)
PetriNet	*Reseau;
int		Buf_Size;
{
  int		i,j;
  double	val;
  int		Max_Time;
  int		Max_Delay;

  Max_Time = Reseau->Sim_Params->Nb_Firings;
  Max_Delay = Reseau->Max_Pop_Init;

  for (i=0; i<Reseau->Nb_Trans; i++) {
    printf("Transition %d ...\n",i);
    j = ( Max_Time-Max_Delay >= 0 ? Max_Time-Max_Delay : 0 );
    for ( ; j<=Max_Time; j++) {
      val = YY[j % Buf_Size ][i];
      if ( val > 100000.0 )
	printf("oo ");
      else
	printf("%8.3f", val );
      printf("\n");
    }
  }
}

/****************************************************************/
/* Procedure de calcul du temps ecoule                          */
/* (AJM | 16/11/92 )                                            */
/****************************************************************/
double elapsed_time( t_start, t_end )
struct timeval t_start, t_end;
{
  return (double)((t_end.tv_sec - t_start.tv_sec) * 1000000.0
		  + t_end.tv_usec - t_start.tv_usec)/1000000.0;
}

void write_stat_results(Reseau, Scale, Buf_Size, t_debut, t_fin)
PetriNet 	*Reseau;
int		Scale;
int		Buf_Size;
struct timeval	t_debut;
struct timeval	t_fin;
{
  int		i;
  int		Max_Time;
  int		Nb_Evt;

  Max_Time = Scale * Reseau->Sim_Params->Nb_Firings;

  for (i=0; i<Reseau->Nb_Trans; i++) {
    printf("Y_%1d[%1d] = %5d      Cycle time: ",
	   i, Max_Time / Scale,
	   YY[Max_Time % Buf_Size ][i] );
    if ( 0 == YY[Max_Time % Buf_Size ][i] )
      printf("    oo    ");
    else
      printf("%10.4f", 
	     Max_Time / (double) YY[Max_Time % Buf_Size ][i] / Scale);

    if ( strlen( Reseau->T[i].Name ) != 0 )
      printf("  (%s)\n", Reseau->T[i].Name );
    else
      printf("\n");
  }
  printf("Total execution time: %f", 
	 elapsed_time( t_debut, t_fin ) );

  Nb_Evt = 0;
  for ( i=0; i<Reseau->Nb_Trans; i++ )
    Nb_Evt += YY[ Max_Time % Buf_Size ][i];

  printf("\t\t(%f evt/sec)\n", (double)Nb_Evt
	 / elapsed_time( t_debut, t_fin ) );
}
