/************************************************************************/
/* Simulation of an On/Off Fluid source.				*/
/************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <values.h>
#include <math.h>
#include <strings.h>
#include "boolean.h"
#include "parseargs.h"
#include "law.h"

#define	TOOLNAME	"onoff_sim"
#define VERSION 	"v1.0"

typedef struct {
  Law_Desc	On_Law;
  Law_Desc	Off_Law;
  double	Peak_Rate;
} SOURCE;

/************************************************************************/
/* Structure describing a plotting range and its properties.		*/
/************************************************************************/
typedef struct {
  double	Start;
  double	Stop;
  double	Step;
  boolean	Queue;
  boolean	Log;
  boolean	Log10;
} RANGE_INFO;
#define DEFAULT_RANGE { 0.0, 10.0, 0.5, FALSE, FALSE, FALSE }

/* Structure for passing around stat tables */
typedef struct RESULT {
  int		Nb_Points;
  double	T_Final;
  double	Total_Empty;
  int*		N_Plus;
  int*		N_Minus;
  double*	Acc_Plus;
  double*	Acc_Minus;
} RESULT;

/* Global variables */
boolean	Debug = FALSE;
boolean	NewLine = TRUE;

/* Parsing procedure */
int Parse_Law_From_Args( char **argv, int index, int argc, Law_Desc The_Law );
extern double Sample( Law_Desc );

/* Procedures for analysis */
double	Load_Factor( SOURCE, double );
int	Nb_Pts(RANGE_INFO);
RESULT 	Simulate( SOURCE, double, int, double W0, RANGE_INFO, boolean );
void	Print_Results( RANGE_INFO, RESULT, double, double );
void	Print_Header_Info( SOURCE, double, int );

int main(int argc, char *argv[])
{
  static char		Result_File_Name[256] = "";
  int			i;
  boolean		Help = FALSE;
  boolean		Args_Ok = TRUE;
  boolean		File_Ok = TRUE;
  boolean		Value = FALSE;
  boolean		Traj = FALSE;
  RANGE_INFO		P_Range = DEFAULT_RANGE; /* Range for printing */
  double		x_Val;
  double		Rho;
  int			Nb_Cycles = 100000;
  double		Capacity = 1.0;
  double		W0 = 0.0;	/* initial workload */
  SOURCE		Source;
  RESULT		Result;

  /* Various initializations before reading arguments */
  Source.On_Law = New_Law();
  Source.Off_Law = New_Law();
  Source.Off_Law->Name = 'E';
  Source.On_Law->Name = 'E';
  Source.Off_Law->Parameters[0] = 1.0;
  Source.On_Law->Parameters[0] = 1.0;
  Source.Peak_Rate = 1.0;

  i = 1;

  if (argc>1) {
    do {
      if GET_STRING( "-o", Result_File_Name )

     /* Option "moment" not available. Computing moments while not
	keeping all the empirical distribution in memory seems
	to require accumulating auxiliary quantities (cf. Acc*) */
     /*      else if GET_INT_SWITCH( "-moment", Order, Moment, TRUE ) */

      else if GET_REAL( "-peak_rate", Source.Peak_Rate )
      else if GET_REAL( "-capacity", Capacity )
      else if GET_REAL( "-W0", W0 )
      else if GET_INT( "-nbcycles", Nb_Cycles )

      else if GET_REAL( "-xfrom", P_Range.Start )
      else if GET_REAL( "-xto", P_Range.Stop )
      else if GET_REAL( "-xstep", P_Range.Step )
      else if GET_REAL_SWITCH( "-xval", x_Val, Value, TRUE )
			       
      else if SWITCH( "-traj", Traj, TRUE )

      else if SWITCH( "-queue", P_Range.Queue, TRUE )
      else if SWITCH( "-log", P_Range.Log, TRUE )
      else if SWITCH( "-log10", P_Range.Log10, TRUE )
      else if SWITCH( "-nonewline", NewLine, FALSE )
      else if SWITCH( "-debug", Debug, TRUE )
      else if IS_VERSION		     

      else if ( 0 == strcmp( "-on_law", argv[i] ) ) {
	i++;
	i = Parse_Law_From_Args( argv, i, argc, Source.On_Law );
      }
      else if ( 0 == strcmp( "-off_law", argv[i] ) ) {
	i++;
	i = Parse_Law_From_Args( argv, i, argc, Source.Off_Law );
      }
      else if ( ( 0 == strcmp( "-h", argv[i] ) )
		|| ( 0 == strcmp( "-?", argv[i] ) ) ) {
	printf("Usage: %s [-o <file>]\n\t", TOOLNAME );
	printf("[-on_law <law>] [-off_law <law>] ");
	printf("[-peak_rate <real>] [-capacity <real>]\n\t[-W0 <real>] ");
	printf("[-nbcycles <int>]\n\t");
	printf("[-xfrom <real>] [-xto <real>] [-xval <real>] ");
	printf("[-xstep <real>]\n\t");
	printf("[-queue] [-log] [-nonewline] [-traj] ");
	printf("[-debug] [-h] [-?]\n");
	Args_Ok = FALSE;
	Help = TRUE;
	i++;
      }
      else
	i++; /* skip unknown argument */
    } while ( Args_Ok && ( i < argc ) );
  }

  if (Args_Ok) {

    if ( Value ) {
      P_Range.Start = x_Val;
      P_Range.Stop = x_Val;
    }

    if ( strlen(Result_File_Name) != 0 )
      File_Ok = ( NULL != freopen( Result_File_Name, "w", stdout ) );

    if ( !File_Ok ) {
      printf("%s: error: could not open file %s\n", 
	     TOOLNAME, Result_File_Name );
      return -1;
    }
    else { 

      Rho = Load_Factor( Source, Capacity );

      Print_Header_Info( Source, Capacity, Nb_Cycles );

      Result = Simulate( Source, Capacity,
			 Nb_Cycles,
			 W0,
			 P_Range,
			 Traj );

      if ( !Traj ) 
	Print_Results( P_Range, Result, 
		       Capacity, Source.Peak_Rate - Capacity );


      return 0;
    }
  }

  else if (Help)
    return 0;
  else
    return 1;
}

int Nb_Pts( RANGE_INFO Range )
{
  if ( Range.Step < 0.0 )
    return 0;
  else if ( Range.Stop == Range.Start )
    return 1;
  else if ( Range.Step == 0.0 )
    return 0;
  else {
    return 1 + (int)floor( ( Range.Stop - Range.Start ) / Range.Step );
  }
}


double Load_Factor( SOURCE	The_Source,
		    double	Capacity )
{
  double Res;
  double MeanOn;
  double MeanOff;

  MeanOff = Mean( The_Source.Off_Law );
  MeanOn = Mean( The_Source.On_Law );

  Res = ( MeanOn * The_Source.Peak_Rate ) / ( MeanOn + MeanOff ) / Capacity;

  return Res;
}
		       
int Parse_Law_From_Args( char **argv, int index, int argc,
			 Law_Desc The_Law )
{
  int		i;
  int		j;
  int		Nb_Params;
  boolean 	Error;
  boolean	Short;
  boolean 	Ok;
  double	Foo;
  
  i = index;
  Error = FALSE;
  Short = FALSE;

  The_Law->Name = argv[i][0];
  i++;

  if ( i < argc ) {
    switch( The_Law->Name ) {
    case 'C':
    case 'E':
    case 'B':
    case 'g':
    case 'G':
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else
	i++;
      break;
    case 'R':
    case 'U':
    case 'W':
    case 'N':
    case 'L':
    case 'p':
      The_Law->Parameters = (double*) 
	realloc ( The_Law->Parameters, 2*sizeof(double) );
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      } 
      else {
	i++;
	if ( i < argc ) {
	  if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[1] ) ) ) {
	    Error = TRUE;
	  }
	  else 
	    i++;
	} 
	else {
	  Short = TRUE;
	}
      }
      break;
    case 'O':
      The_Law->Parameters = (double*) 
	realloc ( The_Law->Parameters, 3*sizeof(double) );
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else {
	i++;
	if ( i < argc ) {
	  if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[1] ) ) ) {
	    Error = TRUE;
	  } 
	  else {
	    i++;
	    if ( i < argc ) {
	      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[2] ) ) ) {
		Error = TRUE;
	      } 
	    }
	  }
	}
	else 
	  Error = TRUE;
      }
      break;
    case 'H':
      /* the number of parameters is not known in advance... have to 	*/
      /* find it first.							*/
      Ok = TRUE;
      for ( j=i; Ok && ( j < argc ); j++ ) {
	Ok = ( 1 == sscanf( argv[j], "%lf", &Foo ) );
      }
      /* counting the number of floats read */
      if ( j < argc )
	j--;
      Nb_Params = (j-i);
      if ( Nb_Params %2 != 0 ) {
	Error = TRUE;
      } else {
	The_Law->Parameters = (double*) 
	  realloc ( The_Law->Parameters, (1+Nb_Params)*sizeof(double) );
	The_Law->Parameters[0] = (double)(Nb_Params);
	for( j=i; j<i+Nb_Params; j++ ) {
	  sscanf( argv[j], "%lf", &(The_Law->Parameters[1+j-i]) );
	}
	i = i+Nb_Params;
      }
      break;
    case 'D':
      /* skip the (a priori) "[" argument */
      i++;
      /* the number of parameters is not known in advance... have to 	*/
      /* find it first, looking for the next "]".			*/
      Ok = TRUE;
      for ( j=i; Ok && ( j < argc ); j++ ) {
	Ok = ( 0 != strcmp( argv[j], "]" ) );
      }
      /* counting the number of floats read */
      if ( j >= argc ) {
	Short = TRUE;
      }
      else {
	j--;
	Nb_Params = (j-i);
	The_Law->Parameters = (double*) 
	  realloc ( The_Law->Parameters, (1+2*Nb_Params)*sizeof(double) );
	The_Law->Parameters[0] = (double)(Nb_Params);
	for( j=0; j<Nb_Params; j++ ) {
	  sscanf( argv[i+j], "%lf", &(The_Law->Parameters[1+j]) );
	}
	/* skip "]" and "[" */
	i += Nb_Params+2;
	/* continue reading. Have to protect against short arg list */
	Ok = TRUE;
	for( j=0; Ok && (j<Nb_Params) && (i+j<argc); j++ ) {
	  Ok = ( 1 == sscanf( argv[i+j], "%lf", 
			      &(The_Law->Parameters[1+Nb_Params+j]) ) );
	}
	if ( !Ok ) {
	  Error = TRUE;
	  i = i+j-1;
	} 
	else if ( i+j >= argc ) {
	  Short = TRUE;
	}
	else {
	  /* skip final "]", hoping it is actually there */
	  i = i + Nb_Params + 1;
	}
      }
      break;
    case 'P':
      /* read first parameter */
      if ( 1 != sscanf( argv[i], "%lf", &(The_Law->Parameters[0] ) ) ) {
	Error = TRUE;
      }
      else {
	i++;
	/* recursively read two laws */
	The_Law->Law_Parameters = (Law_Desc*) calloc( 2, sizeof(Law_Desc) );
	The_Law->Law_Parameters[0] = New_Law();
	The_Law->Law_Parameters[1] = New_Law();
	i = Parse_Law_From_Args( argv, i, argc, The_Law->Law_Parameters[0] );
	i = Parse_Law_From_Args( argv, i, argc, The_Law->Law_Parameters[1] );
      }
    default:
      printf("%s: unknown law id (%c)\n", TOOLNAME, The_Law->Name );
    }
  }
  else {
    Short = TRUE;
  }

  if ( Short ) {
    printf("%s: error parsing law %c: not enough arguments.\n", 
	   TOOLNAME, The_Law->Name );
    i = index;
  }
  else if ( Error ) {
    printf("%s: error parsing law %c at argument %s.\n", 
	   TOOLNAME, The_Law->Name, argv[i] );
    i = index;
  }

  return i;
}

/* Procedure for the gathering of statistics */
void Record_Segment( double w0, double w1,
		     RANGE_INFO Range,
		     RESULT	Result,
		     boolean	Is_On )
{
  int		i_start;
  double	a;
  int		*N;
  double	*Acc;

  if ( Is_On ) {
    N = Result.N_Plus;
    Acc = Result.Acc_Plus;
  } 
  else {
    N = Result.N_Minus;
    Acc = Result.Acc_Minus;
  }

  /* Determine end points */
  if ( w0 < Range.Start ) {
    i_start = 0;
    a = Range.Start;
  }
  else {
    if ( w0 >= Range.Stop ) {
      i_start = Result.Nb_Points;
      a = Range.Stop;
    }
    else {
      i_start = 1 + (int) floor( ( w0 - Range.Start ) / Range.Step );
      a = Range.Start + ( i_start - 1 ) * Range.Step;
    }
  }
  N[i_start] ++;
  Acc[i_start] += ( a - w0 );

  if ( w1 < Range.Start ) {
    i_start = 0;
    a = Range.Start;
  }
  else {
    if ( w1 >= Range.Stop ) {
      i_start = Result.Nb_Points;
      a = Range.Stop;
    }
    else {
      i_start = 1 + (int) floor( ( w1 - Range.Start ) / Range.Step );
      a = Range.Start + ( i_start - 1 ) * Range.Step;
    }
  }
  N[i_start] --;
  Acc[i_start] += ( w1 - a );
}
		     
/* Simulation procedure */
RESULT Simulate( SOURCE 	Source, 
		 double 	Capacity, 
		 int		Nb_Cycles,
		 double		W0,
		 RANGE_INFO 	Range,
		 boolean	Traj )
{
  int		Periods;
  double	Cur_Time;
  double	T;
  double	T_Empty;
  double	Workload;
  double	New_Workload;
  double	Drift;
  double	Total_Empty;
  RESULT	Result;

  Periods = 0.0;
  Cur_Time = 0.0;
  Workload = W0;
  Total_Empty = 0.0;
  Drift = Source.Peak_Rate - Capacity;

  if ( !Traj ) {
    /* prepare host structure for storing results */
    Result.Nb_Points = Nb_Pts( Range );
    Result.N_Plus  = 
      (int*) calloc( (size_t)(1+Result.Nb_Points), sizeof(int) );
    Result.N_Minus = 
      (int*) calloc( (size_t)(1+Result.Nb_Points), sizeof(int) );
    Result.Acc_Plus  = 
      (double*) calloc( (size_t)(1+Result.Nb_Points), sizeof(double) );
    Result.Acc_Minus = 
      (double*) calloc( (size_t)(1+Result.Nb_Points), sizeof(double) );
  }

  do {
    /* Simulate on period */ 
    T = Sample( Source.On_Law );
    New_Workload = Workload + Drift * T;
    if ( New_Workload < 0.0 ) {
      Total_Empty += ( New_Workload / Drift );
      New_Workload = 0.0;
      if ( Traj ) {
	T_Empty = -Workload / Drift;
	printf("%f %f\n", Cur_Time+T_Empty, 0.0);
      }
    }
    Cur_Time += T;

    if ( Traj )
      printf("%f %f\n", Cur_Time, New_Workload);
    else 
      Record_Segment( Workload, New_Workload, Range, Result, TRUE );

    Workload = New_Workload;

    /* Simulate off period */
    T = Sample( Source.Off_Law );
    New_Workload = Workload - Capacity * T;
    if ( New_Workload < 0.0 ) {
      Total_Empty += ( - New_Workload / Capacity );
      New_Workload = 0.0;

      if ( Traj ) {
	T_Empty = Workload / Capacity;
	printf("%f %f\n", Cur_Time+T_Empty, 0.0);
      }
    }
    Cur_Time += T;
    
    if ( Traj )
      printf("%f %f\n", Cur_Time, New_Workload );
    else 
      Record_Segment( New_Workload, Workload, Range, Result, FALSE );

    Workload = New_Workload;

    Periods++;

  } while ( Periods < Nb_Cycles );

  Result.T_Final = Cur_Time;
  Result.Total_Empty = Total_Empty;

  return Result;
}


void Print_Header_Info( SOURCE Source, double Capacity, int Nb_Cycles )
{
    printf("# Statistics on the On/Off fluid queue.\n");
    printf("# On distribution:  ");
    Write_Law( stdout, Source.On_Law, 0 );
    printf("\n");
    printf("# Off distribution: ");
    Write_Law( stdout, Source.Off_Law, 0 );
    printf("\n");
    printf("# Peak rate:        %f\n", Source.Peak_Rate );
    printf("# Capacity:         %f\n", Capacity );
    printf("# Theoretical Load Factor Rho = %f\n", 
	   Load_Factor( Source, Capacity ) );
    printf("# %d On and Off periods.\n", Nb_Cycles );

}

void Dump_Stats( RANGE_INFO 	Range,
		 RESULT		Result )
{
  int	i;
  int	N_Plus = 0;
  int	N_Minus = 0;

  printf("# Final Time: %f\n", Result.T_Final );
  printf("# Interval               N+  [cum]      N-  [cum]     Acc+     Acc-\n");

  for ( i=0; i<=Result.Nb_Points; i++ ) {
    if ( i > 0 ) {
      N_Plus  = Result.N_Plus[i] - Result.N_Plus[i-1];
      N_Minus = Result.N_Minus[i] - Result.N_Minus[i-1];
    }
    else {
      N_Plus = Result.N_Plus[0];
      N_Minus = Result.N_Minus[0];
    }

    if ( i == 0 ) {
      printf("# ]-inf,%6.3f[		%3d  [%3d]	%3d [%3d]    %8.3f %8.3f\n",
	     Range.Start,
	     N_Plus,
	     Result.N_Plus[0],
	     N_Minus,
	     Result.N_Minus[0],
	     Result.Acc_Plus[0],
	     Result.Acc_Minus[0] );
    }
    else if ( i == Result.Nb_Points ) {
      printf("# [%6.3f,inf[		%3d  [%3d]	%3d [%3d]    %8.3f %8.3f\n",
	     Range.Stop,
	     N_Plus,
	     Result.N_Plus[i],
	     N_Minus,
	     Result.N_Minus[i],
	     Result.Acc_Plus[i],
	     Result.Acc_Minus[i] );
    }
    else {
      printf("# [%6.3f,%6.3f[	%3d  [%3d]	%3d [%3d]    %8.3f %8.3f\n",
	     Range.Start + (i-1)*Range.Step,
	     Range.Start + i*Range.Step,
	     N_Plus,
	     Result.N_Plus[i],
	     N_Minus,
	     Result.N_Minus[i],
	     Result.Acc_Plus[i],
	     Result.Acc_Minus[i] );
    }
  }
}

void Print_Results( RANGE_INFO 	P_Range,
		    RESULT	Result,
		    double	Capacity,
		    double	Drift )
{
  int		i;
  double	Val;
  double	Fbar;
  double	Q;
  double	*Res;

  /* prepare result table */
  Res = (double*) calloc( (size_t)(1+Result.Nb_Points), sizeof(double) );

  /* First change N counts into cumulative counts */
  for ( i = 0; i < Result.Nb_Points; i++ ) {
    Result.N_Plus[i+1] += Result.N_Plus[i];
    Result.N_Minus[i+1] += Result.N_Minus[i];
  }
  if ( Debug ) {
    Dump_Stats( P_Range, Result );
  }

  /* Next compute backwards Q */
  /* Res[i] will contain: ( Total Time >= x_i ) / Total Simulation Time */
  Q = 0.0;

  for ( i=Result.Nb_Points-1; i>=0; i-- ) {
    Q += ( P_Range.Step * Result.N_Plus[i+1]  + Result.Acc_Plus[i+1] ) 
      / Drift ;
    Q += ( P_Range.Step * Result.N_Minus[i+1] + Result.Acc_Minus[i+1] ) 
      / Capacity ;

    Res[i] = Q / Result.T_Final;
  }
  /* Finally, print it out */
  printf("# Average proportion of empty time: %8.6f\n",
	 Result.Total_Empty / Result.T_Final );
  printf("# Printing Range: [%6.3f,%8.3f]\n", P_Range.Start, P_Range.Stop );

  for ( i=0; i<Result.Nb_Points; i++ ) {
    Val = P_Range.Start + i * P_Range.Step;
    Fbar = Res[i];

    if ( !P_Range.Queue )
      printf("%10.6f %12.8f", Val, 1.0-Fbar );
    else {
      if ( !P_Range.Log && !P_Range.Log10 )
	printf("%10.6f %14.10e", Val, Fbar );
      else if ( P_Range.Log )
	printf("%10.6f %12.8e", Val, log(Fbar) );
      else /* Log10 == TRUE */
	printf("%10.6f %12.8e", Val, log10(Fbar) );
    }
    if ( NewLine )
      printf("\n");
  }

}

