#
# The "logger" package
#
######################################################################
# Needs the "system" module for file browsing
lappend auto_path "."

######################################################################
# Make_Logger_Window -
# A procedure to create a Logger window with a text area, equipped with
# menus (and buttons) to save/clear the text.
#
# The name of the window and its title are passed as parameters so that
# several loggers can be handled (e.g. output and error).
#
# Parameters:
# - winname   the name of the window, without the leading "."
# - title     the title to be given to the window
#
proc Make_Logger_Window { { winname {} } { title "LogTool" } } {
    global Logger

    # if winname is empty (the default), the logger is created inside .
    if { [string length $winname] == 0 } {
	set top .logger
	frame $top
	pack $top -expand true -fill both
    } else {
	set top .[string tolower $winname]
	
	if [catch {toplevel $top} errtext] {
	    switch -regexp $errtext {
		bad { set message "Bad window name $top" }
		.*already.* { set message "A Logger with name $winname is already running" }
		default { set message $errtext }
	    }
	    ReportError "Logger Error" $message
	    error "Cannot create logger window"
	}

	wm title $top $title
	set name_icon $title
	wm iconname $top $name_icon
    }

    # creating the menubar
    frame $top.topmenu

    # A "file" menu for I/O etc.
    menubutton $top.topmenu.file -text "File" -menu $top.topmenu.file.menu \
	-underline 0
    set m [menu $top.topmenu.file.menu -tearoff 1]
    $m add command -label "Clear" -underline 0 \
	-command "Clear_Logger $top"
    $m add separator
    $m add command -label "Save" -underline 0 -state disabled \
	-command "Save_Log $top 0"
    $m add command -label "Save As..." -underline 0 \
	-command "Save_Log $top 1"
    $m add separator
    $m add command -label "Print"   -underline 0 \
	-command {ReportError "Print Error" "Not implemented. Sorry!"}
    $m add separator
    $m add command -label "Quit" -underline 0 \
	-command "Quit_Logger $top"
    pack $top.topmenu.file -side left 

    # menu for text appearance
    menubutton $top.topmenu.text -text "Text" -menu $top.topmenu.text.menu \
	-underline 0
    set m [menu $top.topmenu.text.menu]
    $m add cascade -label "Font Family" -menu $m.fontfamily
    $m add cascade -label "Font Size" -menu $m.fontsize
    $m add cascade -label "Font Weight" -menu $m.fontweight
    set m [menu $top.topmenu.text.menu.fontfamily]    
    foreach f { fixed times helvetica courier lucida } {
	$m add radio -label "$f" -variable Logger($top.fontfamily) \
	    -value $f -command "Update_Text_Attributes $top"
    }
    set m [menu $top.topmenu.text.menu.fontsize]    
    for { set s 8 } { $s <= 16 } { incr s } {
	$m add radio -label "$s" -variable Logger($top.fontsize) \
	    -value $s -command "Update_Text_Attributes $top"
    }
    set m [menu $top.topmenu.text.menu.fontweight]    
    foreach w { normal medium bold } {
	$m add radio -label "$w" -variable Logger($top.fontweight) \
	    -value $w -command "Update_Text_Attributes $top"
    }
    pack $top.topmenu.text -side left

    # menu for parameter setting and debugging
    menubutton $top.topmenu.options -text "Options" \
	-menu $top.topmenu.options.m -underline 0
    menu $top.topmenu.options.m
    pack $top.topmenu.options -side left
    $top.topmenu.options.m add check -label "Toggle Buttons" \
	-variable Logger($top.buttons) \
	-command "Toggle_Buttons $top" \
	-underline 0

    pack $top.topmenu -side top -anchor w

    # creating the text canvas with its scrollbar
    text $top.t -relief raised -bd 2 \
	-yscrollcommand "$top.s set" \
	-width 60 -height 30 -wrap word -setgrid yes
    scrollbar $top.s -relief flat -command "$top.t yview"
    pack $top.s -side right -fill y
    pack $top.t -expand true -fill both

    # preparing a button frame
    frame $top.action
    button $top.action.ok -text CLOSE -command "Quit_Logger $top"
    button $top.action.refresh -text CLEAR \
	-command "Clear_Logger $top"
    button $top.action.save -text SAVE -command "Save_Log 1"
    pack $top.action.ok $top.action.refresh $top.action.save \
	-side left -fill x -padx 2
    pack $top.action

    # setting up default configurations and bindings for keys and tags
    set Logger($top.fontfamily) fixed
    set Logger($top.fontweight) medium
    set Logger($top.fontsize) 12
    set Logger($top.buttons) 1
    set Logger($top.status) saved

    Update_Text_Attributes $top

}

proc Update_Text_Attributes { win } {
    global Logger

    set font -*-$Logger($win.fontfamily)-$Logger($win.fontweight)-r-*-*-$Logger($win.fontsize)-*

    $win.t tag configure normal -font $font
    $win.t tag configure info -background grey95 -foreground purple \
	-font $font \
	-relief raised -borderwidth 1
    $win.t tag configure warning -background yellow -foreground red \
	-font $font \
	-relief ridge -borderwidth 1
}

######################################################################
# Log_Text -
# Logs a piece of text into some window with some style.
#
# Parameters:
# - win		the window name (without the leading dot) where to log
# - text	the text to log
# - tag		the style in which to log
#
proc Log_Text { win text { tag normal } } {
    global Logger

    set top [string tolower .$win]

    if ![winfo exists .$win] {
	if [catch {Make_Logger_Window $win "Logger($win)"} errtext] {
	    ReportError "Logging Error" "Could not log text: window could not be created."
	    return
	}
    }

    # if text empty, exit. Do it AFTER checking the window, so that 
    # the Logger may be created even with an empty string.
    if { "$text" == "" } { return }

    # if the window is hidden, pop it!
    if { [wm state [winfo toplevel $top]] == "iconic" } {
	wm deiconify [winfo toplevel $top]
    }
    
    $top.t insert insert "$text" $tag
    set Logger($top.status) modified

}

proc Toggle_Buttons { win } {
    global Logger

    if $Logger($win.buttons) {
	pack $win.action
    } else {
	pack forget $win.action
    }
}

proc Clear_Logger { win } {
    global Logger

    if ![string compare $Logger($win.status) "modified"] {
	switch [tk_dialog .question "Log Modified" \
		    "The log has been modified. Save it?" \
		    {warning} 0 {Save & Clear} {Clear, don't save} {Cancel}] {
			0 { if ![Save_Log $win 1] { return } }
			1 { } 
			2 { return }
		    }
    } 

    $win.t delete 0.0 end
    set Logger($win.status) saved

}

proc Quit_Logger { win } {
    global Logger

    if ![string compare $Logger($win.status) "modified"] {
	switch [tk_dialog .question "Log Modified" \
		    "The log has been modified. Save it?" \
		    {warning} 0 {Save & Quit} {Quit, don't save} {Cancel}] {
			0 { if [Save_Log $win 1] { Destroy_Logger $win } }
			1 { Destroy_Logger $win }
			2 { return }
		    }
    } else {
	Destroy_Logger $win 
    }
}

######################################################################
# Destroy_Logger -
# Removes a logger without asking. Cleans up entries in the Logger table
# When the logger is not a toplevel window, this means the application
# has to quit also.
#
# Parameters:
# - win		the logger frame to destroy
#
proc Destroy_Logger { win } {
    global Logger

    unset Logger($win.buttons) 
    unset Logger($win.status) 
    if [info exists Logger($win.filename)] {
	unset Logger($win.filename)
    }

    destroy [winfo toplevel $win]
}

######################################################################
# Save_Log -
# User dialog for saving the contents of the log into a file.
# A nice (future) feature would be to let the possibility of removing
# the info while saving. I don't do it now because I wish to avoid the
# cross product Save/Save As + With/Without...
#
# Parameters:
# - win		the logger frame to save
# - ask		a boolean controlling confirmation (for the "Save" menu).
#
proc Save_Log { win { ask 1 } } {
    global Logger

    set filefield $win.filename

    set types {
	{{Log File}       {.log}  }
	{{All Files}        *     }
    }

    if { !$ask && [info exists Logger($filefield)] } {
	Save_Log_To_File $win $Logger($filefield)
	return 1
    } else {
	set options "-defaultextension log -filetypes {$types} -parent $win"
	if [info exists Logger($filefield)] {
	    set options "$options -initialfile $Logger($filefield)"
	}

	set file_name [GetSaveFile $options]

	if { $file_name != "" } {
	    set Logger($filefield) $file_name
	    Save_Log_To_File $win $Logger($filefield)
	    $win.topmenu.file.menu entryconfigure 3 -state normal
	    return 1
	} else {
	    return 0
	}
    }
}

proc Save_Log_To_File { win filename } {
    global Logger

    if [catch {open $filename w} outfile] {
	ReportError "Save Error" "Could not open file $filename for writing."
	return
    }

    puts $outfile [$win.t get 1.0 end]
    close $outfile

    set Logger($win.status) saved
}

# the procedure ReportError is (re)defined here to improve the 
# "standaloneness" of the package 
proc ReportError { title message } {
    tk_dialog .error "$title" "$message" {error} 0 {Ok}
}

######################################################################
# The procedures LogResult and MakeLogger are devised to be used from
# an application that defines a variable "logstyle". According to this
# style, a different logger is created and used.
######################################################################
# LogResult -
# Logs a result into the current logger, given by the variable "logstyle"
#
# Parameters:
# - info	text to be logged in "info" style
# - result	text to be logged in "result" style (optional)
# - warning	text to be logged in "warning" style (optional)
#
proc LogResult { info { result {} } { warning {} } } {
    global logstyle LoggerId

    if ![info exists logstyle] {
	ReportWarning "Logging Warning" \
	    "No logging style defined. Using \"none\"."
	set logstyle none
    }

    switch $logstyle {
	none {
	    if [catch {toplevel .results}] { 
		destroy .results
		toplevel .results
	    }
	    wm title .results "$info"
	    if { [string length $warning] > 0 } {
		message .results.warning -text "$warning" -aspect 1000 \
		    -justify left -foreground red -background yellow
		pack .results.warning -side top
	    }
	    message .results.text -text "$result" -aspect 1000 -justify left
	    button .results.ok -text "Ok" -command { destroy .results }
	    pack .results.text .results.ok -side top
	}
	local {
	    if ![winfo exists .results] {
		Make_Logger_Window results "Result log of [tk appname]"
	    }
	    Log_Text results "$info\n" info
	    if { [string length $warning] > 0 } {
		Log_Text results "$warning\n" warning
	    }
	    Log_Text results "$result"
	}
	global {
	    # checking the existence of a running "logtool"
	    if [catch {send logtool Log_Text [list logger {}]} err] {
		# puts "Error Exec Send: $err"
		switch -glob $err {
		    "X server insecure*" {
			# no good. Have to change logstyle
			set logstyle local
			ReportWarning "Logging" "X server insecure. Switching to local logging and killing global logger."
			LogResult $info $result $warning
			# Try to kill global logger
			if [info exists LoggerId] {
			    catch [exec /bin/kill -9 $LoggerId >& /dev/null]
			}
		    }
		    "no application named*" {
			# Have to spawn a logger. 
			set LoggerId [MakeLogger]
			# wait a little to let the logger appear...
			after 1000
			# try again
			LogResult $info $result $warning
		    }
		    * {
			# ???
			ReportWarning "Logging" "Send Error not handled: $err"
		    }
		}
		# Return in any case. 
		return -1
	    }
	    send logtool [concat Log_Text [list logger "$info\n" info]]
	    if { [string length $warning] > 0 } {
		send logtool \
		    [concat Log_Text [list logger "$warning\n" warning]]
	    }
	    send logtool [concat Log_Text [list logger "$result"]]
	}
    }
}

######################################################################
# MakeLogger -
# Switching procedure to create a logger according to the prevailing
# logging style.
# Returns the id of the process in case of a remote logger, -1 otherwise.
#
# Parameters:
#
proc MakeLogger { } {
    global logstyle

    if ![info exists logstyle] {
	ReportError "Logging Error" "No logging style defined."
	return
    }

    switch $logstyle {
	local { 
	    Make_Logger_Window results "Results"
	}
	global { 
	    # checking that the X server is ok
	    if [catch {send [tk appname] ERS_Version} err] {
		switch -glob $err {
		    "X server insecure*" {
			# No good. Change logstyle and try again.
			set logstyle local
			ReportWarning "Logging" "X server insecure. Switching to local logging"
			MakeLogger
			return -1
		    }
		    * {
			# unknown error. Switching anyway.
			set logstyle local
			puts "Send Error: $err"
			ReportWarning "Logging" "Problem encountered with \"send\". X server may be insecure. Switching to local logging"
			MakeLogger
			return -1
		    }
		}
	    } else {
		# X server appears to be ok. Checking the existence of a 
		# running "logtool"
		if [catch {send logtool Log_Text [list logger {}]}] {
		    set Id [exec logtool -iconic -title "Result Logger" &]
		    # puts "Id -> $Id"
		    return $Id
		}
	    }
	}
	default {
	    # default (or "none"): do nothing
	    return -1
	}
    } 
}

