#
# The I/O Module for mced
#
# This module contains the procedures to read and write files
# - save/load in the graph format
# - save as exclusion graph
#
######################################################################
# write_graph -
# Saves the current Markov chain in the "mced" format.
# Note that the "weight" of nodes, inherited from "gred" is here although
# it is useless. Perhaps it will be rehabilitated later.
# Returns a boolean telling the success of the operation.
#
# Parameters:
# - win		the canvas containing the chain to save
# - file_name	the name of the file in which to save.
# - rewards	boolean telling whether rewards should be written (def=false)
#
proc write_graph { win file_name { rewards 0 } } {
    global node_list arc_list weight arc_proba arc_reward
    global DEBUG

    if [catch {set outfile [open $file_name w]} err] {
	ReportError "File Error" "Could not open file $file_name for writing."
	return 0
    }

    if $DEBUG {
	puts -nonewline "Saving graph in file $file_name..."
    }

    # renumbering all nodes
    set curnode 1
    foreach node $node_list {
	set no_number([lindex $node 0]) $curnode
	incr curnode 1
    }

    foreach node $node_list {
	set node_id [lindex $node 0]
	puts -nonewline $outfile "Node $no_number($node_id) "
	puts -nonewline $outfile [$win coords $node_id]
# 	regsub -all {[ ]} $name($node_id) "_WHITE_" the_name
# 	puts -nonewline $outfile " \"$the_name\""
	puts $outfile " $weight($node_id)"
    }

    foreach link $arc_list {
	set link_id [lindex $link 0]
	set orig_id [lindex $link 1]
	set dest_id [lindex $link 2]
	puts -nonewline $outfile "Arc "
	puts -nonewline $outfile "$no_number($orig_id) $no_number($dest_id) "
	puts -nonewline $outfile "$arc_proba($link_id) "
	puts -nonewline $outfile "$arc_reward($link_id) "
	puts $outfile [$win coords $link_id]
    }

    close $outfile

    if $DEBUG {
	puts " ...Done"
    }

    return 1
}

proc read_graph { win file_name merge { Xoff 0.0 } { Yoff 0.0 } } {
    global weight arc_proba arc_reward GeomParams
    global node_list arc_list state
    global DEBUG

    if [catch {open $file_name r} infile] {
	puts "Could not open file $file_name for reading"
	return 
    }

    if !$merge {
	Clear_Graph $win
    }

    if $DEBUG {
	puts -nonewline "Reading graph from $file_name..."
    }

    # convert offset into pixels
    set xoff [winfo pixels . ${Xoff}m]
    set yoff [winfo pixels . ${Yoff}m]

    while { [gets $infile line] >= 0 } {
	set args [split $line]
	switch [lindex $args 0] {
	    Node { 
		set xpos [expr $xoff+([lindex $args 2]+[lindex $args 4])/2]
		set ypos [expr $yoff+([lindex $args 3]+[lindex $args 5])/2]
		set newid [make_new_node $win $xpos $ypos]
		set weight($newid) [lindex $args 6]
		set no_id([lindex $args 1]) $newid
	    }
	    Arc { 
		set points [Shift_Line_Coords \
				[lrange $args 5 [llength $args]] \
				$xoff $yoff]
		set newid [eval {$win create line} $points -tag ARC \
			       -arrow last]
		set newlink [list $newid $no_id([lindex $args 1]) \
				 $no_id([lindex $args 2])]
		lappend arc_list $newlink
		set arc_proba($newid) [lindex $args 3]
		set arc_reward($newid) [lindex $args 4]
	    }
	}
    }

    # adjusting the size of the canvas
    set bbox [$win bbox NODE ARC]
    Resize_Graph $win [expr [lindex $bbox 2]+$GeomParams(object_distance)] \
	[expr [lindex $bbox 3]+$GeomParams(object_distance)]

    Update_Probas $win

    # override modification flag
    if !$merge {
	set state($win) saved
    }

    close $infile
    if $DEBUG {
	puts " ...Done"
    }
}

######################################################################
# Clear_Graph -
# User dialog procedure for saving the chain into a file.
# Returns a boolean meaning the success of the operation. Saving may 
# fail if something happened while writing to the file, or after a cancel.
#
# Parameters:
# - win		the canvas window to clean
#
proc Save_Graph { win } {
    global file_name state

    set types {
	{{Markov Chain File}       {.mc .mced}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getSaveFile -defaultextension mced \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension mced -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."

    set file_name [GetSaveFile $options]

    if { $file_name != "" } {
	if [write_graph $win $file_name] {
	    set state($win) saved
	    return 1
	} else {
	    return 0
	}
    } else {
	return 0
    }
}

proc Browse_File { } {
    global file_name 

    set types {
	{{markov Chain File}       {.mced}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getOpenFile -defaultextension pned \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension mced -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."

    set file_name [GetOpenFile $options]

    return $file_name
		  
}

proc Load_Graph { win } {
    global file_name GeomParams

    if [catch {toplevel .params}] { return }

    wm title .params "Load"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar file_name -width 20 \
	-relief sunken
    
    frame .params.offset
    label .params.offset.label -text "Offset (mm)" -width 20 -anchor w
    label .params.offset.xlabel -text "X=" -width 3 -anchor w
    entry .params.offset.xoffset -textvar GeomParams(xoffset) \
	-width 4 -relief sunken
    label .params.offset.ylabel -text "Y=" -width 3 -anchor w
    entry .params.offset.yoffset -textvar GeomParams(yoffset) \
	-width 4 -relief sunken
    pack .params.offset.label .params.offset.xlabel .params.offset.xoffset \
	.params.offset.ylabel .params.offset.yoffset -side left 

    frame .params.actions
    button .params.actions.browse -text "Browse" \
	-underline 0 \
	-command Browse_File
     button .params.actions.ok -text "Load" \
	-underline 0 \
	-command "destroy .params\n read_graph $win \$file_name 0 \$GeomParams(xoffset) \$GeomParams(yoffset) "
    button .params.actions.merge -text "Merge" \
	-underline 0 \
	-command "destroy .params\n read_graph $win \$file_name 1 \$GeomParams(xoffset) \$GeomParams(yoffset)"
    button .params.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy .params\n array set GeomParams {[array get GeomParams]}"

    pack .params.actions.browse .params.actions.ok \
	.params.actions.merge .params.actions.cancel \
	-side left

    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}

    pack .params.filename.label -side left
    pack .params.filename.field -side left -expand true -fill x
    pack .params.filename -side top -fill x
    pack .params.offset -side top
    pack .params.actions -side top

    grab .params

}

######################################################################
# Procedures for exporting the graph in other formats than the gred
# formats.
######################################################################
# write_matrix -
# Exports the current graph as a plain matrix.
#
# Parameters:
# - file_name	the name of the file in which to write the result
# - mode	the representation of the matrix: one of sparse/full
# - rewards	boolean telling whether rewards should be written
#
proc write_matrix { file_name { mode sparse } { rewards 0 } } {
    global node_list arc_list arc_proba arc_reward
    global DEBUG

    # checking because of the EXPLAIN mode
    if ![Is_Proba_Matrix] {
	ReportError "Analysis Error" "Not a Probability Matrix"
	return 0
    }

    set SPA [expr 0 == [string compare $mode sparse]]

    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Generating Markov Chain matrix (mode $mode)..."
    }

    # numbering the nodes
    set curnodenumber [llength $node_list]
    for { set i 0 } { $i < $curnodenumber } { incr i 1 } {
	set no_number([lindex [lindex $node_list $i] 0]) $i
    }

    # type of chain and representation
    puts -nonewline $outfile "discrete"
    if $rewards {
	puts -nonewline $outfile " reward"
    }
    if $SPA {
	puts $outfile " sparse"
    } else {
	puts $outfile ""
    }

    # number of nodes
    puts $outfile $curnodenumber

    # adjacency matrix
    foreach node $node_list {
	set orig [lindex $node 0]

	# Scan arcs looking for neighbors. Work is done for both
	# representations (full/sparse) although not everything is 
	# used in both.
	set neighbors {}
	set o_arcs {}
	set probas {}
	set rew_list {}
	foreach arc $arc_list {
	    if { $orig == [lindex $arc 1] } {
		lappend o_arcs $arc
		lappend neighbors [lindex $arc 2]
		lappend probas $arc_proba([lindex $arc 0])
		lappend rew_list $arc_reward([lindex $arc 0])
	    } 
	}
	if $DEBUG {
	    puts "Node $orig: [llength $neighbors] links ($neighbors)"
	}
	
	if $SPA {
	    set i 0
	    foreach ngb $neighbors {
		set pr [lindex $probas $i]
		puts -nonewline $outfile \
		    "$no_number($orig) $no_number($ngb) $pr"
		if $rewards {
		    set rw [lindex $rew_list $i]
		    puts $outfile " wr $rw"
		} else {
		    puts $outfile ""
		}
		incr i
	    }
	} else {
	    foreach othernode $node_list {
		set dest [lindex $othernode 0]
		set pr 0.0
		# have to find out the arc
		foreach arc $o_arcs {
		    if { $dest == [lindex $arc 2] } {
			set pr $arc_proba([lindex $arc 0])
		    } 
		}
		puts -nonewline $outfile " $pr"
	    }
	    puts $outfile ""
	}
    }
    if $SPA {
	puts $outfile "stop"
    }
    close $outfile

    if $DEBUG {
	puts " ...Done"
    }

    return 1
}

######################################################################
# Gen_Matrix -
# User interface procedure prompting for a file name and the current
# chain in the file in the MCL language.
#
# Parameters:
# - win		the graph canvas to save
# - mode	the matrix mode: full or sparse
# - rewards	boolean telling whether rewards should be saved
#
proc Gen_Matrix { win mode { rewards 0 } } {

    set types {
	{{Markov Chain Language Matrix}       {.mcl}  }
	{{All Files}        *             }
    }

#     set gen_file_name [tk_getSaveFile -defaultextension mcl \
# 			   -filetypes $types -parent .]
    set options "-defaultextension mcl -filetypes {$types} -parent ."

    set gen_file_name [GetSaveFile $options]

    if { $gen_file_name != "" } {
	write_matrix $gen_file_name $mode $rewards
    }
}

proc Gen_Postscript { win } {
    global file_name

    set types {
	{{Postscript File}       {.ps .eps .epsi}  }
	{{All Files}        *             }
    }
    
#     set ps_file_name [tk_getSaveFile -defaultextension ps \
# 			  -initialfile [Make_File_Name $file_name ps] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension ps -initialfile {[Make_File_Name $file_name ps]} -filetypes {$types} -parent ."

    set ps_file_name [GetSaveFile $options]

    if { $ps_file_name != "" } {
	$win postscript -file $ps_file_name 
    }

}
