#
# Procedures related with task graph generation
#
######################################################################
# Make_TG_Standard -
# Converts the Petri Net into a Task Graph with delays. Each transition
# becomes a task with the same coordinates. Places are converted into
# dependencies with delay equal to the marking. Since the convention for
# task graph is NOT to support timing for dependency links, a new task is
# created when there is non-zero time AND mark(s) in a place.
# The task graph is written into a file and the graph editor (tged) is called
# on the result.
# 
# Parameters:
# - win		the graph canvas containing the Petri Net
#
proc Make_TG_Standard { win } {
    global trans_list place_list source_list platra_list trapla_list 
    global mark TransTime PlaceTime source_rate name 
    global curzoomvalue
    global env
    global DEBUG

    # constant taken from tged...
    set task_radius 15

    set file_name "/tmp/pned[pid].tged"
    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Creating task graph in file $file_name..."
    }

    # computing the correction factor
    set scale [expr 100.0/$curzoomvalue]
    # renumbering all places and transitions
    set curtrans 1
    foreach trans [concat $source_list $trans_list] {
	set tr_number([lindex $trans 0]) $curtrans
	incr curtrans 1
    }
    set curplace 1
    foreach place $place_list {
	set pl_number([lindex $place 0]) $curplace
	incr curplace 1
    }

    foreach trans [concat $source_list $trans_list] {
	set trans_id [lindex $trans 0]
	puts -nonewline $outfile "Task $tr_number($trans_id) "

	set box [Scale_Line_Coords [$win coords $trans_id] $scale]
	set posx [expr ([lindex $box 0]+[lindex $box 2])/2]
	set posy [expr ([lindex $box 1]+[lindex $box 3])/2]
	set newbox [list \
			[expr $posx-$task_radius] [expr $posy-$task_radius] \
			[expr $posx+$task_radius] [expr $posy+$task_radius] ]
	puts -nonewline $outfile $newbox
	regsub -all {[ ]} $name($trans_id) "_WHITE_" the_name
	puts -nonewline $outfile " \"$the_name\" "
	if [Is_Type $trans_id SOURCE] {
	    puts $outfile "\{ name exponential mean $source_rate($trans_id) \}"
	} else {
	    puts $outfile "\{ $TransTime($trans_id) \}"
	}
    }
    # Scan places looking for marked AND timed ones. For them, a new task
    # is necessary, with a link with its input transition.
    foreach place $place_list {
	set place_id [lindex $place 0]
	if [expr { $mark($place_id) > 0 } && ![Has_Zero_Time $place_id] ] {
	    puts -nonewline $outfile "Task $curtrans "
	    # puts "PLAPLA $place_id [$win coords $place_id] (x$scale)"
	    set box [Scale_Line_Coords [$win coords $place_id] $scale]
	    set posx [expr ([lindex $box 0]+[lindex $box 2])/2]
	    set posy [expr ([lindex $box 1]+[lindex $box 3])/2]
	    set newbox [list \
			    [expr $posx-$task_radius] \
			    [expr $posy-$task_radius] \
			    [expr $posx+$task_radius] \
			    [expr $posy+$task_radius] ]
	    puts -nonewline $outfile $newbox
	    regsub -all {[ ]} $name($place_id) "_WHITE_" the_name
	    puts -nonewline $outfile " \"$the_name\" "
	    puts $outfile "\{ $PlaceTime($place_id) \}"

	    # make the link
	    set ins [Get_Place_Inputs $place_id]
	    set orig [lindex $ins 0]
	    set orig_id $tr_number($orig)
	    set dest_id $curtrans
	    puts -nonewline $outfile "Dep $orig_id $dest_id "

	    # Have to find input link.
	    foreach link $trapla_list {
		if [expr ( [lindex $link 1] == $orig ) \
			&& ( [lindex $link 2] == $place_id ) ] {
		    set res $link
		}
	    }
	    if ![info exists res] {
		ReportError "Conversion error" \
		    "Can't find link ($orig,$place_id)"
		return
	    }

	    set ptlist [Scale_Line_Coords [$win coords [lindex $res 0]] $scale]
	    # adjust ends and clip.
	    # puts "ORIORI $orig [$win coords $orig] (x$scale)"
	    set box [Scale_Line_Coords [$win coords $orig] $scale]
	    set posx2 [expr ([lindex $box 0]+[lindex $box 2])/2]
	    set posy2 [expr ([lindex $box 1]+[lindex $box 3])/2]
	    set ptlist [lreplace $ptlist 0 1 $posx2 $posy2]
	    set nbpts [llength $ptlist]
	    set ptlist [lreplace $ptlist [expr $nbpts-2] [expr $nbpts-1] \
			    $posx $posy]
	    set ptlist [ClipLine $ptlist $task_radius]
	    puts -nonewline $outfile $ptlist

	    # finally put the law: zero
	    puts $outfile " \{ name constant value 0.0 \}"

	    # Remember id and prepare number for next trans
	    set subst_id($place_id) $curtrans
	    incr curtrans
	}
    }    
    # Scan places for link creation
    foreach place $place_list {
	set place_id [lindex $place 0]
	set ins [Get_Place_Inputs $place_id]
	set outs [Get_Place_Outputs $place_id]
	set orig [lindex $ins 0]
	set dest [lindex $outs 0]
	set orig_id $tr_number($orig)
	set dest_id $tr_number($dest)

	if { $mark($place_id) > 0 } {
	    puts -nonewline $outfile "Prog "
	    # if timed and marked, override the origin
 	    if ![Has_Zero_Time $place_id] {
 		set orig_id $subst_id($place_id)
 	    }
	} else {
	    puts -nonewline $outfile "Dep "
	}

	puts -nonewline $outfile "$orig_id $dest_id"
	if { $mark($place_id) > 0 } {
	    puts -nonewline $outfile " $mark($place_id) "
	}

	# now compute path. Have to find input and output links
	foreach link $trapla_list {
	    if [expr ( [lindex $link 1] == $orig ) \
		    && ( [lindex $link 2] == $place_id ) ] {
		set res $link
	    }
	}
	if ![info exists res] {
	    ReportError "Conversion error" \
		"Can't find link ($orig,$place_id)"
	    return
	}

	set ptlist [Scale_Line_Coords [$win coords [lindex $res 0]] $scale]
	# puts -nonewline "Points = $ptlist"

	unset res
	foreach link $platra_list {
	    if [expr ( [lindex $link 2] == $dest ) \
		    && ( [lindex $link 1] == $place_id ) ] {
		set res $link
	    }
	}
	if ![info exists res] {
	    ReportError "Conversion error" \
		"Can't find link ($place_id,$dest)"
	    return
	}
	# puts "+ [$win coords [lindex $res 0]]"
	# if marked+timed, the "input" link has already been drawn
	if [expr { $mark($place_id) > 0 } && ![Has_Zero_Time $place_id] ] {
	    set ptlist [Scale_Line_Coords [$win coords [lindex $res 0]] \
			 $scale]
	    set box [Scale_Line_Coords [$win coords $place_id] $scale]
	    set posx2 [expr ([lindex $box 0]+[lindex $box 2])/2]
	    set posy2 [expr ([lindex $box 1]+[lindex $box 3])/2]
	    set ptlist [lreplace $ptlist 0 1 $posx2 $posy2]
	} else {
	    # concatenate arcs
	    set ptlist [concat $ptlist \
			    [Scale_Line_Coords [$win coords [lindex $res 0]] \
				 $scale] ]
	}
	    
	# Clip at ends
	set ptlist [ClipLine $ptlist $task_radius]

	puts -nonewline $outfile " $ptlist"
	unset res
	
	# finally put the law if Dep link
	if { $mark($place_id) == 0 } {
	    puts $outfile " \{ $PlaceTime($place_id) \}"
	} else {
	    puts $outfile ""
	}
    }

    close $outfile
    if $DEBUG {
	puts " ...Done"
    }

    # calling tg editor
    exec $env(ERSHOME)/Interfaces/Tged/tged $file_name &

}

######################################################################
# Create_TG_Developped -
# Converts the Petri Net into a developped task graph with a specified
# number of iterations. Transitions becomes a set of tasks arranged into
# a column. Places are converted into dependencies joining these tasks.
# The task graph is written into a file and the graph editor (tged) is called
# on the result.
# 
# Parameters:
# - height	the number of iterations to develop
#
proc Create_TG_Developped { height } {
    global trans_list place_list source_list platra_list trapla_list 
    global mark TransTime PlaceTime source_rate name
    global env object_distance
    global DEBUG

    if ![isNumber $height] { return }

    # constant changed from that of tged...
    set task_radius 5

    set x0 $object_distance
    set y0 [expr 2*$object_distance]

    set file_name "/tmp/pned[pid].tged"
    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Creating task graph in file $file_name..."
    }

    # renumbering all places and transitions
    set curtrans 1
    foreach trans [concat $source_list $trans_list] {
	set tr_number([lindex $trans 0]) $curtrans
	incr curtrans 1
    }
    set curplace 1
    foreach place $place_list {
	set pl_number([lindex $place 0]) $curplace
	incr curplace 1
    }

    set all_trans [concat $source_list $trans_list]
    set width [llength $all_trans]

    # create standard tasks. All of them, because they must be listed
    # before links in the graph file.
    for { set i 0 } { $i < $height } { incr i 1  } {
	foreach trans $all_trans {
	    set trans_id [lindex $trans 0]
	    set j $tr_number($trans_id)
	    puts -nonewline $outfile "Task [expr $i*$width+$j] "

	    set posx [expr $x0+$j*$object_distance]
	    set posy [expr $y0+$i*$object_distance]
	    set newbox [list \
			    [expr $posx-$task_radius] [expr $posy-$task_radius] \
			    [expr $posx+$task_radius] [expr $posy+$task_radius] ]
	    puts -nonewline $outfile $newbox
	    puts -nonewline $outfile [format " %s_%s" "\"T$i" "$j\""]
	    if [Is_Type $trans_id SOURCE] {
		puts $outfile " \{ name exponential mean $source_rate($trans_id) \}"
	    } else {
		puts $outfile " \{ $TransTime($trans_id) \}"
	    }
	}
    }
    # now making the top and bottom tasks
    set top_id [expr $height*$width+1]
    puts -nonewline $outfile "Task $top_id "
    set topx [expr $x0+(1+$width)*$object_distance/2]
    set topy [expr $y0-$object_distance]
    set newbox [list [expr $topx-$task_radius] [expr $topy-$task_radius] \
		    [expr $topx+$task_radius] [expr $topy+$task_radius] ]
    puts -nonewline $outfile $newbox
    puts $outfile " \"Top\" \{ name constant value 0 \}"

    set bot_id [expr $height*$width+2]
    puts -nonewline $outfile "Task $bot_id "
    set botx [expr $x0+(1+$width)*$object_distance/2]
    set boty [expr $y0+($height-1)*$object_distance+$object_distance]
    set newbox [list [expr $botx-$task_radius] [expr $boty-$task_radius] \
		    [expr $botx+$task_radius] [expr $boty+$task_radius] ]
    puts -nonewline $outfile $newbox
    puts $outfile " \"Bottom\" \{ name constant value 0 \}"

    # now create links
    for { set i 0 } { $i < $height } { incr i 1  } {    
	foreach place $place_list {
	    set place_id [lindex $place 0]
	    set the_mark $mark($place_id)

	    set ins [Get_Place_Inputs $place_id]
	    set outs [Get_Place_Outputs $place_id]
	    set orig [lindex $ins 0]
	    set orig_id [expr $tr_number($orig)+$i*$width]

	    # Find out what is the destination. If it is bottom, then
	    # the weight of the arc SHOULD include that of the destination
	    # However, this is not trivial to do if time is stochastic.
	    set dest [lindex $outs 0]
	    if { [expr $i + $the_mark] < $height } {
		set dest_id [expr $tr_number($dest)+($i+$the_mark)*$width]
	    } else {
		set dest_id $bot_id
	    }

	    puts -nonewline $outfile "Dep $orig_id $dest_id"

	    # now compute path.
	    set end1 [list [expr $x0+$tr_number($orig)*$object_distance] \
			  [expr $y0+$i*$object_distance]]
	    if { [expr $i + $the_mark] < $height } {
		set end2 [list [expr $x0+$tr_number($dest)*$object_distance] \
			      [expr $y0+($i+$the_mark)*$object_distance]]
	    } else {
		set end2 [list $botx $boty]
	    }

	    set xmid [expr ([lindex $end1 0]+[lindex $end2 0])/2]
	    # the middle is moved slightly to account for the possibility
	    # of superposed arcs going to bottom
	    if { [expr $i + $the_mark] >= $height } {
		set xmid [expr $xmid+5*($tr_number($dest)-$width/2.0)]
	    }

	    if { $the_mark == 0 } {
		if { [expr abs($tr_number($orig)-$tr_number($dest))] == 1 } {
		    set ymid [expr [lindex $end1 1]]
		} else {
		    set ymid [expr [lindex $end1 1]+$object_distance/3]
		}
	    } else {
		set ymid [expr ([lindex $end1 1]+[lindex $end2 1])/2]
	    }
	    set mid [list $xmid $ymid]
	    set end1 [eval clip $mid $end1 $task_radius]
	    set end2 [eval clip $mid $end2 $task_radius]
	    
	    set ptlist [concat $end1 [concat $mid $end2]]
	    puts -nonewline $outfile " $ptlist"

	    # finally put the time distribution
	    puts $outfile " \{ $PlaceTime($place_id) \}"
	}
    }

    # now the arcs from "Top"
    set orig_id $top_id
    foreach place $place_list {
	set place_id [lindex $place 0]
	set the_mark $mark($place_id)

	set ins  [Get_Place_Inputs $place_id]
	set orig [lindex $ins 0]
	set outs [Get_Place_Outputs $place_id]
	set dest [lindex $outs 0]

	for { set i 0 } { $i < $the_mark } { incr i 1 } {
	    set dest_id [expr $tr_number($dest)+$i*$width]

	    puts -nonewline $outfile "Dep $orig_id $dest_id"
	    
	    # now compute path.
	    set end1 [list $topx $topy]
	    if { $i < $height } {
		set end2 [list [expr $x0+$tr_number($dest)*$object_distance] \
			      [expr $y0+$i*$object_distance]]
	    } else {
		set end2 [list $botx $boty]
	    }

	    set xmid [expr ([lindex $end1 0]+[lindex $end2 0])/2]
	    set ymid [expr ([lindex $end1 1]+[lindex $end2 1])/2]
	    # the middle is moved slightly to account for the possibility
	    # of superposed arcs.
	    set xmid [expr $xmid+5*($tr_number($orig)-$width/2.0)]

	    set mid [list $xmid $ymid]
	    set end1 [eval clip $mid $end1 $task_radius]
	    set end2 [eval clip $mid $end2 $task_radius]
	    
	    set ptlist [concat $end1 [concat $mid $end2]]
	    puts -nonewline $outfile " $ptlist"

	    # finally put the time distribution
	    puts $outfile " \{ name constant value 0 \}"
	}
    }

    close $outfile
    if $DEBUG {
	puts " ...Done"
    }

    # calling tg editor
    exec $env(ERSHOME)/Interfaces/Tged/tged -ts $task_radius $file_name &

}

######################################################################
# Make_TG_Developped -
# Builds a user interface to specify the number of rows for the creation
# of a developped task graph.
#
# Parameters: none
#
proc Make_TG_Developped { } {

    if [catch {toplevel .params}] { return }

    wm title .params "Task Graph Parameters"

    frame .params.height
    label .params.height.label -text "Number of rows" -width 15 -anchor w
    entry .params.height.field -textvar height -width 10 \
	-relief sunken
    
    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-underline 0 \
	-command "destroy .params\nCreate_TG_Developped \$height"
    button .params.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy .params"
    
    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}

    pack .params.height.label .params.height.field -side left
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.height -side top -fill x
    pack .params.actions -side top

}
