#
# The I/O Module for teted
#
# This module contains the procedures to read and write files
# - save/load in the tetris format
#
######################################################################
# write_tetris - 
# Saves the current heap of pieces in the "teted" format, in a given file.
# The optional colorformat controls the way colors names are written.
# When it is "hexa", the format #rrggbb is used even if the original name
# is a string.
#
# Parameters:
# - file_name	name of the file in which to save
# - colorformat normal or hexa
#
proc write_tetris { file_name { colorformat normal } } {
    global Piece_Store cur_piece_nb cur_width state
    global weight color
    global DEBUG

    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Saving tetris in file $file_name..."
    }

    puts $outfile "Columns $cur_width\n" 

    for { set i 1 } { $i <= $cur_piece_nb } { incr i 1 } {
	puts $outfile "Piece $i"
	puts $outfile "Weight $weight($i)"
	if { $colorformat == "normal" } {
	    puts $outfile "Color  $color($i)"
	} elseif [string match "\#......" $color($i)] {
	    puts $outfile "Color  $color($i)"
	} else {
	    # have to translate
	    set rgb [winfo rgb .store.canvas $color($i)]
	    puts -nonewline $outfile "Color  \#"
	    puts $outfile [format "%03x%03x%03x" [lindex $rgb 0 ] \
			       [lindex $rgb 1 ]	[lindex $rgb 2 ] ]
	}
	    
	puts -nonewline $outfile "LoProf"
	for { set j 0 } { $j < $cur_width } { incr j 1 } {
	    set idx [lindex [lindex $Piece_Store($i) 0] [expr 2*$j]]
	    set val [lindex [lindex $Piece_Store($i) 0] [expr 2*$j+1]]
	    if {  $val != "eps" } {
		puts -nonewline $outfile " $idx $val"
	    }
	}
	puts $outfile ""
	puts -nonewline $outfile "HiProf"
	for { set j 0 } { $j < $cur_width } { incr j 1 } {
	    set idx [lindex [lindex $Piece_Store($i) 1] [expr 2*$j]]
	    set val [lindex [lindex $Piece_Store($i) 1] [expr 2*$j+1]]
	    if {  $val != "eps" } {
		puts -nonewline $outfile " $idx $val"
	    }
	}
	puts $outfile "\n"
    }

    close $outfile

    set state(.store) saved

    if $DEBUG {
	puts " ...Done"
    }
}

######################################################################
# read_tetris - 
# Reads a heap of pieces in the "teted" format, from a given file.
#
# Parameters:
# - win		unused
# - file_name	name of the file in which to save
# - merge	boolean; when true the new set is merged with the current
#
proc read_tetris { win file_name merge } {
    global Piece_Store weight color cur_piece_nb store_level cur_width
    global piece_height state exitcode
    global DEBUG

    if [catch {open $file_name r} infile] {
	ReportError "Load Error" "Could not open file $file_name for reading"
	return 
    }

    if $DEBUG {
	puts "Reading a heap of pieces from $file_name..."
    }

    # First of all, it is necessary to check the column issue.
    set width 0
    gets $infile line
    set args [split $line]
    if { [string compare [lindex $args 0] "Columns" ] != 0 } {
	# The input file does not specify the number of columns.
	# If this is is a "merge", I refuse to do it. I could accept at the
	# cost of reading the whole file first. This will be done later.
	if $merge {
	    ReportError "Load Error" \
		"The input file does not specify the number of column.\nCannot merge files."
	    return
	} else {
	    ReportWarning "Load Warning" \
		"Cannot find number of columns. Will have to guess it!"
	}
	# rewinding the file to put back the line in stream!
	close $infile
	set infile [open $file_name r]
    } else {
	# Here, the number of columns has been duly specified.
	set width [lindex $args 1]
	if $merge {
	    if { $width != $cur_width } {
		# widths have to be adjusted.
		switch [tk_dialog .adjust "Width Mismatch" "Incompatible widths\n(cur=$cur_width, new=$width).\nYou have to adjust sizes." {warning} 0 {Adjust} {Cancel}] {
		    0 { 
			for { } { $width != $cur_width } { } {
			    tkwait window [Make_Store_Editor]
			    # There is a global variable saying if the editor
			    # exited with cancel. Otherwise, no way out!
			    if [info exists exitcode] {
				if { $exitcode == "cancel" } { 
				    unset exitcode
				    return
				}
			    }
			}
		    }
		    1 { return }
		}
	    }
	} else {
	    set cur_width $width
	}
    }

    if [expr { !$merge } && [winfo exists .store]] {
	set cur_piece_nb 0
	set level 0
	set store_level(0) 0
	Clear_Store 0 end 
    } else {
	set level $store_level($cur_piece_nb)
    }
    set max $cur_piece_nb

    while { [gets $infile line] >= 0 } {
	# eliminate multiple separators
	regsub -all {[ 	]+} $line " " line
	set args [split $line \ \t]
	switch [lindex $args 0] {
	    \#* {
		# skipping line
		puts "Skipping comment $line"
	    }
	    Piece { 
		set cur_nb [expr $cur_piece_nb+[lindex $args 1]]
		if { $max < $cur_nb } { set max $cur_nb }
		# initialize profiles to epsilon
		for { set i 0 } { $i < $cur_width } { incr i 1 } {
		    set loprof($i) eps
		    set hiprof($i) eps
		}
	    }
	    Weight { 
		set weight($cur_nb) [lindex $args 1]
	    }
	    Color { 
		set color($cur_nb) [lindex $args 1]
	    }
	    LoProf { 
		array set loprof [lrange $args 1 [expr [llength $args]-1] ]
	    }
	    HiProf { 
		array set hiprof [lrange $args 1 [expr [llength $args]-1] ]
		set piece [list [array get loprof] [array get hiprof] ]
		set Piece_Store($cur_nb) $piece
		set piece_height($cur_nb) [high_point $piece]
		incr level $piece_height($cur_nb)
		incr level 2
		set store_level($cur_nb) $level

		if $DEBUG {
		    puts "Got piece $Piece_Store($cur_nb) -> level $store_level($cur_nb)"
		}
	    }
	}
    }

    set cur_piece_nb $max

    if { $width == 0 } {
	# guessing width from the pieces
	for { set i 1 } { $i <= $cur_piece_nb } { incr i 1 } {
	    set piece_w [expr [llength [lindex $Piece_Store($i) 0]]/2]
	    if { $piece_w > $width } { set width $piece_w } 
	}
    }

    set cur_width $width

    if $DEBUG {
	puts " ...Done. Now have $cur_piece_nb pieces of width $width. Displaying..."
    }

    Make_Canvas
    # see if store big enough...
    Expand_Store $level
    Display_Store 1 $cur_piece_nb

    close $infile
}

proc Save_Tetris { } {
    global file_name

    set types {
	{{Tetris File}       {.teted .tet}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getSaveFile -defaultextension teted \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension teted -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."

    set file_name [GetSaveFile $options]

    if { $file_name != "" } {
	write_tetris $file_name
	return 1
    } else {
	return 0
    }
}

proc Browse_File { } {
    global file_name

    set types {
	{{Tetris File}       {.teted .tet}  }
	{{All Files}        *             }
    }

#     set new_file_name [tk_getOpenFile -defaultextension teted \
# 			   -initialfile [file tail $file_name] \
# 			   -filetypes $types -parent .]
    set options "-defaultextension tet -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."

    set new_file_name [GetOpenFile $options]
    
    if { $new_file_name != "" } {
	set file_name $new_file_name
    }
}

proc Load_Tetris { win } {
    global file_name

    if [catch {toplevel .params}] { return }

    wm title .params "Load"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar file_name -width 40 \
	-relief sunken
    
    frame .params.actions
    button .params.actions.browse -text "Browse" \
	-underline 0 \
	-command "Browse_File"
    button .params.actions.ok -text "Load" \
	-underline 0 \
	-command "destroy .params\n read_tetris $win \$file_name 0"
    button .params.actions.merge -text "Merge" \
	-underline 0 \
	-command "destroy .params\n read_tetris $win \$file_name 1"
    button .params.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy .params"

    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}

    pack .params.filename.label -side left
    pack .params.filename.field -side left -expand true -fill x
    pack .params.actions.browse .params.actions.ok \
	.params.actions.merge .params.actions.cancel -side left
    pack .params.filename -side top -expand true -fill x 
    pack .params.actions -side top

    grab .params

}

######################################################################
# write_interaction_graph -
# Exporting the interaction graph of the heap of pieces in the corresponding
# format. Two pieces are in interaction when they span over a common 
# column.
#
# Parameters:
# - file_name	the name of the file into which the graph is written
#
proc write_interaction_graph { file_name } {
    global weight cur_piece_nb Piece_Store cur_width
    global DEBUG

    if { $DEBUG } {
	puts -nonewline "Generating Interaction Graph..."
    }

    if { $cur_piece_nb == 1 } { return }

    # Checking height of pieces...
    if ![Is_Interaction_Graph] {
	ReportError "Export Error" \
	    "Some pieces have height > 1\nCannot make interaction graph"
	return BAD_HEIGHT
    }

    set outfile [open $file_name w]

    puts $outfile $cur_piece_nb

    for { set i 1 } { $i <= $cur_piece_nb } { incr i 1 } {
	puts -nonewline $outfile " $weight($i)"
    }
    puts $outfile ""

    # adjacency matrix
    for { set i 1 } { $i <= $cur_piece_nb } { incr i 1 } {
	array set Prof_i [lindex $Piece_Store($i) 0]
	for { set j 1 } { $j <= $cur_piece_nb } { incr j 1 } {
	    array set Prof_j [lindex $Piece_Store($j) 0]

	    if { $i == $j } {
		set interact 1 
	    } else {
		set interact 0
		for { set k 0 } { [expr { !$interact } && { $k < $cur_width } ] } \
		    { incr k }  {
			set interact \
			    [expr { [string compare $Prof_i($k) eps] != 0 } \
				 && { 0 != [string compare $Prof_j($k) eps] } ] 
		    }
	    }
		
	    if { $interact } {
		puts -nonewline $outfile " 1"
	    } else {
		puts -nonewline $outfile " 0"
	    }
	}
	puts $outfile ""
    }
    close $outfile

    if $DEBUG {
	puts " ...Done"
    }
}

######################################################################
# gen_interaction_graph -
# User interface for specifying a file name into which to write an
# the interaction graph of the game.
#
# Parameters:
# - gr_file_name	a proposed name for the file
#
proc gen_interaction_graph { gr_file_name } {

    set types {
	{{Graph file}       {.gr .gred .graph}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getSaveFile -defaultextension gr \
# 		       -initialfile [file tail $gr_file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension gred -initialfile {$gr_file_name} -filetypes {$types} -parent ."

    set gen_file_name [GetSaveFile $options]

    if { $gen_file_name != "" } {
	write_interaction_graph $gen_file_name
    }
}

######################################################################
# Gen_Interaction_Graph -
# Generate the interaction graph corresponding to a heap of pieces
# and calling the interaction graph editor (gred) on it.
#
# Parameters: none
#
proc Gen_Interaction_Graph { } {
    global env
    global KEEP

    set TMP_GR_FILE "/tmp/teted[pid].gr"
    
    set res [write_interaction_graph $TMP_GR_FILE]

    if { $res != "BAD_HEIGHT" } {
	# calling gr editor
	exec $env(ERSHOME)/Interfaces/Gred/gred -matrix $TMP_GR_FILE &

	if !$KEEP {
	    RemoveFile $TMP_GR_FILE
	}
    }
}

######################################################################
# Write_MaxPlus_Matrices -
# Procedure for effectively writing the (max,+) matrices of the pieces
# of the heap.
#
# Parameters:
# - file_name	the file into which to save
# - format	the matrix format (scilab|standard|sparse)
# - prefix	the name of the matrix (for the Scilab format)
#
proc Write_MaxPlus_Matrices { file_name format prefix } {
    global weight cur_piece_nb Piece_Store cur_width
    global DEBUG

    set SCI [expr 0 == [string compare $format scilab]]
    set SPA [expr 0 == [string compare $format sparse]]

    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Generating the set of (max,+) matrices...";
    }

    # preparing small window for controlling progress.
    if [catch {toplevel .progress}] { return }
    wm title .progress "Matrix Generation"
    label .progress.label -text "$cur_piece_nb matrices to generate"
    canvas .progress.view -background magenta \
	-height 10 -width 200
    set objid [.progress.view create rectangle 0 0 10 10 -fill blue]

    pack .progress.label -side top
    pack .progress.view -side top

    # number of matrices and their size
    if !$SCI {
	puts $outfile $cur_width
	puts $outfile $cur_piece_nb
    }

    if $SCI {
	puts -nonewline $outfile "Proba_$prefix = \["
    }
    # Writing out the weights, not probas (mat_sim allows that).
    for { set i 1 } { $i <= $cur_piece_nb } { incr i 1 } {
	puts -nonewline $outfile " $weight($i)"
    }
    if $SCI {
	puts $outfile " \]"
    } else {
	puts $outfile ""
    }

    # now the matrices
    for { set k 1 } { $k <= $cur_piece_nb } { incr k 1 } {

	array set LoProf [lindex $Piece_Store($k) 0]
	array set HiProf [lindex $Piece_Store($k) 1]
	# a message for patienting
	if $DEBUG {
	    puts "Matrix $k of $cur_piece_nb"
	}
	.progress.view coords $objid 0 0 \
	    [expr int(200*$k/$cur_piece_nb)] 10
	update

	if $DEBUG {
	    puts "Mat pour [array get LoProf] / [array get HiProf]"
	}

	if $SCI {
	    puts -nonewline $outfile "$prefix[expr $k-1] = \#(\[\n\t"
	} elseif $SPA {
	    puts $outfile "sp"
	}
 
	for { set i 0 } { $i < $cur_width } { incr i 1 } {
	    for { set j 0 } { $j < $cur_width } { incr j 1 } {
		if [expr { [string compare $LoProf($i) eps] != 0 } \
			  && { [string compare $LoProf($j) eps] != 0 } ] {
		    # a standard non-eps element
		    if $SPA {
			puts $outfile [format "%3d %3d %3d" $i $j \
					   [expr $HiProf($i)-$LoProf($j)]]
		    } else {
			puts -nonewline $outfile \
			    [format " %3d " [expr $HiProf($i)-$LoProf($j)]]
		    }
		} elseif { $i == $j } {
		    # the diagonal element: a 0.
		    if $SPA {
			puts $outfile [format "%3d %3d %3d" $i $i 0] 
		    } else {
			puts -nonewline $outfile "  0  "
		    }
		} elseif !$SPA {
		    if $SCI {
			puts -nonewline $outfile "  %0 "
		    } else {
			puts -nonewline $outfile " eps "
		    }
		}
	    }
	    # end of line.
	    if $SCI {
		if { $i < [expr $cur_width-1] } {
		    puts -nonewline $outfile " ;\n\t"
		} else {
		    puts -nonewline $outfile " ])\n"
		}
	    } elseif !$SPA {
		puts $outfile ""
	    }
	}
	if $SPA {
	    puts $outfile "end"
	} else {
	    puts $outfile ""
	}
    }

    close $outfile
    destroy .progress
    update 

    if $DEBUG {
	puts " ...Done"
    }

    return 1
}

######################################################################
# Gen_MaxPlus_Matrices -
# Builds a user dialog interface for specifying a file name and other
# parameters for the (max,+) matrix corresponding to the current sequence.
# This interface may be used for exporting the matrices of the heap of
# pieces, but also for the matrix of a sequence. The option "what" 
# tells which.
#
# Parameters: 
# - mat_file_name	a proposed name for the file
# - what	the matrix to make (heap|sequence)
#
proc Gen_MaxPlus_Matrices { mat_file_name { what heap } } {
    global gen_file_name matrix_format matrix_prefix

    set types {
	{{MaxPlus matrix}       {.mat}  }
	{{All Files}        *           }
    }

    set options "-defaultextension mat -initialfile {$mat_file_name} -filetypes {$types} -parent ."

    set gen_file_name $mat_file_name

    if [catch {toplevel .params}] { return }

    wm title .params "Generation Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar gen_file_name -width 20 \
	-relief sunken
    pack .params.filename.label -side left
    pack .params.filename.field -side left -expand true -fill x

    set f [frame .params.format]
    label $f.label -text "Format" -width 15 -anchor w
    radiobutton $f.standard -text "Standard" \
	-variable matrix_format -value normal
    radiobutton $f.sparse -text "Sparse" \
	-variable matrix_format -value sparse
    radiobutton $f.scilab -text "Scilab" \
	-variable matrix_format -value scilab
    pack $f.label $f.standard $f.sparse $f.scilab -side left -fill x

    frame .params.prefix
    label .params.prefix.label -text "Matrix Name" -width 15 -anchor w
    entry .params.prefix.field -textvar matrix_prefix -width 20 -relief sunken
    pack .params.prefix.label .params.prefix.field -side left -fill x

    set f [frame .params.actions]
    button $f.browse -text "Browse" -underline 0 \
	-command "set gen_file_name \[GetOpenFile \"$options\"\]"
    switch $what {
	heap {
	    button $f.ok -text "Save" -underline 0 \
		-command "destroy .params\n\Write_MaxPlus_Matrices \$gen_file_name \$matrix_format \$matrix_prefix"
	}
	sequence {
	    button $f.ok -text "Save" -underline 0 \
		-command "destroy .params\n\Compute_And_Write_Matrix \$gen_file_name \$matrix_format \$matrix_prefix"
	}
    }
    button $f.cancel -text "Cancel" -underline 0 \
	-command "destroy .params"
    pack $f.browse $f.ok $f.cancel -side left
    
    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}

    pack .params.filename .params.format .params.prefix -side top \
	-expand true -fill x
    pack .params.actions -side top

}

proc gen_postscript { win } {

    global ps_file_name

    if [catch {toplevel .params}] { return }

    wm title .params "Postscript Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar ps_file_name -width 20 \
	-relief sunken
    
    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\print_tetris \$win \$ps_file_name"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"
    
    pack .params.filename.label .params.filename.field -side left
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.filename -side top -fill x
    pack .params.actions -side top

}

######################################################################
# Export_Tetris -
# The general dispatching procedure for exporting the heap of pieces
# in whatever is not its native format "teted".
# Before calling the proper procedure, it fabricates a tentative file
# name from the current one using a type-specific extension.
#
# Parameters:
# - type	a string representing the type of object to obtain
# - extension	the extension to be given to the file name
#
proc Export_Tetris { type extension } {
    global file_name

    set name [Make_File_Name $file_name $extension]

    if { $type == "IG" } {
	if ![Is_Interaction_Graph] {
	    ReportError "Export Error" \
		"Some pieces have height > 1\nCannot make interaction graph"
	} else {
	    # saving into a file has been superseded by calling the editor.
	    # gen_interaction_graph $name
	    Gen_Interaction_Graph
	}
    } elseif { $type == "MAXPLUS" } {
	Gen_MaxPlus_Matrices $name heap
    } elseif { $type == "PS" } {
	gen_postscript $name
    } {
	ReportError "Export Error" "Exporting in $type not available yet. Sorry!"
    }
}    
